/******************************************************************************
 * $Header: /CommonBe/agmsmith/Programming/AGMSRAMFileSystem/RCS/AGMSFileSystemPacker.cpp,v 1.34 2003/09/02 20:31:32 agmsmith Exp $
 *
 * This is a BeOS program for packing file systems into a single file as well
 * as doing the opposite unpacking operation.  Besides handling files,
 * directories, symbolic links and attributes like conventional archiver
 * programs do, it also can store or recreate BeOS features like indices and
 * new oddities like AGMS links.  It can restore to an existing directory, or
 * it can create a RAM file system and restore the files there.  The packed
 * file format allows for files of unlimited length, supports compression, and
 * optionally can be in an uncompressed text format for manual editing.
 *
 * Note that this uses the AGMS coding style, not the OpenTracker one.  That
 * means no tabs, indents are two spaces, m_ is the prefix for member
 * variables, g_ is the prefix for global names, C or C++ style comments,
 * constants are in all capital letters and most other things are mixed case,
 * it's word wrapped to fit in 79 characters per line to make proofreading on
 * paper easier, and functions are listed in reverse dependency order so that
 * forward declarations (function prototypes with no code) aren't needed.
 *
 * $Log: AGMSFileSystemPacker.cpp,v $
 * Revision 1.34  2003/09/02 20:31:32  agmsmith
 * Added Modify Thing command.
 *
 * Revision 1.33  2003/08/29 21:46:46  agmsmith
 * Now restores the file metadata (date stamps, owner, permission flags).
 *
 * Revision 1.32  2003/08/27 18:29:34  agmsmith
 * Reading a Thing now starting to work.
 *
 * Revision 1.31  2003/08/21 17:54:54  agmsmith
 * Added the read volume functionality, and all that that implies.
 *
 * Revision 1.30  2003/08/12 01:10:43  agmsmith
 * Now ready for command processing code to be added.
 *
 * Revision 1.29  2003/08/11 23:30:32  agmsmith
 * Fixed up buffering code, added ReadNextLine.
 *
 * Revision 1.28  2003/08/09 21:51:39  agmsmith
 * Now with working decompression methods.
 *
 * Revision 1.27  2003/08/08 20:34:42  agmsmith
 * UnPacking under construction.
 *
 * Revision 1.26  2003/08/01 21:09:23  agmsmith
 * Renaming and sorting of recursive variables.
 *
 * Revision 1.25  2003/07/31 21:55:34  agmsmith
 * Write the volume description.
 *
 * Revision 1.24  2003/07/25 21:40:43  agmsmith
 * Spit and polish.
 *
 * Revision 1.23  2003/07/24 20:29:38  agmsmith
 * Now packs indices.
 *
 * Revision 1.22  2003/07/24 19:16:00  agmsmith
 * Now updates the progress indicator while packing.
 *
 * Revision 1.21  2003/07/23 03:08:12  agmsmith
 * Now packs attributes nicely!
 *
 * Revision 1.20  2003/07/23 00:04:33  agmsmith
 * Without the special text file handling.
 *
 * Revision 1.19  2003/07/22 23:57:21  agmsmith
 * Now dumps file data, including special handling of text files.
 *
 * Revision 1.18  2003/07/22 20:52:52  agmsmith
 * Under construction.
 *
 * Revision 1.17  2003/07/21 17:50:43  agmsmith
 * Under construction.
 *
 * Revision 1.16  2003/07/20 04:00:06  agmsmith
 * Under construction.
 *
 * Revision 1.15  2003/07/18 21:01:19  agmsmith
 * Added escape code processing for mount arguments.
 *
 * Revision 1.14  2003/07/17 17:34:51  agmsmith
 * Added file system mounting.
 *
 * Revision 1.13  2003/07/16 01:37:12  agmsmith
 * Added a command for just creating a file system, and GUI too.
 *
 * Revision 1.12  2003/07/13 01:52:00  agmsmith
 * Now with colour cycling arrow contents, and non-flickering backing
 * bitmap for the progress arrow image.
 *
 * Revision 1.11  2003/07/12 23:29:15  agmsmith
 * Now draws the outline of the progress arrow, and the percentage.
 *
 * Revision 1.10  2003/06/30 01:20:18  agmsmith
 * Now has all GUI controls installed.
 *
 * Revision 1.9  2003/06/29 21:57:38  agmsmith
 * Unpacking options under construction.
 *
 * Revision 1.8  2003/06/29 18:00:02  agmsmith
 * Simulates packing progress.
 *
 * Revision 1.7  2003/06/25 21:09:31  agmsmith
 * Under construction.
 *
 * Revision 1.6  2003/06/25 18:10:53  agmsmith
 * A few more buttons added.
 *
 * Revision 1.5  2003/06/21 20:08:35  agmsmith
 * More button creation.
 *
 * Revision 1.4  2003/06/19 01:24:08  agmsmith
 * Added packed file display and notes.
 *
 * Revision 1.3  2003/06/16 20:18:36  agmsmith
 * GUI under construction.  It compiles, but doesn't do anything.
 *
 * Revision 1.2  2003/06/15 14:31:33  agmsmith
 * Added a few more commands and options.
 *
 * Revision 1.1  2003/06/14 16:30:52  agmsmith
 * Initial revision
 */

/* Standard C Library. */

#include <ctype.h>
#include <errno.h>
#include <stdio.h>
#include <stdlib.h>
#include <unistd.h>
#include <zlib.h>

/* Standard C++ library. */

#include <iostream>

/* BeOS (Be Operating System) headers. */

#include <Alert.h>
#include <Application.h>
#include <Beep.h>
#include <Bitmap.h>
#include <Box.h>
#include <Button.h>
#include <CheckBox.h>
#include <Cursor.h>
#include <Directory.h>
#include <Entry.h>
#include <File.h>
#include <FilePanel.h>
#include <FindDirectory.h>
#include <fs_attr.h>
#include <fs_index.h>
#include <fs_info.h>
#include <Message.h>
#include <MessageQueue.h>
#include <Mime.h>
#include <Node.h>
#include <NodeInfo.h>
#include <Path.h>
#include <Point.h>
#include <PropertyInfo.h>
#include <RadioButton.h>
#include <Resources.h>
#include <String.h>
#include <StringView.h>
#include <SymLink.h>
#include <TextControl.h>
#include <View.h>



/******************************************************************************
 * Global variables, and not-so-variable things too.  Grouped by functionality.
 */

#define READERR_END_OF_FILE (B_ERRORS_END+1)
#define READERR_END_OF_NESTING (B_ERRORS_END+2)
#define READERR_UNKNOWN_FUTURE_THING (B_ERRORS_END+3)

static float g_MarginBetweenControls; /* Space of a letter "M" between them. */
static float g_LineOfTextHeight;      /* Height of text the current font. */
static float g_StringViewHeight;      /* Height of a string view text box. */
static float g_ButtonHeight;          /* How many pixels tall buttons are. */
static float g_CheckBoxHeight;        /* Same for check boxes. */
static float g_RadioButtonHeight;     /* Also for radio buttons. */
static float g_TextBoxHeight;         /* Ditto for editable text controls. */
static float g_RowHeight;             /* All controls fit in a row height. */

static const char *g_AFSPAppSignature =
  "application/x-vnd.agmsmith.AGMSFileSystemPacker";

static const char *g_AboutText =
"AGMSFileSystemPacker\n\n"

"This is a BeOS program for saving a whole file system and later recreating \
files, directories and modern (1990s) metadata (indices, AGMS links, \
symbolic links, attributes).  Its main purpose is to recreate an \
AGMSRAMFileSystem volume from saved data, but it can also be used to save \
and recreate ordinary directories, like conventional archive programs \
do.\n\n"

"Command line usage: Specify the operation (see the scripting commands list) \
as the first argument followed by more information as appropriate.  You may \
need to use several commands to get the desired effect.  The program's \
configuration will affect the actual operation (things like the name of the \
packed file, compression settings and so on).  In command line mode it will \
do the one operation and exit.  You can also use BeOS scripting (see the \
\"Hey\" command which you can get from http://www.bebits.com/app/2042) to \
control the program.  Finally, there's also a GUI interface which shows up \
if you start it without any command line arguments.\n\n"

"Written and copyright © by Alexander G. M. Smith, summer 2003.\n\n"

"This program is free software; you can redistribute it and/or modify it \
under the terms of the GNU Lesser General Public License as published by \
the Free Software Foundation; either version 2.1 of the License, or (at \
your option) any later version.\n\n"

"This program is distributed in the hope that it will be useful, but WITHOUT \
ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or \
FITNESS FOR A PARTICULAR PURPOSE.  See the GNU Lesser General Public \
License for more details.\n\n"

"You should have received a copy of the GNU Lesser General Public License \
along with this program; if not, write to the Free Software Foundation, \
Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA.\n\n"

"Send e-mail with suggestions for improvement to agmsmith@rogers.com.  If \
that address doesn't work, search the web for \"Alexander G. M. Smith\" to \
find my home page and contact info.\n\n"

"$Revision: 1.34 $\n"
"Compiled on " __DATE__ " at " __TIME__ ".";

static const char *g_AFSPTextFileMIMEType =
  "text/x-vnd.agmsmith.packed_file_system";

static const char *g_AFSPBinaryFileMIMEType =
  "application/x-vnd.agmsmith.packed_file_system_binary";

static const char *g_AFSPTextFileRecognitionString =
  "# AGMSFileSystemPacker Text File";

static const char *g_AFSPBinaryFileRecognitionString =
  "# AGMSFileSystemPacker Binary File";

static const char *g_DataName = "data";
static const char *g_ResultName = "result";

static const char *g_BeepPackDone = "AGMSFSPack Done";
static const char *g_UnknownText = "Unknown...";


static const char *g_SettingsDirectoryName =
  "AGMSFileSystemPacker";
static const char *g_SettingsFileName =
  "AGMSFileSystemPacker Settings";
static const uint32 g_SettingsWhatCode = 'AFSP';

static bool g_CommandLineMode;
  /* TRUE if the program was started from the command line (and thus should
  exit after processing the command), FALSE if it is running with a graphical
  user interface. */

static int g_QuitCountdown = -1;
  /* Set to the number of pulse timing events (about one every half second) to
  count down before the program quits.  Negative means stop counting.  Zero
  means quit at the next pulse event.  It also checks to see that there is no
  more work to do before trying to quit.  The idea is to process any pending
  scripting commands before quitting. */

static class CommanderLooper *g_CommanderLooperPntr = NULL;
static BMessenger *g_CommanderMessenger = NULL;
  /* Some globals for use with the looper which processes external commands
  (arguments received, file references received), needed for avoiding deadlocks
  which would happen if the BApplication sent a scripting message to itself. */

static BCursor *g_BusyCursor = NULL;
  /* The busy cursor, will be loaded from the resource file during application
  startup. */

typedef enum PropertyNumbersEnum
{
  PN_PACKED_FILE = 0,
  PN_TOP_LEVEL_DIRECTORY,
  PN_PACK,
  PN_UNPACK,
  PN_CREATE_FILE_SYSTEM,
  PN_COMPRESSION_METHOD,
  PN_ZLIB_COMPRESSION_LEVEL,
  PN_MAKE_RAM_DISK,
  PN_MAX_RAM_DISK_SIZE,
  PN_RAM_DISK_VIRTUAL_MEMORY,
  PN_RECREATE_INDICES,
  PN_NOTES,
  PN_RESET_TO_DEFAULTS,
  PN_INSTALL_THINGS,
  PN_MAX
} PropertyNumbers;

static char * g_PropertyNames [PN_MAX] =
{
  "PackedFile",
  "TopLevelDirectory",
  "Pack",
  "Unpack",
  "CreateFileSystem",
  "CompressionMethod",
  "ZLibLevel",
  "MakeRAMDisk",
  "MaxRAMDiskSize",
  "RAMDiskVirtualMemory",
  "RecreateIndices",
  "Notes",
  "ResetToDefaults",
  "InstallThings",
};

/* This array lists the scripting commands we can handle, in a format that the
scripting system can understand too. */

static struct property_info g_ScriptingPropertyList [] =
{
  /* *name; commands[10]; specifiers[10]; *usage; extra_data; ... */
  {g_PropertyNames[PN_PACKED_FILE], {B_GET_PROPERTY, 0},
    {B_DIRECT_SPECIFIER, 0}, "Get the pathname of the packed file.  This is "
    "the file which will contain the whole file system image.  "
    "The default name is something like B_USER_DIRECTORY / "
    "MyDiskImage.pafs", PN_PACKED_FILE, {}, {}, {}},
  {g_PropertyNames[PN_PACKED_FILE], {B_SET_PROPERTY, 0},
    {B_DIRECT_SPECIFIER, 0}, "Change the pathname of the packed file to "
    "use.  It will automatically be converted to an absolute path name, "
    "so make sure the parent directories exist before setting it.  "
    "If the packed file already exists, its header will be read and will be "
    "used to change a few settings.  So use this command first before trying "
    "to change the affected settings, which include the top level directory "
    "name, the option to create a RAM disk, the RAM disk maximum size, "
    "the option to recreate indices and the notes.  They all get set to make "
    "it easy to recreate an AGMSRAMFileSystem by default (or find out info "
    "about a packed file).",
    PN_PACKED_FILE, {}, {}, {}},
  {g_PropertyNames[PN_TOP_LEVEL_DIRECTORY], {B_GET_PROPERTY, 0},
    {B_DIRECT_SPECIFIER, 0}, "Get the pathname of the top level file system "
    "root directory.  When unpacking, the items inside the packed file will "
    "be recreated in this directory.  If you asked for a RAM disk to be "
    "created as part of unpacking, it will use this directory as its mount "
    "point.  When packing, the things in this directory (and all "
    "subdirectories) will be copied to the packed file.  Usually the root of "
    "a file system.",
    PN_TOP_LEVEL_DIRECTORY, {}, {}, {}},
  {g_PropertyNames[PN_TOP_LEVEL_DIRECTORY], {B_SET_PROPERTY, 0},
    {B_DIRECT_SPECIFIER, 0}, "Set the pathname of the top level directory.",
    PN_TOP_LEVEL_DIRECTORY, {}, {}, {}},
  {g_PropertyNames[PN_PACK], {B_EXECUTE_PROPERTY, 0},
    {B_DIRECT_SPECIFIER, 0}, "Packs the top level file system root directory "
    "and all the things inside it into a packed file, using the current "
    "settings (file names, compression levels, etc).  Can take a long while.",
    PN_PACK, {}, {}, {}},
  {g_PropertyNames[PN_UNPACK], {B_EXECUTE_PROPERTY, 0},
    {B_DIRECT_SPECIFIER, 0}, "Unpacks the packed file into the top level "
    "file system root directory (optionally creating a RAM file system "
    "there), using the current settings.  This recreates the things which "
    "had been originally copied into the pack file.  Can take a while.",
    PN_UNPACK, {}, {}, {}},
  {g_PropertyNames[PN_CREATE_FILE_SYSTEM], {B_EXECUTE_PROPERTY, 0},
    {B_DIRECT_SPECIFIER, 0}, "Creates a RAM file system mounted at the top "
    "level directory, using the current settings.",
    PN_CREATE_FILE_SYSTEM, {}, {}, {}},
  {g_PropertyNames[PN_COMPRESSION_METHOD], {B_GET_PROPERTY, 0},
    {B_DIRECT_SPECIFIER, 0}, "Get the compression method.  Currently returns "
    "\"None\" or \"ZLibDeflate\", others may be added later.  Only used "
    "when making a packed file.", PN_COMPRESSION_METHOD, {}, {}, {}},
  {g_PropertyNames[PN_COMPRESSION_METHOD], {B_SET_PROPERTY, 0},
    {B_DIRECT_SPECIFIER, 0}, "Set the compression method.",
    PN_COMPRESSION_METHOD, {}, {}, {}},
  {g_PropertyNames[PN_ZLIB_COMPRESSION_LEVEL], {B_GET_PROPERTY, 0},
    {B_DIRECT_SPECIFIER, 0}, "Get the zlib compression level.  It's a number "
    "from 0 (no compression, just data integrity) to 9 (max compression).  "
    "Only used when making a packed file with zlib deflate compression.",
    PN_ZLIB_COMPRESSION_LEVEL, {}, {}, {}},
  {g_PropertyNames[PN_ZLIB_COMPRESSION_LEVEL], {B_SET_PROPERTY, 0},
    {B_DIRECT_SPECIFIER, 0}, "Set the zlib compression level.",
    PN_ZLIB_COMPRESSION_LEVEL, {}, {}, {}},
  {g_PropertyNames[PN_MAKE_RAM_DISK], {B_SET_PROPERTY, 0},
    {B_DIRECT_SPECIFIER, 0}, "Change the boolean option to make a RAM disk "
    "when unpacking.  TRUE means that a AGMSRAMFileSystem disk will be "
    "created, using the top level directory as its mount point.  FALSE means "
    "that the unpacked files will be put in the top level directory.",
    PN_MAKE_RAM_DISK, {}, {}, {}},
  {g_PropertyNames[PN_MAKE_RAM_DISK], {B_GET_PROPERTY, 0},
    {B_DIRECT_SPECIFIER, 0}, "Get the state of the boolean option for making "
    "a RAM disk when unpacking.", PN_MAKE_RAM_DISK, {}, {}, {}},
  {g_PropertyNames[PN_MAX_RAM_DISK_SIZE], {B_SET_PROPERTY, 0},
    {B_DIRECT_SPECIFIER, 0}, "Change the maximum size of the RAM disk "
    "that's optionally created when unpacking.  Best set to at most the "
    "amount of real memory the computer has, since unusued space doesn't "
    "take up any memory.  In any case, it's limited to the kernel memory "
    "space, which has at most 800MB for allocation.  Zero means use the file "
    "system's internal default.",
    PN_MAX_RAM_DISK_SIZE, {}, {}, {}},
  {g_PropertyNames[PN_MAX_RAM_DISK_SIZE], {B_GET_PROPERTY, 0},
    {B_DIRECT_SPECIFIER, 0}, "Get the maximum RAM disk size setting.",
    PN_MAX_RAM_DISK_SIZE, {}, {}, {}},
  {g_PropertyNames[PN_RAM_DISK_VIRTUAL_MEMORY], {B_SET_PROPERTY, 0},
    {B_DIRECT_SPECIFIER, 0}, "Change the boolean option to enable or disable "
    "the use of virtual memory for the optional RAM disk.  TRUE means that "
    "it uses virtual memory.  FALSE means that it uses real memory (a scarce "
    "resource that can crash your system if you use too much).  Besides "
    "being faster, real memory is more secure since it won't get copied to "
    "the swap file (we hope).", PN_RAM_DISK_VIRTUAL_MEMORY, {}, {}, {}},
  {g_PropertyNames[PN_RAM_DISK_VIRTUAL_MEMORY], {B_GET_PROPERTY, 0},
    {B_DIRECT_SPECIFIER, 0}, "Get the state of the boolean option for using "
    "virtual memory for the optional RAM disk.",
    PN_RAM_DISK_VIRTUAL_MEMORY, {}, {}, {}},
  {g_PropertyNames[PN_RECREATE_INDICES], {B_SET_PROPERTY, 0},
    {B_DIRECT_SPECIFIER, 0}, "Change the boolean option to create indices on "
    "the disk volume that the top level directory resides on, when "
    "unpacking.  TRUE means recreate the indices, if they don't already "
    "exist.  FALSE means don't, so only existing indices are used.  In case "
    "you don't know, the indices are file system entities which have a list "
    "of all the files which use an attribute of the same name, so that you "
    "can quickly find the files which have a given attribute value.  The "
    "index needs to be created before the files with the indexed attribute "
    "are created, otherwise they don't get indexed.",
    PN_RECREATE_INDICES, {}, {}, {}},
  {g_PropertyNames[PN_RECREATE_INDICES], {B_GET_PROPERTY, 0},
    {B_DIRECT_SPECIFIER, 0}, "Get the state of the boolean option for "
    "recreating the indices when unpacking.", PN_RECREATE_INDICES, {}, {}, {}},
  {g_PropertyNames[PN_NOTES], {B_SET_PROPERTY, 0},
    {B_DIRECT_SPECIFIER, 0}, "Change the notes text associated with the "
    "packed file system file.  These notes will be saved in the packed file "
    "during packing and are displayed during unpacking.  They can be any "
    "UTF-8 encoded text you wish, including LF characters for multiple "
    "lines, ended with a NUL byte.",
    PN_NOTES, {}, {}, {}},
  {g_PropertyNames[PN_NOTES], {B_GET_PROPERTY, 0},
    {B_DIRECT_SPECIFIER, 0}, "Get the notes associated with the packed file.",
    PN_NOTES, {}, {}, {}},
  {g_PropertyNames[PN_RESET_TO_DEFAULTS], {B_EXECUTE_PROPERTY, 0},
    {B_DIRECT_SPECIFIER, 0}, "Resets all the configuration options to the "
    "default values.", PN_RESET_TO_DEFAULTS, {}, {}, {}},
  {g_PropertyNames[PN_INSTALL_THINGS], {B_EXECUTE_PROPERTY, 0},
    {B_DIRECT_SPECIFIER, 0}, "Sets up the MIME type for the packed file.  "
    "Only needs to be done once, usually during installation.  Will replace "
    "any existing MIME type settings, so if you have customized it (such as "
    "changing the preferred program), the changes will be lost.",
    PN_INSTALL_THINGS, {}, {}, {}},
  {0, {0}, {0}, 0, 0, {}, {}, {}} /* End of list of property commands. */
};


/* Standard questions and default answers.  The answer can be changed depending
on the user's request.  That allows them to answer "Always Yes" to the
"Overwrite existing thing" question.  The default answers are also stored in
the configuration file, so that they can be permanently changed. */

// bleeble;

/* The various compression methods as text and enums. */

typedef enum CompressionMethodEnum
{
  CM_NONE = 0,
  CM_ZLIB,
  CM_MAX
} CompressionMethod;

static const char * g_CompressionMethodNames [CM_MAX] =
{
  "None",
  "ZLib"
};



/******************************************************************************
 * Due to deadlock problems with the BApplication posting scripting messages to
 * itself, we need to add a second Looper.  Its job is to just to convert
 * command line arguments and arguments from the Tracker (refs received) into a
 * series of scripting commands sent to the main BApplication.  It also prints
 * out the replies received (to stdout for command line replies).  An instance
 * of this class will be created and run by the main() function, and shut down
 * by it too.
 */

class CommanderLooper : public BLooper
{
public:
  CommanderLooper ();
  ~CommanderLooper ();
  virtual void MessageReceived (BMessage *MessagePntr);

  void CommandArguments (int argc, char **argv);
  void CommandReferences (BMessage *MessagePntr);
  bool IsBusy ();

private:
  void ProcessArgs (BMessage *MessagePntr);
  void ProcessRefs (BMessage *MessagePntr);

  static const uint32 MSG_COMMAND_ARGUMENTS = 'CArg';
  static const uint32 MSG_COMMAND_FILE_REFS = 'CRef';

  bool m_IsBusy;
};



/******************************************************************************
 * A custom BView for displaying the progress arrow.  The arrow can be
 * inverted, and is drawn to fit the BView area.  A progress percentage text is
 * displayed on top of the arrow.
 */

class ProgressArrow : public BView
{
public:
  /* Constructor and destructor. */
  ProgressArrow (BRect NewBounds, bool PointingUp);
  ~ProgressArrow ();

  /* BeOS virtual functions. */
  virtual void AttachedToWindow ();
  virtual void Draw (BRect UpdateRect);
  virtual void Pulse ();

  /* User callable functions. */
  void SetProportion (float Proportion);

private:
  BBitmap *m_BackingBitmapPntr;
  BView m_BackingView;
    /* Used for making the off-screen picture of the arrow and the background
    around it.  Needed to avoid flicker when it redraws. */

  int m_Red, m_Green, m_Blue;
    /* The current drawing colour for the tip of the arrow.  The rest of the
    arrow is drawn with related colours. */

  int m_DRed, m_DGreen, m_DBlue;
    /* The delta colours, which control the colour cycling pattern.  Preferably
    should be coprime for a longer cycle.  Signs get reversed as the cycle hits
    the ends of the colour ranges. */

  BFont m_PercentageTextFont;
    /* Font to use for drawing the percentage text. */

  float m_PercentageTextHeightOffset;
    /* How much to place the baseline below the center of the view so that the
    total height of the letters (ascending plus descending) is centered. */

  bool m_PointingUp;
    /* TRUE for an arrow pointing upwards, FALSE for downwards. */

  float m_Proportion;
    /* Amount complete.  Varies fractionally from 0 to 1.  If it is between
    zero and one, the arrow animates.  If it is negative, the percentage number
    doesn't get drawn. */
};



/******************************************************************************
 * This view contains the various buttons and other controls for setting
 * configuration options.  Various subviews are created and deleted depending
 * on the operating mode.
 */

class ControlsView : public BView
{
public:
  /* Constructor and destructor. */
  ControlsView (BRect NewBounds);
  ~ControlsView ();

  /* BeOS virtual functions. */
  virtual void AttachedToWindow ();
  virtual void MessageReceived (BMessage *MessagePntr);
  virtual void Pulse ();

private:
  /* Various message codes for various buttons etc. */
  static const uint32 MSG_COMPRESSION_METHOD = 'Cmp0';
  static const uint32 MSG_DISPLAY_MODE_CREATE_FS = 'DMCF';
  static const uint32 MSG_DISPLAY_MODE_PACK = 'DMPk';
  static const uint32 MSG_DISPLAY_MODE_UNPACK = 'DMUP';
  static const uint32 MSG_DO_OPERATION_OR_CANCEL = 'DoOp';
  static const uint32 MSG_MAKE_RAM_DISK_CHECK = 'MRDK';
  static const uint32 MSG_MAX_RAM_DISK_SIZE_TEXT = 'MRST';
  static const uint32 MSG_NOTES_TEXT = 'NtsT';
  static const uint32 MSG_PACK_FILE_NAME_BROWSE = 'PFnB';
  static const uint32 MSG_PACK_FILE_NAME_TEXT = 'PFnT';
  static const uint32 MSG_RAM_VIRTUAL_MEMORY_CHECK = 'RVMT';
  static const uint32 MSG_RECREATE_INDICES_CHECK = 'RInK';
  static const uint32 MSG_TOP_LEVEL_PATH_BROWSE = 'TPaB';
  static const uint32 MSG_TOP_LEVEL_PATH_TEXT = 'TPaT';
  static const uint32 MSG_ZLIB_LEVEL_TEXT = 'ZLvT';

  /* The various different display modes.  In each mode different controls are
  displayed.  Switching between modes destroys all the old controls and creates
  the ones required for the new mode. */
  typedef enum DisplayModeEnum {
    DM_UNPACKING,
    DM_PACKING,
    DM_CREATE_FILE_SYSTEM,
    DM_MAX
  } DisplayModes;

  /* Different states the operation (pack / unpack) button can be in. */
  typedef enum OperButtonStatesEnum {
    OBS_DO_OPERATION,
    OBS_CANCEL,
    OBS_NOW_CANCELLING,
    OBS_MAX
  } OperButtonStates;

  /* Our member functions. */
  void BrowseForPackFile ();
  void BrowseForTopLevelDirectory ();
  void NullifyChildViewPointers ();
  void PollServerForChanges ();
  void ReadAllTextControls ();
  void SwitchToDisplayMode (DisplayModes NewMode);

  /* Member variables. */
  BButton          *m_AboutButtonPntr;
  class AFSPApp    *m_AppPntr;
  CompressionMethod m_CompressionMethodCachedValue;
  BRadioButton     *m_CompressionMethodRadioButtons [CM_MAX];
  OperButtonStates  m_DoOperationButtonCachedState;
  BButton          *m_DoOperationButtonPntr;
  BButton          *m_InstallThingsButtonPntr;
  bool              m_MakeRAMDiskCachedValue;
  BCheckBox        *m_MakeRAMDiskCheckboxPntr;
  int32             m_MaxRAMDiskSizeCachedValue;
  BTextControl     *m_MaxRAMDiskSizeTextboxPntr;
  BRadioButton     *m_ModeCreateFSRadioButtonPntr;
  BRadioButton     *m_ModePackRadioButtonPntr;
  BRadioButton     *m_ModeUnpackRadioButtonPntr;
  BString           m_NotesCachedValue;
  BTextControl     *m_NotesTextboxPntr;
  DisplayModes      m_NowShowing;
  BFilePanel       *m_PackFileBrowsePanelPntr;
  BButton          *m_PackFileBrowseButtonPntr;
  BString           m_PackFileNameCachedValue;
  BTextControl     *m_PackFileNameTextboxPntr;
  ProgressArrow    *m_ProgressArrowPntr;
  BTextControl     *m_ProgressPathTextboxPntr;
  float             m_ProgressProportionCachedValue;
  bool              m_RAMDiskVirtualMemoryCachedValue;
  BCheckBox        *m_RAMDiskVirtualMemoryCheckboxPntr;
  bool              m_RecreateIndicesCachedValue;
  BCheckBox        *m_RecreateIndicesCheckboxPntr;
  BButton          *m_ResetToDefaultsButtonPntr;
  bool              m_TextControlsContainInvalidData;
  bigtime_t         m_TimeOfLastPoll;
  BFilePanel       *m_TopLevelDirectoryBrowsePanelPntr;
  BButton          *m_TopLevelDirectoryBrowseButtonPntr;
  BString           m_TopLevelDirectoryNameCachedValue;
  BTextControl     *m_TopLevelDirectoryNameTextboxPntr;
  uint32            m_ZLibCompressionLevelCachedValue;
  BTextControl     *m_ZLibCompressionLevelTextboxPntr;
};



/******************************************************************************
 * The BWindow class for this program.  It displays the settings in real time,
 * and has various buttons and gadgets for changing settings (live changes
 * which reflect changes done by other programs too).  Most operations go
 * through the scripting message system, but it also peeks at the BApplication
 * data for examining simple things.
 */

class AFSPSettingsWindow : public BWindow
{
public:
  /* Constructor and destructor. */
  AFSPSettingsWindow ();

  /* BeOS virtual functions. */
  virtual void MessageReceived (BMessage *MessagePntr);
  virtual bool QuitRequested ();

  /* Various message codes. */
  static const uint32 MSG_FASTER_PULSE = 'FstP';
  static const uint32 MSG_SLOWER_PULSE = 'SloP';

private:
  /* Member variables. */
  ControlsView *m_ControlsViewPntr;
};



/******************************************************************************
 * ABSApp is the top level class for this program.  This handles messages from
 * the outside world (requests to pack or unpack files, change settings).  It
 * responds to command line arguments (if you start up the program a second
 * time, the system will just send the arguments to the existing running
 * program).  It responds to scripting messages.  And it responds to messages
 * from the window.  Its thread does the main work of tracking settings and
 * packing or unpacking.
 */

class AFSPApp : public BApplication
{
public:
  /* Constructor and destructor. */
  AFSPApp ();
  ~AFSPApp ();

  /* BeOS virtual functions. */
  virtual void AboutRequested ();
  virtual void ArgvReceived (int32 argc, char **argv);
  virtual status_t GetSupportedSuites (BMessage *MessagePntr);
  virtual void MessageReceived (BMessage *MessagePntr);
  virtual void Pulse ();
  virtual bool QuitRequested ();
  virtual void ReadyToRun ();
  virtual void RefsReceived (BMessage *MessagePntr);
  virtual BHandler *ResolveSpecifier (BMessage *MessagePntr, int32 Index,
    BMessage *SpecifierMsgPntr, int32 SpecificationKind, const char *Property);

private:
  static const int INPUT_OUTPUT_BUFFER_SIZE = 256 * 1024;
  static const int ERROR_MESSAGE_SIZE = PATH_MAX + 1024;

private:
  /* Our member functions. */
  void DefaultSettings ();
  status_t DoCreateFS (char *ErrorMessage);
  status_t DoPacking (char *ErrorMessage);
  status_t DoUnpacking (char *ErrorMessage);
  status_t DumpEscapedData (const char *DataPntr, int DataLength,
    bool Encode8Bit, int32 *SavedState);
  status_t DumpEscapedString (const char *StringPntr,
    int BytesAlreadyPrinted = 0, bool Encode8Bit = false);
  status_t DumpFlushRawData ();
  status_t DumpRawData (const char *DataPntr, int DataLength);
  status_t DumpRawString (const char *StringPntr);
  status_t DumpSwitchCompressionMode (CompressionMethod NewMethod);
  status_t DumpUnixStandardMetadata (bool DumpSize);
  status_t InstallThings (char *ErrorMessage);
public:
  status_t LoadSaveSettings (bool DoLoad);
private:
  status_t PackAttributes (BNode &CurrentNode);
  status_t PackDirectoryRecursively (BDirectory &CurrentNode);
  status_t PackFile (BFile &CurrentNode);
  status_t PackIndices (BDirectory &TopLevelDirectory);
  status_t PackSymbolicLink (BSymLink &CurrentNode, BDirectory &ParentNode);
  void ProcessScriptingMessage (BMessage *MessagePntr,
    struct property_info *PropInfoPntr);
  status_t ReadAndUnescapeValue (BString &ValueString);
  status_t ReadFillInputBuffer ();
  status_t ReadIndexCommand ();
  status_t ReadLinkCommand (bool CreateCommand);
  status_t ReadNextLine (BString &InitialText);
  status_t ReadNextParameterName (BString &ParameterName,
    bool SkipNestedBlocks = true);
  bool ReadMetaData (const char *NamePntr, const char *ValuePntr,
    BEntry *StatablePntr, bool IsAttribute, off_t ActualDataSize,
    const char *UserPathName, status_t &ErrorCode);
  status_t ReadThingCommand (bool CreateCommand);
  status_t ReadVolumeCommand (bool OverrideUserSettings = false);
  status_t RecursiveTotalSizeEstimate (BDirectory &CurrentNode);
  void UpdateNowProcessingPath (const char *StringPntr);

public:
  /* Member variables.  Many are asynchronously read by the window thread to
  see if it needs updating.  However, the other threads will lock the
  BApplication or use scripting commands if they want to make changes. */

  volatile bool m_CancelRequestedByGUI;
    /* The GUI will set this to TRUE when the user clicks the Cancel button, to
    stop a packing / unpacking operation in progress.  This BApplication will
    set it to FALSE when it cancels an operation. */

  CompressionMethod m_CompressionMethod;
    /* Which compression algorithm to use when creating the packed file
    contents.  Not relevant when unpacking. */

  bool m_MakeRAMDisk;
    /* TRUE to create a RAM disk when unpacking, using the top level directory
    as the mount point.  FALSE to just use the top level directory as-is. */

  int32 m_MaxRAMDiskSize;
    /* Maximum size limit of the RAM disk.  Zero or negative means use the file
    system's internal default. */

  BString m_Notes;
    /* User provided notes describing the packed file system.  UTF-8 encoding,
    can contain multiple lines of text. */

  char m_NowProcessingPath [B_PATH_NAME_LENGTH];
    /* Path name to the file currently being processed.  Not a BString so that
    other threads may read it without worrying about locking problems (though
    they may get the occasional mangled name). */

  off_t m_NowProcessingCurrentTotal;
    /* Number of bytes of data which have been packed or unpacked so far
    (counts uncompressed data bytes in files and attributes).  Set to a
    negative number if it isn't currently doing any processing.  Read
    asynchronously by other threads that display a progress indicator, so they
    may occasionally read a garbage value if half a word is updated while the
    other half isn't (since it is a 64 bit value which isn't written in one
    atomic operation on 32 bit processors). */

  off_t m_NowProcessingEstimatedFinalTotal;
    /* Total number of bytes to be processed, so that progress can be displayed
    by other threads.  Zero if it isn't known.  Calculated by adding up the
    sizes of files and counting non-file things (directories, links) as 1 byte
    of size.  Doesn't include attributes nor parent directory links.  An
    estimate since it is calculated by a scan of the files before the actual
    packing is done, and the files may be changed by other programs during
    packing.  For unpacking, it will be a lot more accurate. */

  BPath m_PackedFileName;
    /* Path to the packed file.  This file contains the whole file system (or
    directory tree) in Human readable text and (optionally) binary compressed
    data. */

  bool m_RAMDiskVirtualMemory;
    /* TRUE to use virtual memory for the RAM disk sometimes created when
    unpacking.  FALSE to use real memory. */

  bool m_RecreateIndices;
    /* TRUE to recreate indices specified in unpacking, FALSE to skip the
    instructions for creating indices. */

  struct RecursiveCallStaticDataStruct {

    bool blockDeviceWarningAlreadyIssued;
      /* Controls display of error messages which you only want to see once, to
      avoid having to click the OK button dozens of times. */

    int commandNestingLevel;
      /* When reading a packed archive, this keeps track of the current command
      block nesting level.  Zero at the highest level, 1 while inside the first
      command block, and temporarily incremented while inside each subcommand
      block by the ReadNextLine function (which also decrements it when it
      finds the end of the block parenthesis marker).  Mostly useful for
      skipping unrecognized blocks and their nested subblocks. */

    CompressionMethod currentCompressionMethod;
      /* Compression method currently being used to write the output.  Can be
      changed on the fly by calling DumpSwitchCompressionMode, which will flush
      data from intermediate buffers used by the previous compression method
      before switching to the new method. */

    BPath currentPathName;
      /* Absolute path to the item currently being dumped, or loaded. */

    int discardedPathPrefixLength;
      /* When packing, this is the number of characters to discard from an
      absolute path to get the relative path to the top level directory.  The
      relative path also should not start with a slash.  In other words, the
      first discardedPathPrefixLength characters of currentPathName are to be
      discarded when making relative paths for the packed archive.  When
      unpacking, this is the number of characters (and length of
      oldVolumePrefix) to be removed from the paths in the archive file, not
      including the slash after the volume name, so you may want to remove one
      more character before prepending m_TopLevelDirectoryName. */

    char *errorMessage;
      /* Points to the global error message string which will be displayed if
      an error occurs during a packing or unpacking operation.  Has space for
      ERROR_MESSAGE_SIZE bytes, which is at least as long as the maximum path
      name, plus a bit for the error text. */

    struct stat fileStatus;
      /* To avoid doing multiple stat() operations, the current file's data is
      stored here. */

    char *inputBuffer;
    int   inputLength;
    int   inputLengthExported;
    int   inputUnusedDataStart;
      /* An input buffer.  When writing a packed archive, it is used for
      reading the data from files before converting it to escape codes and
      writing it out, as well as a place for formatting strings.  When reading
      a packed archive, the decompressed archive data is put into this buffer
      as needed to keep it full.  Length is number of bytes already in the
      buffer, which is INPUT_OUTPUT_BUFFER_SIZE bytes long when it exists (NULL
      if it doesn't exist).  LengthExported is the length which is available
      for use by the next layer (varies between 0 and Length), since some of
      the data has to be kept aside for read ahead comparison to find the next
      compression header start and end, for switching compression mode
      purposes.  UnusedDataStart is the index to the first unused data in the
      buffer.  This is incremented, rather than moving the unused data to the
      front of the buffer (which can be several hundred kilobytes) every time a
      few bytes are used. */

    char *oldVolumePrefix;
      /* The path name prefix for the old volume name.  Does not include the
      trailing slash.  Used when reading in packed file system archives - all
      the things have this string as the initial part of their name.  The new
      absolute path name for the dearchived things is made by removing the
      oldVolumePrefix (after checking that it matches) and replacing it with
      the current m_TopLevelDirectoryName.  This string is malloc'd and free'd
      as needed. */

    char *outputBuffer;
    int   outputLength;
      /* When writing a packed archive file, this is used for the final lowest
      level of buffering before the data actually gets written to the archive
      file.  When reading a packed archive file, this is used for storing the
      results of de-escaping the data.  Length is number of bytes already in
      the buffer, which is INPUT_OUTPUT_BUFFER_SIZE bytes long when it exists
      (NULL if it doesn't exist). */

    char *rawDataBuffer;
    int   rawDataLength;
      /* When reading a packed archive, this buffer is used to hold the raw
      data (compressed or not), read directly from the packed archive file.
      The raw data is copied or decompressed from the rawDataBuffer into the
      inputBuffer as needed.  Not used when writing an archive.  Length is
      number of bytes already in the buffer, which is INPUT_OUTPUT_BUFFER_SIZE
      bytes long when it exists (NULL if it doesn't exist). */

    FILE *stream;
      /* The archive file being read or written. */

    bool zlibInitialised;
    z_stream zlibStream;
      /* For use with ZLib compression.  zlibInitialised if it has been set up,
      and thus needs closing operations before the operation finishes.
      Initialised or closed as compression methods are changed.  See z_stream
      in zlib.h for documentation about it. */

    void Clear ()
    {
      blockDeviceWarningAlreadyIssued = false;
      commandNestingLevel = 0;
      currentCompressionMethod = CM_NONE;
      currentPathName.Unset ();
      discardedPathPrefixLength = 0;
      errorMessage = NULL;
      memset (&fileStatus, 0, sizeof (fileStatus));
      inputBuffer = NULL;
      inputLength = 0;
      inputLengthExported = 0;
      inputUnusedDataStart = 0;
      oldVolumePrefix = NULL;
      outputBuffer = NULL;
      outputLength = 0;
      rawDataBuffer = NULL;
      rawDataLength = 0;
      stream = NULL;
      zlibInitialised = false;
      memset (&zlibStream, 0, sizeof (zlibStream));
    };

  } m_Recursive;
    /* Various semi-global variables used in the recursive calls to dump the
    file system.  Stored here rather than specifying them as unchanging
    recursive function arguments which would waste stack space. */

// bleeble; // application class variables

  BPath m_TopLevelDirectoryName;
    /* Path to the root directory where a tree of files is restored to, or
    saved from.  Also the mount point and implied volume name for the
    AGMSRAMFileSystem if one is being used. */

  BPath m_SettingsDirectoryPath;
    /* The constructor initialises this to the settings directory path.  It
    never changes after that. */

  bool m_SettingsHaveChanged;
    /* Set to TRUE when the settings are changed (different than the ones which
    were loaded).  When the application exits, the settings will be written out
    if they have changed. */

  uint32 m_zlibCompressionLevel;
    /* The compression level to use with the zlib library, 0 for none up to 9
    for maximum compression.  Tradeoff between compression speed and
    compactness. */
};



/******************************************************************************
 * Global utility function to display an error message and return.  The message
 * part describes the error, and if ErrorNumber is non-zero, gets the string
 * ", error code $X (standard description)." appended to it.  If the message
 * is NULL then it gets defaulted to "Something went wrong".  The title part
 * doesn't get displayed (no title bar in the dialog box, but you can see it in
 * the debugger as the window thread name), and defaults to "Error Message" if
 * you didn't specify one.  The error also gets printed to stderr.
 */

static void DisplayErrorMessage (
  const char *MessageString = NULL,
  int ErrorNumber = 0,
  const char *TitleString = NULL)
{
  BAlert *AlertPntr;
  char ErrorBuffer [PATH_MAX + 1500];

  if (TitleString == NULL)
    TitleString = "AGMSFileSystemPacker Error Message";

  if (MessageString == NULL)
  {
    if (ErrorNumber == 0)
      MessageString = "No error, no message, why bother?";
    else
      MessageString = "Something went wrong";
  }

  if (ErrorNumber != 0)
  {
    sprintf (ErrorBuffer, "%s, error code $%X/%d (%s) has occured.",
      MessageString, ErrorNumber, ErrorNumber, strerror (ErrorNumber));
    MessageString = ErrorBuffer;
  }

  cerr << TitleString << ": " << MessageString << endl;

  if (!g_CommandLineMode)
  {
    AlertPntr = new BAlert (TitleString, MessageString,
      "Acknowledge", NULL, NULL, B_WIDTH_AS_USUAL, B_STOP_ALERT);
    if (AlertPntr != NULL)
      AlertPntr->Go ();
  }
}



/******************************************************************************
 * Word wrap a long line of text into shorter 79 column lines and print the
 * result on the given output stream.
 */

static void WrapTextToStream (ostream& OutputStream, const char *TextPntr)
{
  const int LineLength = 79;
  char     *StringPntr;
  char      TempString [LineLength+1];

  TempString[LineLength] = 0; /* Only needs to be done once. */

  while (*TextPntr != 0)
  {
    while (isspace (*TextPntr))
      TextPntr++; /* Skip leading spaces. */
    if (*TextPntr == 0)
      break; /* It was all spaces, don't print any more. */

    strncpy (TempString, TextPntr, LineLength);

    /* Advance StringPntr to the end of the temp string, partly to see how long
    it is (rather than doing strlen). */

    StringPntr = TempString;
    while (*StringPntr != 0)
      StringPntr++;

    if (StringPntr - TempString < LineLength)
    {
      /* This line fits completely. */
      OutputStream << TempString << endl;
      TextPntr += StringPntr - TempString;
      continue;
    }

    /* Advance StringPntr to the last space in the temp string. */

    while (StringPntr > TempString)
    {
      if (isspace (*StringPntr))
        break; /* Found the trailing space. */
      else /* Go backwards, looking for the trailing space. */
        StringPntr--;
    }

    /* Remove more trailing spaces at the end of the line, in case there were
    several spaces in a row. */

    while (StringPntr > TempString && isspace (StringPntr[-1]))
      StringPntr--;

    /* Print the line of text and advance the text pointer too. */

    if (StringPntr == TempString)
    {
      /* This line has no spaces, don't wrap it, just split off a chunk. */
      OutputStream << TempString << endl;
      TextPntr += strlen (TempString);
      continue;
    }

    *StringPntr = 0; /* Cut off after the first trailing space. */
    OutputStream << TempString << endl;
    TextPntr += StringPntr - TempString;
  }
}



/******************************************************************************
 * Print the usage info to the stream.  Includes a list of all commands.
 */

ostream& PrintUsage (ostream& OutputStream); /* PPC needs a declaration. */
ostream& PrintUsage (ostream& OutputStream)
{
  struct property_info *PropInfoPntr;
  WrapTextToStream (OutputStream, g_AboutText);

  OutputStream << "\n\n"
    "Commands:\n"
    "\n"
    "Quit\n"
    "Stop the program.\n"
    "\n";

  /* Go through all our scripting commands and add a description of each one to
  the usage text. */

  for (PropInfoPntr = g_ScriptingPropertyList + 0;
  PropInfoPntr->name != 0;
  PropInfoPntr++)
  {
    switch (PropInfoPntr->commands[0])
    {
      case B_GET_PROPERTY:
        OutputStream << "Get " << PropInfoPntr->name << endl;
        break;

      case B_SET_PROPERTY:
        OutputStream << "Set " << PropInfoPntr->name << " NewValue" << endl;
        break;

      case B_COUNT_PROPERTIES:
        OutputStream << "Count " << PropInfoPntr->name << endl;
        break;

      case B_CREATE_PROPERTY:
        OutputStream << "Create " << PropInfoPntr->name << endl;
        break;

      case B_DELETE_PROPERTY:
        OutputStream << "Delete " << PropInfoPntr->name << endl;
        break;

      case B_EXECUTE_PROPERTY:
        OutputStream << PropInfoPntr->name << endl;
        break;

      default:
        OutputStream << "Buggy Command: " << PropInfoPntr->name << endl;
        break;
    }
    WrapTextToStream (OutputStream, PropInfoPntr->usage);
    OutputStream << endl;
  }

  return OutputStream;
}



/******************************************************************************
 * A utility function to send a command to the application, will return after a
 * short delay if the application is busy (doesn't wait for it to be executed).
 * The reply from the application is also thrown away.  It used to be an
 * overloaded function, but the system couldn't distinguish between bool and
 * int, so now it has slightly different names depending on the arguments.
 */

static void SubmitCommand (BMessage& CommandMessage)
{
  status_t ErrorCode;

  ErrorCode = be_app_messenger.SendMessage (&CommandMessage,
    be_app_messenger /* reply messenger, throw away the reply */,
    1000000 /* delivery timeout */);

  if (ErrorCode != B_OK)
    cerr << "SubmitCommand failed to send a command, code " <<
    ErrorCode << " (" << strerror (ErrorCode) << ")." << endl;
}


static void SubmitCommandString (
  PropertyNumbers Property,
  uint32 CommandCode,
  const char *StringArgument = NULL)
{
  BMessage CommandMessage (CommandCode);

  if (Property < 0 || Property >= PN_MAX)
  {
    DisplayErrorMessage ("SubmitCommandString bug.");
    return;
  }
  CommandMessage.AddSpecifier (g_PropertyNames [Property]);
  if (StringArgument != NULL)
    CommandMessage.AddString (g_DataName, StringArgument);
  SubmitCommand (CommandMessage);
}


static void SubmitCommandInt32 (
  PropertyNumbers Property,
  uint32 CommandCode,
  int32 Int32Argument)
{
  BMessage CommandMessage (CommandCode);

  if (Property < 0 || Property >= PN_MAX)
  {
    DisplayErrorMessage ("SubmitCommandInt32 bug.");
    return;
  }
  CommandMessage.AddSpecifier (g_PropertyNames [Property]);
  CommandMessage.AddInt32 (g_DataName, Int32Argument);
  SubmitCommand (CommandMessage);
}


static void SubmitCommandBool (
  PropertyNumbers Property,
  uint32 CommandCode,
  bool BoolArgument)
{
  BMessage CommandMessage (CommandCode);

  if (Property < 0 || Property >= PN_MAX)
  {
    DisplayErrorMessage ("SubmitCommandBool bug.");
    return;
  }
  CommandMessage.AddSpecifier (g_PropertyNames [Property]);
  CommandMessage.AddBool (g_DataName, BoolArgument);
  SubmitCommand (CommandMessage);
}



/******************************************************************************
 * Convert problematic characters into escape codes (RemoveBackslashEscapeCodes
 * has a list of the codes).  The resulting text is easy to edit in vi, or a
 * UTF-8 editor.  Returns 0 if successful, else an error code (usually because
 * of NULL pointers, or a too small output buffer).
 *
 * Input and output have to be different strings, since the output text is
 * usually larger than the input text (output buffer must be at least 6 bytes
 * long to hold the worst case escape code).  You can also have a NULL output
 * string pointer if you just wish to count up the amount of space the output
 * would take.  NUL bytes in the input string get converted to escape codes in
 * the output, which will never have NUL bytes (you'll have to add one if you
 * need it).  On input, source and destination lengths describe the sizes of
 * the buffers.  On output, they describe the number of bytes actually used.
 * Typically you should repeatedly call this function until the returned
 * DestinationLength is zero.  Initialise your state variable to zero for the
 * first iteration.  If you are not doing multiple iterations, you can pass in
 * a NULL StatePntr instead.  Internally the state is used for counting the
 * number of characters output since the last line break - an invisible line
 * break escape sequence (backslash followed by line feed) is inserted after
 * around 78 characters are output.  If you use TRUE for Encode8BitCharacters
 * then all bytes with the high bit set are output as numeric escape codes
 * (useful for arbitrary binary data).  If it is FALSE then they are output as
 * themselves (useful for UTF-8 strings).
 */

static status_t AddBackslashEscapeCodes (
  const char *SourceString,
  int *SourceLengthPntr,
  char *DestinationString,
  int *DestinationLengthPntr,
  int32 *StatePntr,
  bool Encode8BitCharacters)
{
  int                  InputChar;
  const unsigned char *InputPntr;
  int                  InputRemaining;
  bool                 OutputAllowed;
  unsigned char       *OutputPntr;
  int                  OutputRemaining;
  int                  OutputStringLength;
  const unsigned char *OutputString;
  unsigned char        TempString [6];
  int                  WrapCount;

  if (SourceString == NULL ||
  SourceLengthPntr == NULL || DestinationLengthPntr == NULL ||
  *SourceLengthPntr < 0 || *DestinationLengthPntr < 6)
  {
    DisplayErrorMessage ("ARFS:AddBackslashEscapeCodes: "
      "Called with invalid inputs", EINVAL);
    return EINVAL;
  }

  InputRemaining = *SourceLengthPntr;
  InputPntr = (const unsigned char *) SourceString;
  OutputRemaining = *DestinationLengthPntr;
  OutputPntr = (unsigned char *) DestinationString;
  OutputAllowed = (DestinationString != NULL);
  WrapCount = 0;
  if (StatePntr != NULL)
    WrapCount = *StatePntr;

  while (InputRemaining > 0)
  {
    InputChar = *InputPntr;
    switch (InputChar)
    {
      case 0x07:
        OutputString = (const unsigned char *) "\\a";
        OutputStringLength = 2;
        break;
      case 0x08:
        OutputString = (const unsigned char *) "\\b";
        OutputStringLength = 2;
        break;
      case 0x09:
        OutputString = (const unsigned char *) "\\t";
        OutputStringLength = 2;
        break;
      case 0x0A:
        OutputString = (const unsigned char *) "\\n";
        OutputStringLength = 2;
        break;
      case 0x0B:
        OutputString = (const unsigned char *) "\\v";
        OutputStringLength = 2;
        break;
      case 0x0C:
        OutputString = (const unsigned char *) "\\f";
        OutputStringLength = 2;
        break;
      case 0x0D:
        OutputString = (const unsigned char *) "\\r";
        OutputStringLength = 2;
        break;
      case '\\': /* Hide the backslashes to make later parsing easier. */
        OutputString = (const unsigned char *) "\\134";
        OutputStringLength = 4;
        break;
      case 0x7F: /* The DEL character, awkward to edit. */
        OutputString = (const unsigned char *) "\\177";
        OutputStringLength = 4;
        break;
      default:
        if (InputChar < 32 || (InputChar >= 128 && Encode8BitCharacters))
        {
          /* Use an octal number for this unprintable character. */

          if (InputRemaining >= 2 &&
          (InputPntr [1] < '0' || InputPntr [1] > '7'))
          {
            /* Can remove leading zeroes - the next letter isn't an octal digit
            so that won't cause confusion.  No octal in kernel, so can't use:
            sprintf (TempString, "\\%o", InputChar); */

            TempString [0] = '\\';
            if (InputChar < 8)
            {
              TempString [1] = InputChar + '0';
              OutputStringLength = 2;
            }
            else if (InputChar < 8 * 8)
            {
              TempString [1] = (InputChar >> 3) + '0';
              TempString [2] = (InputChar & 7) + '0';
              OutputStringLength = 3;
            }
            else
            {
              TempString [1] = (InputChar >> 6) + '0';
              TempString [2] = ((InputChar >> 3) & 7) + '0';
              TempString [3] = (InputChar & 7) + '0';
              OutputStringLength = 4;
            }
          }
          else /* Need the full 3 digit number to avoid ambiguity. */
          {
            /* Equivalent of: sprintf (TempString, "\\%03o", InputChar); */
            TempString [0] = '\\';
            TempString [1] = (InputChar >> 6) + '0';
            TempString [2] = ((InputChar >> 3) & 7) + '0';
            TempString [3] = (InputChar & 7) + '0';
            OutputStringLength = 4;
          }
          OutputString = TempString;
        }
        else /* Character doesn't need encoding. */
        {
          TempString [0] = InputChar;
          OutputStringLength = 1;
          OutputString = TempString;
        }
        break;
    }

    /* See if the output escape codes fit in the remaining buffer space.  Need
    to account for line wrap too.  Just stop if it doesn't fit.  No need to
    fiddle with partial escape codes (assume the output buffer is at least 6
    bytes long - backslash, LF, backslash, three octal digits). */

    if (OutputRemaining < OutputStringLength +
    ((WrapCount + OutputStringLength > 78) ? 2 : 0))
      break;

    if (WrapCount + OutputStringLength > 78)
    {
      /* Time for a line wrap.  Insert the backslash linefeed escape code. */
      if (OutputAllowed)
      {
        OutputPntr[0] = '\\';
        OutputPntr[1] = '\n';
      }
      OutputPntr += 2;
      OutputRemaining -= 2;
      WrapCount = 0;
    }

    if (OutputAllowed)
    {
      if (OutputStringLength == 1)
        *OutputPntr = *OutputString;
      else
        memcpy (OutputPntr, OutputString, OutputStringLength);
    }
    OutputPntr += OutputStringLength;
    OutputRemaining -= OutputStringLength;
    WrapCount += OutputStringLength;

    InputPntr++;
    InputRemaining--;
  }

  /* Return the amount of data processed back to the user. */

  *SourceLengthPntr = InputPntr - (const unsigned char *) SourceString;
  *DestinationLengthPntr = OutputPntr - (unsigned char *) DestinationString;
  if (StatePntr != NULL)
    *StatePntr = WrapCount;

  return 0;
}



/******************************************************************************
 * Process the escape codes in the input string and write the result to the
 * output string.  Returns 0 if successful, else an error code (usually because
 * of NULL pointers).
 *
 * Input and output can be the same string, but only safely on the first pass.
 * You can also have a NULL output string pointer if you just wish to count up
 * the amount of space the output would take.  NUL bytes in the strings work
 * OK, they get treated like any other character, not as end of string markers
 * (that's why buffer size is specified by a number for the input).  If you
 * want the output to be NUL terminated, include the NUL in the input string
 * and its length count.  On input, source and destination lengths describe the
 * sizes of the buffers.  On output, they describe the number of bytes actually
 * used.  Typically you should repeatedly call this function until the returned
 * DestinationLength is zero (it can read all the input and still have some
 * more output to produce if you happen to hit the end of the output buffer at
 * just the wrong time).  Initialise your state variable to zero for the first
 * iteration.  If you are not doing multiple iterations, you can pass in a NULL
 * StatePntr instead.  Internally the state contains a really short string with
 * the characters not processed from the last iteration, usually with the
 * backslash character in the front.
 *
 * Backslash (\) is the first character in an escape sequence.  The
 * following ones are cribbed from printf:
 *
 * \a      Write a <bell> 0x07 character (Alert noise).
 * \b      Write a <backspace> 0x08 character.
 * \f      Write a <form-feed> 0x0C character.
 * \n      Write a <new-line> 0x0A character.
 * \r      Write a <carriage return> 0x0D character.
 * \t      Write a <tab> 0x09 character.
 * \v      Write a <vertical tab> 0x0B character.
 * \num    Write an 8-bit character whose ASCII value is the 1-, 2-,
 * \0num   or 3-digit octal number num.  Preferred way of encoding backslashes.
 * \LF     Write nothing, used for making line breaks invisible.
 * \Other  Write the Other character.
 */

static status_t RemoveBackslashEscapeCodes (
  const char *SourceString,
  int *SourceLengthPntr,
  char *DestinationString,
  int *DestinationLengthPntr,
  int32 *StatePntr)
{
  unsigned int         i;
  int                  InputChar;
  const unsigned char *InputPntr;
  int                  InputRemaining;
  int                  LoopsWithoutAnInputAdvance;
  bool                 OutputAllowed;
  int                  OutputChar;
  unsigned char       *OutputPntr;
  int                  OutputRemaining;
  unsigned char        StateString [5];

  if (SourceString == NULL ||
  SourceLengthPntr == NULL || DestinationLengthPntr == NULL ||
  *SourceLengthPntr < 0 || *DestinationLengthPntr < 0)
  {
    DisplayErrorMessage ("ARFS:RemoveBackslashEscapeCodes: "
      "Called with invalid inputs", EINVAL);
    return EINVAL;
  }

  InputRemaining = *SourceLengthPntr;
  InputPntr = (const unsigned char *) SourceString;
  OutputRemaining = *DestinationLengthPntr;
  OutputPntr = (unsigned char *) DestinationString;
  OutputAllowed = (DestinationString != NULL);
  memset (StateString, 0, sizeof (StateString));
  if (StatePntr != NULL)
    memcpy (StateString, StatePntr, 4 /* string size 5 - last byte zeroed */);

  LoopsWithoutAnInputAdvance = 0;
  while (LoopsWithoutAnInputAdvance++ < 2)
  {
    /* Look ahead at the next character to see if the stuff accumulated in
    StateString should dump out a character (happens for a non-digit ending an
    octal number). */

    if (InputRemaining <= 0)
      InputChar = -1;
    else
      InputChar = *InputPntr;

    OutputChar = -1; /* Negative for no output character. */

    if (StateString [0] == 0)
    {
      /* No stored up characters from before.  NULs handled elsewhere. */
    }
    else if (StateString [0] != '\\')
      OutputChar = StateString [0]; /* Not an escape code, char is itself. */
    else /* Have an escape code building up in StateString. */
    {
      switch (StateString [1])
      {
        case 0: /* No second character available yet in the escape code. */
          break;

        case 'a': /* Write a <bell> 0x07 character (Alert noise). */
          OutputChar = 7;
          break;

        case 'b': /* Write a <backspace> 0x08 character. */
          OutputChar = 8;
          break;

        case 'f': /* Write a <form-feed> 0x0C character. */
          OutputChar = 12;
          break;

        case 'n': /* Write a <new-line> 0x0A character. */
          OutputChar = 10;
          break;

        case 'r': /* Write a <carriage return> 0x0D character. */
          OutputChar = 13;
          break;

        case 't': /* Write a <tab> 0x09 character. */
          OutputChar = 9;
          break;

        case 'v': /* Write a <vertical tab> 0x0B character. */
          OutputChar = 11;
          break;

        case '0':
        case '1':
        case '2':
        case '3':
        case '4':
        case '5':
        case '6':
        case '7': /* See if the end of the octal number has been reached. */
          if (StateString [2] != 0)
          {
            if (StateString [3] != 0)
            {
              /* Have a 3 digit octal number, no more digits allowed. */
              OutputChar =
                (StateString [1] - '0') * 8 * 8 +
                (StateString [2] - '0') * 8 +
                (StateString [3] - '0');
            }
            else if (InputChar >= 0 && (InputChar < '0' || InputChar > '7'))
            {
              /* Have a two digit number, followed by something else. */
              OutputChar =
                (StateString [1] - '0') * 8 +
                (StateString [2] - '0');
            }
          }
          else if (InputChar >= 0 && (InputChar < '0' || InputChar > '7'))
          {
            /* Have a one digit number, followed by something else. */
            OutputChar =
              (StateString [1] - '0');
          }
          break;

        case '\n': /* Write nothing, used for making line breaks invisible. */
          StateString [0] = 0; /* Poof!  The backslash linefeed vanishes. */
          break;

        default: /* Write the Other character. */
          OutputChar = StateString [1];
          break;
      }
    }

    /* Output the character derived from the current state contents. */

    if (OutputChar >= 0)
    {
      if (OutputRemaining <= 0)
        break; /* No more output space when we need it. */

      if (OutputChar > 255)
        OutputChar = 255; /* Peg large octal numbers at the max. */
      if (OutputAllowed)
        *OutputPntr++ = OutputChar;
      else
        OutputPntr++; /* Just counting, not a valid pointer. */
      OutputRemaining--;
      StateString [0] = 0;
    }

    if (InputChar == 0)
    {
      /* Special NUL case handling, since it can't be represented in the
      StateString, output it now. */

      if (OutputRemaining <= 0)
        break; /* No more output space when we need it. */

      if (OutputAllowed)
        *OutputPntr++ = 0;
      else
        OutputPntr++; /* Just counting, not a valid pointer. */
      OutputRemaining--;
      InputPntr++;
      InputRemaining--;
      LoopsWithoutAnInputAdvance = 0;
      /* Can clear StateString since the NUL forced it to dump any octal
      numbers, and if we trash a simple backslash in the StateString, that's OK
      too since it would output a NUL anyway. */
      StateString [0] = 0;
    }
    else if (InputChar > 0)
    {
      /* Append the character to the state string. */

      for (i = 0; i < sizeof (StateString); i++)
      {
        if (StateString [i] == 0)
        {
          StateString [i] = InputChar;
          StateString [i+1] = 0;
          break;
        }
      }
      InputPntr++;
      InputRemaining--;
      LoopsWithoutAnInputAdvance = 0;
    }
  }

  /* Return the amount of data processed back to the user. */

  *SourceLengthPntr = InputPntr - (const unsigned char *) SourceString;
  *DestinationLengthPntr = OutputPntr - (unsigned char *) DestinationString;
  if (StatePntr != NULL)
    strncpy ((char *) StatePntr, (const char *) StateString,
    4); /* Zeroes trailing bytes. */

  return 0;
}



/******************************************************************************
 * Utility function for the ZLib library to allocate memory.  We will just
 * use the kernel malloc, even though it does lock pages in memory etc.
 */

voidpf MyAllocForZLib (voidpf opaque, uInt items, uInt size)
{
  unsigned long Amount;
  void         *Pntr;

  Amount = items * (unsigned long) size;
  Pntr = malloc (Amount);
  if (Pntr == NULL)
  {
    sprintf (((class AFSPApp *) opaque)->m_Recursive.errorMessage,
      "MyAllocForZLib: Failed to allocate %lu bytes (%d items of %d size).\n",
      Amount, (int) items, (int) size);
    return Z_NULL;
  }

  return Pntr;
}



/******************************************************************************
 * Utility function for the ZLib library to deallocate memory.
 */

void MyFreeForZLib (voidpf opaque, voidpf address)
{
  free (address);
}



/******************************************************************************
 * Implementation of the AFSPApp class.  Constructor, destructor and the rest
 * of the member functions in mostly alphabetical order.
 */

AFSPApp::AFSPApp ()
: BApplication (g_AFSPAppSignature),
  m_CancelRequestedByGUI (false),
  m_NowProcessingCurrentTotal (-1),
  m_NowProcessingEstimatedFinalTotal (0),
  m_SettingsHaveChanged (false)
{
  status_t    ErrorCode;
  const void *ResourceData;
  size_t      ResourceSize;
  BResources *ResourcesPntr;

  strncpy (m_NowProcessingPath, "None.", sizeof (m_NowProcessingPath));

  /* Set up the pathname which identifies our settings directory.  Note that
  the actual settings are loaded later on (or set to defaults) by the main()
  function, before this BApplication starts running.  So we don't bother
  initialising the other setting related variables here. */

  ErrorCode =
    find_directory (B_USER_SETTINGS_DIRECTORY, &m_SettingsDirectoryPath);
  if (ErrorCode == B_OK)
    ErrorCode = m_SettingsDirectoryPath.Append (g_SettingsDirectoryName);
  if (ErrorCode != B_OK)
    m_SettingsDirectoryPath.SetTo (".");

  /* Initialise the busy cursor from data in the application's resources. */

  if ((ResourcesPntr = AppResources ()) != NULL && (ResourceData =
  ResourcesPntr->LoadResource ('CURS', "Busy Cursor", &ResourceSize)) != NULL
  && ResourceSize >= 68 /* Size of a raw 2x16x16x8+4 cursor is 68 bytes */)
    g_BusyCursor = new BCursor (ResourceData);
}


AFSPApp::~AFSPApp ()
{
  if (m_SettingsHaveChanged)
    LoadSaveSettings (false /* DoLoad */);
  delete g_BusyCursor;
  g_BusyCursor = NULL;
}


/* Display a box showing information about this program. */

void AFSPApp::AboutRequested ()
{
  BAlert *AboutAlertPntr;

  AboutAlertPntr = new BAlert ("About", g_AboutText, "Done");
  if (AboutAlertPntr != NULL)
  {
    AboutAlertPntr->SetShortcut (0, B_ESCAPE);
    AboutAlertPntr->Go ();
  }
}


/* The user has provided a command line.  This could actually be from a
separate attempt to invoke the program (this application's resource/attributes
have the launch flags set to "single launch", so the shell doesn't start the
program but instead sends the arguments to the already running instance).  In
either case, the command is sent to an intermediary thread where it is
asynchronously converted into a scripting message(s) that are sent back to this
BApplication.  The intermediary is needed since we can't recursively execute
scripting messages while processing a message (this ArgsReceived one). */

void AFSPApp::ArgvReceived (int32 argc, char **argv)
{
  if (g_CommanderLooperPntr != NULL)
    g_CommanderLooperPntr->CommandArguments (argc, argv);
}


/* Set the settings to the defaults.  Needed in case there isn't a settings
file or it is obsolete (doesn't specify all settings). */

void AFSPApp::DefaultSettings ()
{
  status_t ErrorCode;

  m_CompressionMethod = CM_ZLIB;
  m_MakeRAMDisk = true;
  m_MaxRAMDiskSize = 100000000;
  m_Notes = "";
  ErrorCode = find_directory (B_USER_DIRECTORY, &m_PackedFileName);
  if (ErrorCode != B_OK)
  {
    DisplayErrorMessage (
      "DefaultSettings: Unable to find user's home directory", ErrorCode);
    m_PackedFileName.SetTo ("/boot/home");
  }
  m_PackedFileName.Append ("MyDiskImage.pafs");
  m_RAMDiskVirtualMemory = true;
  m_RecreateIndices = true;
  m_TopLevelDirectoryName = "/MyDiskImage";
  m_zlibCompressionLevel = 6;

  m_SettingsHaveChanged = true;
}


status_t AFSPApp::DoCreateFS (char *ErrorMessage)
{
  BFile       AddOnFile;
  const char *AddOnName =
              "/boot/home/config/add-ons/kernel/file_systems/ARFSv0";
  status_t    ErrorCode;
  int         EscapedLength;
  char        EscapedString [B_PATH_NAME_LENGTH * 3];
  BString     Parameters;
  int         SourceLength;
  const char *SourceString;

  /* Check that the file system add-on is there. */

  ErrorCode = AddOnFile.SetTo (AddOnName, B_READ_ONLY);
  if (ErrorCode != B_OK)
  {
    sprintf (ErrorMessage, "AFSPApp::DoCreateFS: Can't find file system "
      "add-on \"%s\", so the subsequent mount operation may fail "
      "(will try it anyway)", AddOnName);
    DisplayErrorMessage (ErrorMessage, ErrorCode, "Early Warning");
  }
  AddOnFile.Unset ();

  /* Make sure the mount directory exists.  create_directory succeeds even if
  the directory exists (though will correctly fail if a file with that name is
  there). */

  ErrorCode = create_directory (m_TopLevelDirectoryName.Path (), 0755);
  if (ErrorCode != B_OK)
  {
    sprintf (ErrorMessage, "AFSPApp::DoCreateFS: Can't create mount point "
      "directory \"%s\"", m_TopLevelDirectoryName.Path ());
    return ErrorCode;
  }

  Parameters = "MountDir=";
  SourceString = m_TopLevelDirectoryName.Path ();
  SourceLength = strlen (SourceString);
  EscapedLength = sizeof (EscapedString);
  AddBackslashEscapeCodes (SourceString, &SourceLength,
    EscapedString, &EscapedLength,
    NULL /* state */, false /* encode 8 bit */);
  EscapedString [EscapedLength] = 0;
  Parameters.Append (EscapedString);

  Parameters << "\tDumpFilePath=";
  SourceString = m_PackedFileName.Path ();
  SourceLength = strlen (SourceString);
  EscapedLength = sizeof (EscapedString);
  AddBackslashEscapeCodes (SourceString, &SourceLength,
    EscapedString, &EscapedLength,
    NULL /* state */, false /* encode 8 bit */);
  EscapedString [EscapedLength] = 0;
  Parameters.Append (EscapedString);

  Parameters << "\tMaxVolumeSize=" << m_MaxRAMDiskSize <<
    "\tDumpCompressionMethod=" <<
    g_CompressionMethodNames [m_CompressionMethod] <<
    "\tDumpZlibCompressionLevel=" << m_zlibCompressionLevel <<
    "\tUseVirtualMemory=" << (int) m_RAMDiskVirtualMemory;

  strncpy (m_NowProcessingPath, Parameters.String (),
    sizeof (m_NowProcessingPath) - 1);
  m_NowProcessingPath [sizeof (m_NowProcessingPath) - 1] = 0;

  ErrorCode = mount ("ARFSv0", m_TopLevelDirectoryName.Path (),
    NULL /* Don't specify a device, needs to be NULL, not an empty string! */,
    0 /* flags: B_MOUNT_READ_ONLY or zero */,
    (void *) Parameters.String (), Parameters.Length () + 1);

  if (ErrorCode != B_OK)
  {
    sprintf (ErrorMessage, "AFSPApp::DoCreateFS: Failed to mount RAM file "
      "system on \"%s\"", m_TopLevelDirectoryName.Path ());
    return ErrorCode;
  }

  return B_OK;
}


status_t AFSPApp::DoPacking (char *ErrorMessage)
{
  status_t   ReturnCode;
  BDirectory TopLevelDirectory;

  /* Can't just memset everything to zero (contains a BString) to make error
  exit cleanup safe... */

  m_Recursive.Clear ();
  m_Recursive.errorMessage = ErrorMessage;

  ReturnCode = TopLevelDirectory.SetTo (m_TopLevelDirectoryName.Path ());
  if (ReturnCode != B_OK)
  {
    sprintf (ErrorMessage, "DoPacking: Unable to open top level "
      "directory \"%s\"", m_TopLevelDirectoryName.Path ());
    goto ErrorExit;
  }

  /* Find the total size to be processed for use in the progress indicator.
  Add up the sizes of files and count non-file things (directories, links, but
  not attributes) as 1 byte of size, just in case it is packing an empty
  directory tree.  This results in a GetStat for every file, so it is a bit
  slow. */

  UpdateNowProcessingPath ("Scanning files and adding up their sizes...");
  m_NowProcessingCurrentTotal = 0;
  m_NowProcessingEstimatedFinalTotal = 0;
  ReturnCode = TopLevelDirectory.GetStat (&m_Recursive.fileStatus);
  if (ReturnCode != B_OK)
  {
    sprintf (ErrorMessage, "DoPacking: Unable to get status of top level "
      "directory \"%s\"", m_TopLevelDirectoryName.Path ());
    goto ErrorExit;
  }
  m_Recursive.currentPathName = m_TopLevelDirectoryName;

  ReturnCode = RecursiveTotalSizeEstimate (TopLevelDirectory);
  if (ReturnCode != B_OK) goto ErrorExit;

  /* Set things up for the actual traversal. */

  m_Recursive.discardedPathPrefixLength =
    strlen (m_TopLevelDirectoryName.Path ()) -
    strlen (m_TopLevelDirectoryName.Leaf ());
  m_Recursive.currentPathName = m_TopLevelDirectoryName;
  m_Recursive.outputLength = 0;
  m_Recursive.outputBuffer = (char *) malloc (INPUT_OUTPUT_BUFFER_SIZE);
  if (m_Recursive.outputBuffer == NULL)
  {
    ReturnCode = ENOMEM;
    sprintf (ErrorMessage, "DoPacking: Unable to allocate %d byte "
      "raw output buffer", INPUT_OUTPUT_BUFFER_SIZE);
    goto ErrorExit;
  }
  m_Recursive.inputLength = 0;
  m_Recursive.inputBuffer = (char *) malloc (INPUT_OUTPUT_BUFFER_SIZE);
  if (m_Recursive.inputBuffer == NULL)
  {
    ReturnCode = ENOMEM;
    sprintf (ErrorMessage, "DoPacking: Unable to allocate %d byte "
      "input buffer", INPUT_OUTPUT_BUFFER_SIZE);
    goto ErrorExit;
  }
  m_Recursive.stream = fopen (m_PackedFileName.Path (), "wb");
  if (m_Recursive.stream == NULL)
  {
    ReturnCode = errno;
    sprintf (ErrorMessage, "DoPacking: Unable to open packed file "
      "\"%s\" for writing", m_PackedFileName.Path ());
    goto ErrorExit;
  }

  /* Write the volume description block.  Done in uncompressed mode so you can
  easily see what's in the file. */

  ReturnCode = DumpRawString ("{ Create Volume\n"
    "  Name=");
  if (ReturnCode != B_OK) goto ErrorExit;

  ReturnCode = DumpEscapedString (m_TopLevelDirectoryName.Leaf (),
    7 /* already output */);
  if (ReturnCode != B_OK) goto ErrorExit;

  sprintf (m_Recursive.inputBuffer, "\n"
    "  MaxVolumeSize=%ld\n"
    "  RoughSizeEstimateNoAttributes=%Ld\n"
    "  Notes=",
    m_MaxRAMDiskSize, m_NowProcessingEstimatedFinalTotal);
  ReturnCode = DumpRawString (m_Recursive.inputBuffer);
  if (ReturnCode != B_OK) goto ErrorExit;

  ReturnCode = DumpEscapedString (m_Notes.String (), 8 /* already output */);
  if (ReturnCode != B_OK) goto ErrorExit;

  ReturnCode = DumpRawString ("\n}\n\n");
  if (ReturnCode != B_OK) goto ErrorExit;

  ReturnCode = DumpSwitchCompressionMode (m_CompressionMethod);
  if (ReturnCode != B_OK) goto ErrorExit;

  ReturnCode = PackIndices (TopLevelDirectory);
  if (ReturnCode != B_OK) goto ErrorExit;

  ReturnCode = TopLevelDirectory.GetStat (&m_Recursive.fileStatus);
  if (ReturnCode != B_OK)
  {
    sprintf (ErrorMessage, "DoPacking: Unable to get status again of top "
      "level directory \"%s\"", m_TopLevelDirectoryName.Path ());
    goto ErrorExit;
  }

  ReturnCode = PackDirectoryRecursively (TopLevelDirectory);
  if (ReturnCode != B_OK) goto ErrorExit;

  /* Force it to flush ZLib compressed data and the lowest level buffers. */

  ReturnCode = DumpSwitchCompressionMode (CM_NONE);
  if (ReturnCode != B_OK) goto ErrorExit;
  ReturnCode = DumpFlushRawData ();
  if (ReturnCode != B_OK) goto ErrorExit;

  /* Successful. */
  ReturnCode = B_OK;
  strcpy (ErrorMessage, "Finished successfully!");

ErrorExit:
  if (m_Recursive.zlibInitialised)
  {
    deflateEnd (&m_Recursive.zlibStream);
    m_Recursive.zlibInitialised = false;
  }
  if (m_Recursive.stream != NULL)
  {
    fclose (m_Recursive.stream);
    m_Recursive.stream = NULL;
  }
  if (m_Recursive.outputBuffer != NULL)
  {
    free (m_Recursive.outputBuffer);
    m_Recursive.outputBuffer = NULL;
  }
  if (m_Recursive.inputBuffer != NULL)
  {
    free (m_Recursive.inputBuffer);
    m_Recursive.inputBuffer = NULL;
  }
  m_Recursive.currentPathName.Unset ();
  UpdateNowProcessingPath (ErrorMessage);
  m_NowProcessingCurrentTotal = -1;
  m_CancelRequestedByGUI = false;
  return ReturnCode;
}


status_t AFSPApp::DoUnpacking (char *ErrorMessage)
{
  BString     CommandBString;
  const char *CommandPntr;
  bool        CreateCommand;
  status_t    ReturnCode;

  /* Can't just memset everything to zero (contains a BString) to make error
  exit cleanup safe... */

  m_Recursive.Clear ();
  m_Recursive.errorMessage = ErrorMessage;
  m_NowProcessingCurrentTotal = 0;
  m_NowProcessingEstimatedFinalTotal = 0;

  /* Make sure the top level directory exists.  Create it if needed. */

  if (lstat (m_TopLevelDirectoryName.Path (), &m_Recursive.fileStatus))
  {
    ReturnCode = errno;
    if (ReturnCode != ENOENT)
    {
      sprintf (ErrorMessage, "DoUnpacking: "
        "Unable to check for existance of top level directory \"%s\"",
        m_TopLevelDirectoryName.Path ());
      goto ErrorExit;
    }
    if (mkdir (m_TopLevelDirectoryName.Path (), 0700))
    {
      ReturnCode = errno;
      sprintf (ErrorMessage, "DoUnpacking: "
        "Unable to make top level directory \"%s\"",
        m_TopLevelDirectoryName.Path ());
      goto ErrorExit;
    }
  }
  else if (!S_ISDIR (m_Recursive.fileStatus.st_mode))
  {
    ReturnCode = ENOTDIR;
    sprintf (ErrorMessage, "DoUnpacking: "
      "Thing named as top level directory \"%s\" is not a directory",
      m_TopLevelDirectoryName.Path ());
    goto ErrorExit;
  }

  /* Allocate the various buffers. */

  m_Recursive.outputBuffer = (char *) malloc (INPUT_OUTPUT_BUFFER_SIZE);
  if (m_Recursive.outputBuffer == NULL)
  {
    ReturnCode = ENOMEM;
    sprintf (ErrorMessage, "DoUnpacking: Unable to allocate %d byte "
      "output buffer", INPUT_OUTPUT_BUFFER_SIZE);
    goto ErrorExit;
  }

  m_Recursive.inputBuffer = (char *) malloc (INPUT_OUTPUT_BUFFER_SIZE);
  if (m_Recursive.inputBuffer == NULL)
  {
    ReturnCode = ENOMEM;
    sprintf (ErrorMessage, "DoUnpacking: Unable to allocate %d byte "
      "input buffer", INPUT_OUTPUT_BUFFER_SIZE);
    goto ErrorExit;
  }

  m_Recursive.rawDataBuffer = (char *) malloc (INPUT_OUTPUT_BUFFER_SIZE);
  if (m_Recursive.rawDataBuffer == NULL)
  {
    ReturnCode = ENOMEM;
    sprintf (ErrorMessage, "DoUnpacking: Unable to allocate %d byte "
      "raw data buffer", INPUT_OUTPUT_BUFFER_SIZE);
    goto ErrorExit;
  }

  m_Recursive.stream = fopen (m_PackedFileName.Path (), "rb");
  if (m_Recursive.stream == NULL)
  {
    ReturnCode = errno;
    sprintf (ErrorMessage, "DoUnpacking: Unable to open packed file "
      "\"%s\" for reading", m_PackedFileName.Path ());
    goto ErrorExit;
  }

  /* First find the volume header.  Need it for the progress bar size estimate,
  and to find the volume name (which gets removed from the paths of its
  contents). */

  UpdateNowProcessingPath ("Looking for volume information...");
  while (true)
  {
    ReturnCode = ReadNextLine (CommandBString);
    if (ReturnCode == READERR_END_OF_FILE)
    {
      sprintf (m_Recursive.errorMessage, "DoUnpacking: "
        "Perhaps this isn't a packed file system archive file?  Didn't find "
        "volume information (a line with \"{ Create Volume\") in file \"%s\"",
        m_PackedFileName.Path ());
      ReturnCode = -1;
      goto ErrorExit;
    }
    else if (ReturnCode != B_OK)
      goto ErrorExit;

    CommandPntr = CommandBString.String();
    if (CommandPntr == NULL || *CommandPntr != '{')
      continue;
    CommandPntr++;
    while (*CommandPntr != 0 && isspace (*CommandPntr))
      CommandPntr++;
    if (strncasecmp (CommandPntr, "Create", 6) != 0)
      continue;
    CommandPntr += 6;
    while (*CommandPntr != 0 && isspace (*CommandPntr))
      CommandPntr++;
    if (strncasecmp (CommandPntr, "Volume", 6) != 0)
      continue;
    ReturnCode = ReadVolumeCommand (false /* OverrideUserSettings */);
    if (ReturnCode != B_OK) goto ErrorExit;
    break;
  }

  /* Now rip through the contents, creating things as we find them. */

  m_Recursive.commandNestingLevel = 0;
  while (true)
  {
    if (m_CancelRequestedByGUI)
    {
      sprintf (m_Recursive.errorMessage, "DoUnpacking: "
        "Cancelled by user while between Things from archive \"%s\"",
        m_PackedFileName.Path ());
      ReturnCode = B_INTERRUPTED;
      goto ErrorExit;
    }

    ReturnCode = ReadNextLine (CommandBString);
    if (ReturnCode == READERR_END_OF_FILE)
      break; /* Got to the end of the file. */
    if (ReturnCode != B_OK)
      goto ErrorExit;

    CommandPntr = CommandBString.String ();
    if (CommandPntr == NULL)
      continue;

    if (*CommandPntr == '{' && m_Recursive.commandNestingLevel == 1)
    {
      CommandPntr++;
      while (*CommandPntr != 0 && isspace (*CommandPntr))
        CommandPntr++;
      if (strncasecmp (CommandPntr, "Create", 6) == 0)
        CreateCommand = true;
      else if (strncasecmp (CommandPntr, "Modify", 6) == 0)
        CreateCommand = false;
      else
        continue;
      CommandPntr += 6;
      while (*CommandPntr != 0 && isspace (*CommandPntr))
        CommandPntr++;
      if (strncasecmp (CommandPntr, "Thing", 5) == 0)
        ReturnCode = ReadThingCommand (CreateCommand);
      else if (CreateCommand && strncasecmp (CommandPntr, "Index", 5) == 0)
        ReturnCode = ReadIndexCommand ();
      else if (strncasecmp (CommandPntr, "Link", 4) == 0)
        ReturnCode = ReadLinkCommand (CreateCommand);
      else if (strncasecmp (CommandPntr, "Volume", 6) == 0)
      {
        sprintf (m_Recursive.errorMessage, "DoUnpacking: "
          "Got an unexpected second Create/Modify Volume command, "
          "in file \"%s\", stopping", m_PackedFileName.Path ());
        ReturnCode = -1;
      }
      else
      {
        CommandBString.Truncate (PATH_MAX);
        sprintf (m_Recursive.errorMessage, "DoUnpacking: "
          "Unrecognized (future?) command of \"%s\"",
          CommandBString.String ());
        ReturnCode = READERR_UNKNOWN_FUTURE_THING;
      }

      if (ReturnCode == READERR_UNKNOWN_FUTURE_THING)
      {
        /* Not really an error, just print to standard output and continue. */
        printf ("%s.\n", m_Recursive.errorMessage);
        continue; /* Skip rest of command until it ends (by nesting level). */
      }
      if (ReturnCode != 0) goto ErrorExit;
    }
  }

  /* Successful. */
  ReturnCode = B_OK;
  strcpy (ErrorMessage, "Finished successfully!");

ErrorExit:
  if (m_Recursive.zlibInitialised)
  {
    inflateEnd (&m_Recursive.zlibStream);
    m_Recursive.zlibInitialised = false;
  }

  if (m_Recursive.stream != NULL)
  {
    fclose (m_Recursive.stream);
    m_Recursive.stream = NULL;
  }

  free (m_Recursive.outputBuffer);
  m_Recursive.outputBuffer = NULL;

  free (m_Recursive.inputBuffer);
  m_Recursive.inputBuffer = NULL;

  free (m_Recursive.rawDataBuffer);
  m_Recursive.rawDataBuffer = NULL;

  free (m_Recursive.oldVolumePrefix);
  m_Recursive.oldVolumePrefix = NULL;

  m_Recursive.currentPathName.Unset ();
  UpdateNowProcessingPath (ErrorMessage);
  m_NowProcessingCurrentTotal = -1;
  m_CancelRequestedByGUI = false;
  return ReturnCode;
}


/* Convert the data into escape codes, so that some codes which would otherwise
confuse the data parsing become escaped text.  Encode8Bit can be specified to
encode pure binary, should be false only for things which can be UTF-8 text.
Will wrap to 78 bytes per line.  Use SavedState (see AddBackslashEscapeCodes)
to treat several calls in a row as if they were for one big buffer or to
initialize the line wrap count, or leave it NULL. */

status_t AFSPApp::DumpEscapedData (
  const char *DataPntr,
  int DataLength,
  bool Encode8Bit,
  int32 *SavedState)
{
  int      DestLength = 0;
  status_t ErrorCode;
  int32    LocalState;
  int      SourceLength;
  char     TempString [4 * 1024];

  if (SavedState == NULL)
  {
    SavedState = &LocalState;
    LocalState = 0;
  }

  while (DataLength > 0 || DestLength != 0)
  {
    /* Convert the input to escaped characters in our temporary buffer.  Fully
    converted if DestLength is zero when no input is provided. */

    SourceLength = DataLength;
    DestLength = sizeof (TempString);
    ErrorCode = AddBackslashEscapeCodes (DataPntr, &SourceLength,
      TempString, &DestLength, SavedState, Encode8Bit);
    if (ErrorCode != B_OK)
    {
      sprintf (m_Recursive.errorMessage,
        "DumpEscapedData: Failed to encode string while working on \"%s\"",
        m_PackedFileName.Path ());
      return ErrorCode;
    }
    DataPntr += SourceLength;
    DataLength -= SourceLength;

    ErrorCode = DumpRawData (TempString, DestLength);
    if (ErrorCode != B_OK) return ErrorCode;
  }
  return B_OK;
}


status_t AFSPApp::DumpEscapedString (
  const char *StringPntr,
  int BytesAlreadyPrinted,
  bool Encode8Bit)
{
  int32 State = BytesAlreadyPrinted;

  return DumpEscapedData (StringPntr, strlen (StringPntr), Encode8Bit, &State);
}


/* Write out whatever is in the raw data buffer. */

status_t AFSPApp::DumpFlushRawData ()
{
  size_t   AmountWritten;
  status_t ErrorCode;

  if (m_Recursive.outputLength <= 0)
    return B_OK;

  AmountWritten = fwrite (m_Recursive.outputBuffer,
    1, m_Recursive.outputLength, m_Recursive.stream);
  if ((int) AmountWritten != m_Recursive.outputLength)
  {
    ErrorCode = errno;
    sprintf (m_Recursive.errorMessage, "DumpFlushRawData: "
      "Tried to write %d bytes to \"%s\", but only wrote %ld",
      m_Recursive.outputLength, m_PackedFileName.Path (), AmountWritten);
    if (ErrorCode >= 0)
      ErrorCode = -1;
    return ErrorCode;
  }
  m_Recursive.outputLength = 0;
  return B_OK;
}


status_t AFSPApp::DumpRawData (const char *DataPntr, int DataLength)
{
  int      AmountCopied;
  status_t ErrorCode;

  while (DataLength > 0)
  {
    if (m_CancelRequestedByGUI)
    {
      sprintf (m_Recursive.errorMessage, "DumpRawData: "
        "Cancelled by user while processing \"%s\"",
        m_Recursive.currentPathName.Path ());
      return B_INTERRUPTED;
    }

    if (m_Recursive.currentCompressionMethod == CM_NONE)
    {
      /* Append the data to the raw output buffer. */

      AmountCopied = INPUT_OUTPUT_BUFFER_SIZE - m_Recursive.outputLength;
      if (AmountCopied > DataLength)
        AmountCopied = DataLength;
      memcpy (m_Recursive.outputBuffer + m_Recursive.outputLength,
        DataPntr, AmountCopied);
      m_Recursive.outputLength += AmountCopied;
      DataLength -= AmountCopied;
      DataPntr += AmountCopied;
    }
    else if (m_Recursive.currentCompressionMethod == CM_ZLIB)
    {
      m_Recursive.zlibStream.next_in = (uint8 *) DataPntr;
      m_Recursive.zlibStream.avail_in = DataLength;
      m_Recursive.zlibStream.next_out = (uint8 *)
        m_Recursive.outputBuffer + m_Recursive.outputLength;
      m_Recursive.zlibStream.avail_out =
        INPUT_OUTPUT_BUFFER_SIZE - m_Recursive.outputLength;

      ErrorCode = deflate (&m_Recursive.zlibStream, Z_NO_FLUSH);
      if (ErrorCode != Z_OK)
      {
        sprintf (m_Recursive.errorMessage, "DumpRawData: "
          "ZLib compression failed, ZLib error code %ld, message \"%s\"",
          ErrorCode, m_Recursive.zlibStream.msg);
        return -1;
      }
      AmountCopied = DataLength - m_Recursive.zlibStream.avail_in;
      DataLength -= AmountCopied;
      DataPntr += AmountCopied;
      AmountCopied = INPUT_OUTPUT_BUFFER_SIZE - m_Recursive.outputLength -
        m_Recursive.zlibStream.avail_out;
      m_Recursive.outputLength += AmountCopied;
    }
    else
    {
      sprintf (m_Recursive.errorMessage, "DumpRawData: Bug!  "
        "Unimplemented compression method #%d selected",
        (int) m_Recursive.currentCompressionMethod);
      return -1;
    }

    /* Flush the raw output buffer if it is full. */

    if (m_Recursive.outputLength >= INPUT_OUTPUT_BUFFER_SIZE)
    {
      ErrorCode = DumpFlushRawData ();
      if (ErrorCode != B_OK)
        return ErrorCode;
    }
  }
  return B_OK;
}


status_t AFSPApp::DumpRawString (const char *StringPntr)
{
  return DumpRawData (StringPntr, strlen (StringPntr));
}


/* Switch compression mode.  Takes care of flushing intermediate buffers, and
also writing the compression change commands.  Presumably done at the start of
a new line (previous byte written was LF), else it won't work. */

status_t AFSPApp::DumpSwitchCompressionMode (CompressionMethod NewMethod)
{
  int      AmountCopied;
  status_t ErrorCode;
  char     TempString [80];

  if (NewMethod == m_Recursive.currentCompressionMethod)
    return B_OK;

  sprintf (TempString, "{ Compression\n  Method=%s\n}\n",
    g_CompressionMethodNames [NewMethod]);
  ErrorCode = DumpRawData (TempString, strlen (TempString));
  if (ErrorCode != B_OK) return ErrorCode;

  if (m_Recursive.currentCompressionMethod == CM_ZLIB &&
  m_Recursive.zlibInitialised)
  {
    /* Flush out any zLib internal buffered data before switching off zLib. */

    while (true)
    {
      if (m_Recursive.outputLength >= INPUT_OUTPUT_BUFFER_SIZE)
      {
        ErrorCode = DumpFlushRawData ();
        if (ErrorCode != B_OK) return ErrorCode;
      }
      m_Recursive.zlibStream.next_in = NULL;
      m_Recursive.zlibStream.avail_in = 0;
      m_Recursive.zlibStream.next_out =
        (uint8 *) m_Recursive.outputBuffer + m_Recursive.outputLength;
      m_Recursive.zlibStream.avail_out =
        INPUT_OUTPUT_BUFFER_SIZE - m_Recursive.outputLength;

      ErrorCode = deflate (&m_Recursive.zlibStream, Z_FINISH);
      AmountCopied = INPUT_OUTPUT_BUFFER_SIZE - m_Recursive.outputLength -
        m_Recursive.zlibStream.avail_out;
      m_Recursive.outputLength += AmountCopied;
      if (ErrorCode == Z_STREAM_END)
        break; /* Successfully finished flushing everything. */
      if (ErrorCode != Z_OK)
      {
        sprintf (m_Recursive.errorMessage, "DumpSwitchCompressionMode: "
          "ZLib compression failed, ZLib error code %ld, message \"%s\"",
          ErrorCode, m_Recursive.zlibStream.msg);
        return -1;
      }
    }
    deflateEnd (&m_Recursive.zlibStream);
    m_Recursive.zlibInitialised = false;
  }

  m_Recursive.currentCompressionMethod = NewMethod;

  if (NewMethod == CM_ZLIB)
  {
    memset (&m_Recursive.zlibStream, 0, sizeof (m_Recursive.zlibStream));
    m_Recursive.zlibStream.zalloc = MyAllocForZLib;
    m_Recursive.zlibStream.zfree = MyFreeForZLib;
    m_Recursive.zlibStream.opaque = this;

    ErrorCode = deflateInit (&m_Recursive.zlibStream, m_zlibCompressionLevel);
    if (ErrorCode != Z_OK)
    {
      sprintf (m_Recursive.errorMessage, "DumpSwitchCompressionMode: "
        "ZLib initialisation failed, ZLib error code %ld, message \"%s\"",
        ErrorCode, m_Recursive.zlibStream.msg);
      return -1;
    }
    m_Recursive.zlibInitialised = true;
  }

  return B_OK;
}


status_t AFSPApp::DumpUnixStandardMetadata (bool DumpSize)
{
  status_t ErrorCode;

  if (DumpSize)
  {
    sprintf (m_Recursive.inputBuffer, "  MetaSize=%Lu\n",
      m_Recursive.fileStatus.st_size);

    ErrorCode = DumpRawString (m_Recursive.inputBuffer);
    if (ErrorCode != B_OK) return ErrorCode;
  }

  sprintf (m_Recursive.inputBuffer,
    "  MetaOwnerID=%d\n"
    "  MetaGroupID=%d\n"
    "  MetaModeBits=%05o\n"
    "  MetaDateCreatedMicrosecondsSince1970=%Lu\n"
    "  MetaDateAccessedMicrosecondsSince1970=%Lu\n"
    "  MetaDateModifiedMicrosecondsSince1970=%Lu\n"
    "  MetaDateMetaChangedMicrosecondsSince1970=%Lu\n",
    m_Recursive.fileStatus.st_uid,
    m_Recursive.fileStatus.st_gid,
    m_Recursive.fileStatus.st_mode & S_IUMSK,
    m_Recursive.fileStatus.st_crtime * 1000000ULL,
    m_Recursive.fileStatus.st_atime * 1000000ULL,
    m_Recursive.fileStatus.st_mtime * 1000000ULL,
    m_Recursive.fileStatus.st_ctime * 1000000ULL);

  return DumpRawString (m_Recursive.inputBuffer);
}


/* Tell other programs about the scripting commands we support.  Try this
command: "hey application/x-vnd.agmsmith.AGMSFileSystemPacker getsuites" to see
it in action (this program has to be already running for it to work). */

status_t AFSPApp::GetSupportedSuites (BMessage *MessagePntr)
{
  BPropertyInfo TempPropInfo (g_ScriptingPropertyList);

  MessagePntr->AddString ("suites",
    "suite/x-vnd.agmsmith.AGMSFileSystemPacker");
  MessagePntr->AddFlat ("messages", &TempPropInfo);
  return BApplication::GetSupportedSuites (MessagePntr);
}


/* Set up the pack file MIME type (provide a description and associate it with
this program so that it picks up the right icon).  Replaces the previous MIME
type if it was there before. */

status_t AFSPApp::InstallThings (char *ErrorMessage)
{
  status_t    ErrorCode = B_OK;
  BMimeType   MimeType;

  /* Set up the MIME type for the text version of the packed file. */

  sprintf (ErrorMessage, "Problems with setting up MIME type (%s) for "
    "the pack file", g_AFSPTextFileMIMEType); /* A generic message. */

  ErrorCode = MimeType.SetTo (g_AFSPTextFileMIMEType);
  if (ErrorCode != B_OK)
    return ErrorCode;

  if (MimeType.IsInstalled ())
    MimeType.Delete ();
  ErrorCode = MimeType.Install ();
  if (ErrorCode != B_OK)
    return ErrorCode;

  MimeType.SetShortDescription ("Packed File System - Text");
  MimeType.SetLongDescription ("AGMSFileSystemPacker "
    "packed file system file - contains everything needed to recreate a "
    "modern file system or directory tree, editable text variety.");
  sprintf (ErrorMessage, "1.0 ('%s')", g_AFSPTextFileRecognitionString);
  MimeType.SetSnifferRule (ErrorMessage);
  MimeType.SetPreferredApp (g_AFSPAppSignature);

  /* Set up the MIME type for the binary version of the packed file. */

  sprintf (ErrorMessage, "Problems with setting up MIME type (%s) for "
    "the pack file", g_AFSPBinaryFileMIMEType); /* A generic message. */

  ErrorCode = MimeType.SetTo (g_AFSPBinaryFileMIMEType);
  if (ErrorCode != B_OK)
    return ErrorCode;

  if (MimeType.IsInstalled ())
    MimeType.Delete ();
  ErrorCode = MimeType.Install ();
  if (ErrorCode != B_OK)
    return ErrorCode;

  MimeType.SetShortDescription ("Packed File System - Binary");
  MimeType.SetLongDescription ("AGMSFileSystemPacker "
    "packed file system file - contains everything needed to recreate a "
    "modern file system or directory tree, compressed binary variety.");
  sprintf (ErrorMessage, "1.0 ('%s')", g_AFSPBinaryFileRecognitionString);
  MimeType.SetSnifferRule (ErrorMessage);
  MimeType.SetPreferredApp (g_AFSPAppSignature);

  /* Register our beep sound effect. */

  add_system_beep_event (g_BeepPackDone);

  return B_OK;
}


/* Either load the settings (DoLoad is TRUE) from the configuration file or
write them (DoLoad is FALSE) to it.  The configuration file is a flattened
BMessage containing the various program settings.  If it doesn't exist (and its
parent directories don't exist) then it will be created when saving.  If it
doesn't exist when loading, the settings will be set to default values. */

status_t AFSPApp::LoadSaveSettings (bool DoLoad)
{
  status_t    ErrorCode;
  const char *NamePntr;
  BMessage    Settings;
  BDirectory  SettingsDirectory;
  BFile       SettingsFile;
  const char *StringPntr;
  bool        TempBool;
  int32       TempInt32;
  char        TempString [ERROR_MESSAGE_SIZE];

  /* Preset things to default values if loading, in case of an error or it's an
  older version of the settings file which doesn't have every field defined. */

  if (DoLoad)
    DefaultSettings ();

  /* Look for our settings directory.  When saving we can try to create it. */

  ErrorCode = SettingsDirectory.SetTo (m_SettingsDirectoryPath.Path ());
  if (ErrorCode != B_OK)
  {
    if (DoLoad || ErrorCode != B_ENTRY_NOT_FOUND)
    {
      sprintf (TempString, "Can't find settings directory \"%s\"",
        m_SettingsDirectoryPath.Path ());
      goto ErrorExit;
    }
    ErrorCode = create_directory (m_SettingsDirectoryPath.Path (), 0755);
    if (ErrorCode == B_OK)
      ErrorCode = SettingsDirectory.SetTo (m_SettingsDirectoryPath.Path ());
    if (ErrorCode != B_OK)
    {
      sprintf (TempString, "Can't create settings directory \"%s\"",
        m_SettingsDirectoryPath.Path ());
      goto ErrorExit;
    }
  }

  ErrorCode = SettingsFile.SetTo (&SettingsDirectory, g_SettingsFileName,
    DoLoad ? B_READ_ONLY : B_READ_WRITE | B_CREATE_FILE | B_ERASE_FILE);
  if (ErrorCode != B_OK)
  {
    sprintf (TempString, "Can't open settings file \"%s\" in directory \"%s\" "
      "for %s", g_SettingsFileName, m_SettingsDirectoryPath.Path(),
      DoLoad ? "reading" : "writing");
    goto ErrorExit;
  }

  if (DoLoad)
  {
    ErrorCode = Settings.Unflatten (&SettingsFile);
    if (ErrorCode != 0 || Settings.what != g_SettingsWhatCode)
    {
      sprintf (TempString, "Corrupt data detected while reading settings "
        "file \"%s\" in directory \"%s\", will revert to defaults",
        g_SettingsFileName, m_SettingsDirectoryPath.Path());
      goto ErrorExit;
    }
  }

  /* Transfer the settings between the BMessage and our various global
  variables.  For loading, if the setting isn't present, leave it at the
  default value.  Note that loading and saving are intermingled here to make
  code maintenance easier (less chance of forgetting to update it if load and
  save were separate functions). */

  ErrorCode = B_OK; /* So that saving settings can record an error. */

  NamePntr = "CompressionMethod";
  if (DoLoad)
  {
    if (Settings.FindString (NamePntr, &StringPntr) == B_OK)
      for (TempInt32 = 0; TempInt32 < CM_MAX; TempInt32++)
        if (strcasecmp (g_CompressionMethodNames [TempInt32], StringPntr) == 0)
        {
          m_CompressionMethod = (CompressionMethod) TempInt32;
          break;
        }
  }
  else if (ErrorCode == B_OK)
    ErrorCode = Settings.AddString (NamePntr,
      g_CompressionMethodNames [m_CompressionMethod]);

  NamePntr = "MakeRAMDisk";
  if (DoLoad)
  {
    if (Settings.FindBool (NamePntr, &TempBool) == B_OK)
      m_MakeRAMDisk = TempBool;
  }
  else if (ErrorCode == B_OK)
    ErrorCode = Settings.AddBool (NamePntr, m_MakeRAMDisk);

  NamePntr = "MaxRAMDiskSize";
  if (DoLoad)
  {
    if (Settings.FindInt32 (NamePntr, &TempInt32) == B_OK)
      m_MaxRAMDiskSize = TempInt32;
    if (m_MaxRAMDiskSize < 0)
      m_MaxRAMDiskSize = 0;
  }
  else if (ErrorCode == B_OK)
    ErrorCode = Settings.AddInt32 (NamePntr, m_MaxRAMDiskSize);

  NamePntr = "Notes";
  if (DoLoad)
  {
    if (Settings.FindString (NamePntr, &StringPntr) == B_OK)
      m_Notes = StringPntr;
  }
  else if (ErrorCode == B_OK)
    ErrorCode = Settings.AddString (NamePntr, m_Notes.String ());

  NamePntr = "PackedFileName";
  if (DoLoad)
  {
    if (Settings.FindString (NamePntr, &StringPntr) == B_OK)
      m_PackedFileName.SetTo (StringPntr);
  }
  else if (ErrorCode == B_OK && m_PackedFileName.InitCheck () == B_OK)
    ErrorCode = Settings.AddString (NamePntr, m_PackedFileName.Path ());

  NamePntr = "RAMDiskVirtualMemory";
  if (DoLoad)
  {
    if (Settings.FindBool (NamePntr, &TempBool) == B_OK)
      m_RAMDiskVirtualMemory = TempBool;
  }
  else if (ErrorCode == B_OK)
    ErrorCode = Settings.AddBool (NamePntr, m_RAMDiskVirtualMemory);

  NamePntr = "RecreateIndices";
  if (DoLoad)
  {
    if (Settings.FindBool (NamePntr, &TempBool) == B_OK)
      m_RecreateIndices = TempBool;
  }
  else if (ErrorCode == B_OK)
    ErrorCode = Settings.AddBool (NamePntr, m_RecreateIndices);

  NamePntr = "TopLevelDirectoryName";
  if (DoLoad)
  {
    if (Settings.FindString (NamePntr, &StringPntr) == B_OK)
      m_TopLevelDirectoryName.SetTo (StringPntr, NULL /* leaf */,
      true /* normalize */);
  }
  else if (ErrorCode == B_OK && m_TopLevelDirectoryName.InitCheck () == B_OK)
    ErrorCode = Settings.AddString (NamePntr, m_TopLevelDirectoryName.Path ());

  NamePntr = "ZLibCompressionLevel";
  if (DoLoad)
  {
    if (Settings.FindInt32 (NamePntr, &TempInt32) == B_OK)
      m_zlibCompressionLevel = TempInt32;
    if (m_zlibCompressionLevel < 0)
      m_zlibCompressionLevel = 0;
    else if (m_zlibCompressionLevel > 9)
      m_zlibCompressionLevel = 9;
  }
  else if (ErrorCode == B_OK)
    ErrorCode = Settings.AddInt32 (NamePntr, m_zlibCompressionLevel);

  if (ErrorCode != B_OK)
  {
    strcpy (TempString, "Unable to stuff the program settings into a "
      "temporary BMessage, settings not saved");
    goto ErrorExit;
  }

  /* Save the settings BMessage to the settings file. */

  if (!DoLoad)
  {
    Settings.what = g_SettingsWhatCode;
    ErrorCode = Settings.Flatten (&SettingsFile);
    if (ErrorCode != 0)
    {
      sprintf (TempString, "Problems while writing settings file \"%s\" in "
        "directory \"%s\"", g_SettingsFileName,
        m_SettingsDirectoryPath.Path ());
      goto ErrorExit;
    }
  }

  m_SettingsHaveChanged = false;
  return B_OK;

ErrorExit: /* Error message in TempString, code in ErrorCode. */
  DisplayErrorMessage (TempString, ErrorCode, DoLoad ?
    "Loading Settings Error" : "Saving Settings Error");
  return ErrorCode;
}


void AFSPApp::MessageReceived (BMessage *MessagePntr)
{
  const char           *PropertyName;
  struct property_info *PropInfoPntr;
  int32                 SpecifierIndex;
  int32                 SpecifierKind;
  BMessage              SpecifierMessage;

  /* See if it is a scripting message that applies to one of the operations
  this program supports.  Pass on other scripting messages to the inherited
  parent MessageReceived function (they're usually scripting messages for the
  BApplication). */

  switch (MessagePntr->what)
  {
    case B_GET_PROPERTY:
    case B_SET_PROPERTY:
    case B_COUNT_PROPERTIES:
    case B_CREATE_PROPERTY:
    case B_DELETE_PROPERTY:
    case B_EXECUTE_PROPERTY:
      if (MessagePntr->GetCurrentSpecifier (&SpecifierIndex, &SpecifierMessage,
      &SpecifierKind, &PropertyName) == B_OK &&
      SpecifierKind == B_DIRECT_SPECIFIER)
      {
        for (PropInfoPntr = g_ScriptingPropertyList + 0; true; PropInfoPntr++)
        {
          if (PropInfoPntr->name == 0)
            break; /* Ran out of commands. */

          if (PropInfoPntr->commands[0] == MessagePntr->what &&
          strcasecmp (PropInfoPntr->name, PropertyName) == 0)
          {
            ProcessScriptingMessage (MessagePntr, PropInfoPntr);
            return;
          }
        }
      }
      break;
  }

  /* Pass the unprocessed message to the inherited function, maybe it knows
  what to do.  This includes replies to messages we sent ourselves. */

  BApplication::MessageReceived (MessagePntr);
}


status_t AFSPApp::PackAttributes (BNode &CurrentNode)
{
  ssize_t   AmountRead;
  attr_info AttributeInfo;
  char      AttributeName [B_ATTR_NAME_LENGTH+1];
  BString   AttributePath;
  int       DestLength;
  bool      DoubleEscaped;
  status_t  ErrorCode;
  int       SourceLength;
  int32     State;
  off_t     TotalWritten;

  ErrorCode = CurrentNode.RewindAttrs ();
  if (ErrorCode == B_BAD_VALUE)
    return B_OK; /* Attributes not supported. */
  if (ErrorCode != B_OK)
  {
    sprintf (m_Recursive.errorMessage, "PackAttributes: "
      "Unable to rewind the attributes list for \"%s\"",
      m_Recursive.currentPathName.Path ());
    return ErrorCode;
  }

  while (true)
  {
    ErrorCode = CurrentNode.GetNextAttrName (AttributeName);
    if (ErrorCode == B_ENTRY_NOT_FOUND)
      break; /* Got to the end of the list of attributes. */
    if (ErrorCode != B_OK)
    {
      sprintf (m_Recursive.errorMessage, "PackAttributes: "
        "Unable to get next attribute name for \"%s\"",
        m_Recursive.currentPathName.Path ());
      return ErrorCode;
    }

    /* Some attribute names have slashes in them.  Double escape the attribute
    name as a hack to allow such names.  Real fildirute systems won't have them
    since they can have sub-attributes. */

    DoubleEscaped = (strchr (AttributeName, '/') != NULL);
    if (DoubleEscaped)
    {
      SourceLength = strlen (AttributeName);
      DestLength = INPUT_OUTPUT_BUFFER_SIZE;
      AddBackslashEscapeCodes (AttributeName, &SourceLength,
        m_Recursive.inputBuffer, &DestLength,
        NULL /* State */, false /* Encode8BitCharacters */);
      m_Recursive.inputBuffer[DestLength] = 0;
      AttributePath = m_Recursive.inputBuffer;
      AttributePath.ReplaceAll ("/", "\\057");
    }
    else /* Don't need to double escape, so don't. */
      AttributePath = AttributeName;
    AttributePath.Prepend ("/");
    AttributePath.Prepend (m_Recursive.currentPathName.Path ());

    ErrorCode = CurrentNode.GetAttrInfo (AttributeName, &AttributeInfo);
    if (ErrorCode != B_OK)
    {
      sprintf (m_Recursive.errorMessage, "PackAttributes: "
        "Unable to get info on the attribute \"%s\"",
        AttributePath.String ());
      return ErrorCode;
    }

    UpdateNowProcessingPath (AttributePath.String ());

    ErrorCode = DumpRawString ("{ Create Thing\n"
      "  EquivalentOldThing=Attribute\n");
    if (ErrorCode != B_OK) return ErrorCode;
    if (DoubleEscaped)
    {
      ErrorCode = DumpRawString ("  PathDoubleEscapedLeaf=true\n");
      if (ErrorCode != B_OK) return ErrorCode;
    }
    ErrorCode = DumpRawString ("  Path=");
    if (ErrorCode != B_OK) return ErrorCode;

    ErrorCode = DumpEscapedString (AttributePath.String () +
      m_Recursive.discardedPathPrefixLength, 7 /* already output */);
    if (ErrorCode != B_OK) return ErrorCode;

    ErrorCode = DumpRawString ("\n  PrimitiveType=");
    if (ErrorCode != B_OK) return ErrorCode;

    m_Recursive.inputBuffer[0] = (char) (AttributeInfo.type >> 24);
    m_Recursive.inputBuffer[1] = (char) (AttributeInfo.type >> 16);
    m_Recursive.inputBuffer[2] = (char) (AttributeInfo.type >> 8);
    m_Recursive.inputBuffer[3] = (char) (AttributeInfo.type);
    ErrorCode = DumpEscapedData (m_Recursive.inputBuffer, 4,
      true /* Encode8Bit */, NULL /* SavedState */);
    if (ErrorCode != B_OK) return ErrorCode;

    ErrorCode = DumpRawString ("\n  Data=");
    if (ErrorCode != B_OK) return ErrorCode;

    /* Write out the contents of the attribute as escaped data. */

    State = 7; /* Number of bytes already in the current output text line. */
    TotalWritten = 0;
    while (TotalWritten < AttributeInfo.size)
    {
      AmountRead = CurrentNode.ReadAttr (AttributeName, AttributeInfo.type,
        TotalWritten /* read offset */, m_Recursive.inputBuffer,
        INPUT_OUTPUT_BUFFER_SIZE);
      if (AmountRead == 0)
      {
        sprintf (m_Recursive.errorMessage, "PackAttributes: "
          "Bug!  Early end of data while reading attribute \"%s\"",
          AttributePath.String ());
        return -1;
      }
      if (AmountRead < 0)
      {
        sprintf (m_Recursive.errorMessage, "PackAttributes: "
          "Error while reading attribute \"%s\"",
          AttributePath.String ());
        return AmountRead;
      }

      ErrorCode = DumpEscapedData (m_Recursive.inputBuffer,
        AmountRead, true /* Encode8Bit */, &State);
      if (ErrorCode != B_OK) return ErrorCode;

      TotalWritten += AmountRead;
    }

    sprintf (m_Recursive.inputBuffer,
      "\n  MetaSize=%Lu\n}\n\n", AttributeInfo.size);
    ErrorCode = DumpRawString (m_Recursive.inputBuffer);
    if (ErrorCode != B_OK) return ErrorCode;
  }

  return B_OK;
}


/* Dump out the packed data for the given directory and all child objects.
Returns B_OK if successful.  Or an error if it fails or was interrupted by the
user, with a written description of the problem in m_Recursive.errorMessage.
Uses the output stream and other settings in the application instance's
m_Recursive variable to save on recursive stack space usage. */

status_t AFSPApp::PackDirectoryRecursively (BDirectory &CurrentNode)
{
  BEntry      ChildEntry;
  struct stat DirectoryStat;
  status_t    ErrorCode;

  /* Dump out the information needed to create the basic object first, so that
  attributes and other things can be added soon after when the file is
  unpacked. */

  UpdateNowProcessingPath (m_Recursive.currentPathName.Path ());
  m_NowProcessingCurrentTotal++; /* Directories count as one. */
  memcpy (&DirectoryStat, &m_Recursive.fileStatus, sizeof (DirectoryStat));

  ErrorCode = DumpRawString ("{ Create Thing\n"
    "  EquivalentOldThing=Directory\n"
    "  Path=");
  if (ErrorCode != B_OK) return ErrorCode;

  ErrorCode = DumpEscapedString (m_Recursive.currentPathName.Path () +
    m_Recursive.discardedPathPrefixLength, 7 /* already output */);
  if (ErrorCode != B_OK) return ErrorCode;

  ErrorCode = DumpRawString ("\n"
    "  Data=\n" /* Need a Data= line even for directories. */
    "}\n\n");
  if (ErrorCode != B_OK) return ErrorCode;

  ErrorCode = PackAttributes (CurrentNode);
  if (ErrorCode != B_OK) return ErrorCode;

  /* Write out the things in the directory. */

  CurrentNode.Rewind ();
  while ((ErrorCode = CurrentNode.GetNextEntry (&ChildEntry,
  false /* traverse */)) == B_OK)
  {
    ChildEntry.GetName (m_Recursive.errorMessage);
    m_Recursive.currentPathName.Append (m_Recursive.errorMessage);

    ErrorCode = ChildEntry.GetStat (&m_Recursive.fileStatus);
    if (ErrorCode != B_OK)
    {
      sprintf (m_Recursive.errorMessage, "PackDirectoryRecursively: "
        "Unable to get status of entry \"%s\"",
        m_Recursive.currentPathName.Path ());
      return ErrorCode;
    }

    if (S_ISDIR (m_Recursive.fileStatus.st_mode))
    {
      /* Dump a child subdirectory. */

      BDirectory ChildDir;
      ErrorCode = ChildDir.SetTo (&ChildEntry);
      if (ErrorCode != B_OK)
      {
        sprintf (m_Recursive.errorMessage, "PackDirectoryRecursively: "
          "Unable to open directory \"%s\"",
          m_Recursive.currentPathName.Path ());
        return ErrorCode;
      }
      ErrorCode = PackDirectoryRecursively (ChildDir);
      if (ErrorCode != B_OK) return ErrorCode;
    }
    else if (S_ISREG (m_Recursive.fileStatus.st_mode))
    {
      /* Dump a regular data file. */

      BFile ChildFile;
      ErrorCode = ChildFile.SetTo (&ChildEntry, B_READ_ONLY);
      if (ErrorCode != B_OK)
      {
        sprintf (m_Recursive.errorMessage, "PackDirectoryRecursively: "
          "Unable to open file \"%s\"",
          m_Recursive.currentPathName.Path ());
        return ErrorCode;
      }
      ErrorCode = PackFile (ChildFile);
      if (ErrorCode != B_OK) return ErrorCode;
    }
    else if (S_ISLNK (m_Recursive.fileStatus.st_mode))
    {
      /* Dump a symbolic or AGMS link. */

      BSymLink ChildLink;
      ErrorCode = ChildLink.SetTo (&ChildEntry);
      if (ErrorCode != B_OK)
      {
        sprintf (m_Recursive.errorMessage, "PackDirectoryRecursively: "
          "Unable to open symbolic (or other kind of) link \"%s\"",
          m_Recursive.currentPathName.Path ());
        return ErrorCode;
      }
      ErrorCode = PackSymbolicLink (ChildLink, CurrentNode);
      if (ErrorCode != B_OK) return ErrorCode;
    }
    else /* A nonstandard thing - block file, disk device etc. */
    {
      if (!m_Recursive.blockDeviceWarningAlreadyIssued)
      {
        m_Recursive.blockDeviceWarningAlreadyIssued = true;
        sprintf (m_Recursive.errorMessage, "PackDirectoryRecursively: "
          "Thing \"%s\" isn't a normal thing, maybe it's a block device "
          "or FIFO (mode flags $%X).  Ignoring it, and all further "
          "non-files.",
          m_Recursive.currentPathName.Path (),
          m_Recursive.fileStatus.st_mode & S_IFMT);
        DisplayErrorMessage (m_Recursive.errorMessage);
      }
    }

    m_Recursive.currentPathName.GetParent (&m_Recursive.currentPathName);
  }

  if (ErrorCode != B_ENTRY_NOT_FOUND)
  {
    sprintf (m_Recursive.errorMessage, "PackDirectoryRecursively: "
      "Had problems reading directory \"%s\" contents",
      m_Recursive.currentPathName.Path ());
    return ErrorCode;
  }

  /* Finally, dump out the directory's time stamps and other metadata.  Done
  after the directory contents so that restoration side effects (date last
  modified set to current time) get wiped out.  Also makes it easier to add
  things to a directory if the permissions allow it. */

  memcpy (&m_Recursive.fileStatus, &DirectoryStat, sizeof (DirectoryStat));

  ErrorCode = DumpRawString ("{ Modify Thing\n"
    "  EquivalentOldThing=Directory\n"
    "  Path=");
  if (ErrorCode != B_OK) return ErrorCode;

  ErrorCode = DumpEscapedString (m_Recursive.currentPathName.Path () +
    m_Recursive.discardedPathPrefixLength, 7 /* already output */);
  if (ErrorCode != B_OK) return ErrorCode;

  ErrorCode = DumpRawString ("\n");
  if (ErrorCode != B_OK) return ErrorCode;

  ErrorCode = DumpUnixStandardMetadata (false /* DumpSize */);
  if (ErrorCode != B_OK) return ErrorCode;

  ErrorCode = DumpRawString ("}\n\n");
  if (ErrorCode != B_OK) return ErrorCode;

  return B_OK;
}


status_t AFSPApp::PackFile (BFile &CurrentNode)
{
  ssize_t   AmountRead;
  status_t  ErrorCode;
  int32     State;

  UpdateNowProcessingPath (m_Recursive.currentPathName.Path ());

  ErrorCode = DumpRawString ("{ Create Thing\n"
    "  EquivalentOldThing=File\n"
    "  Path=");
  if (ErrorCode != B_OK) return ErrorCode;

  ErrorCode = DumpEscapedString (m_Recursive.currentPathName.Path () +
    m_Recursive.discardedPathPrefixLength, 7 /* already output */);
  if (ErrorCode != B_OK) return ErrorCode;

  ErrorCode = DumpRawString ("\n  Data=");
  if (ErrorCode != B_OK) return ErrorCode;

  ErrorCode = CurrentNode.Seek (0, SEEK_SET);
  if (ErrorCode < 0)
  {
    sprintf (m_Recursive.errorMessage, "PackFile: "
      "Seek to start of file \"%s\" failed",
      m_Recursive.currentPathName.Path ());
    return ErrorCode;
  }

  /* Write out the contents of the file as escaped data. */

  State = 7; /* Number of bytes already in the current output text line. */
  while (true)
  {
    AmountRead = CurrentNode.Read (m_Recursive.inputBuffer,
      INPUT_OUTPUT_BUFFER_SIZE);
    if (AmountRead == 0)
      break; /* Reached end of file. */
    if (AmountRead < 0)
    {
      sprintf (m_Recursive.errorMessage, "PackFile: "
        "Error while reading file \"%s\"",
        m_Recursive.currentPathName.Path ());
      return AmountRead;
    }

    ErrorCode = DumpEscapedData (m_Recursive.inputBuffer,
      AmountRead, true /* Encode8Bit */, &State);
    if (ErrorCode != B_OK) return ErrorCode;

    m_NowProcessingCurrentTotal += AmountRead;
  }

  ErrorCode = DumpRawString ("\n");
  if (ErrorCode != B_OK) return ErrorCode;

  ErrorCode = DumpUnixStandardMetadata (true /* DumpSize */);
  if (ErrorCode != B_OK) return ErrorCode;

  ErrorCode = DumpRawString ("}\n\n");
  if (ErrorCode != B_OK) return ErrorCode;

  ErrorCode = PackAttributes (CurrentNode);
  if (ErrorCode != B_OK) return ErrorCode;

  return B_OK;
}


status_t AFSPApp::PackIndices (BDirectory &TopLevelDirectory)
{
  dirent_t   *DirEntPntr;
  DIR        *IndexDirHandle = NULL;
  index_info  IndexInfo;
  const char *IndexName;
  node_ref    NodeRef;
  status_t    ReturnCode;

  UpdateNowProcessingPath ("Packing indices...");

  ReturnCode = TopLevelDirectory.GetNodeRef (&NodeRef);
  if (ReturnCode != B_OK)
  {
    sprintf (m_Recursive.errorMessage, "PackIndices: "
      "Unable to get node_ref of \"%s\"",
      m_Recursive.currentPathName.Path ());
    goto ErrorExit;
  }

  IndexDirHandle = fs_open_index_dir (NodeRef.device);
  if (IndexDirHandle == NULL)
  {
    ReturnCode = errno;
    sprintf (m_Recursive.errorMessage, "PackIndices: "
      "Unable to open indices directory for volume containing \"%s\", "
      "will continue without indices",
      m_Recursive.currentPathName.Path ());
    DisplayErrorMessage (m_Recursive.errorMessage, ReturnCode);
    ReturnCode = B_OK;
    goto ErrorExit;
  }

  while (true)
  {
    DirEntPntr = fs_read_index_dir (IndexDirHandle);
    if (DirEntPntr == NULL)
    {
      ReturnCode = errno;
      if (ReturnCode == B_ENTRY_NOT_FOUND)
        break; /* End of list of indices. */
      sprintf (m_Recursive.errorMessage, "PackIndices: "
        "Unable to read indices directory for volume containing \"%s\"",
        m_Recursive.currentPathName.Path ());
      goto ErrorExit;
    }
    IndexName = DirEntPntr->d_name;

    if (fs_stat_index (NodeRef.device, IndexName, &IndexInfo))
    {
      ReturnCode = errno;
      sprintf (m_Recursive.errorMessage, "PackIndices: "
        "Unable to read index \"%s\" info on volume containing \"%s\"",
        IndexName, m_Recursive.currentPathName.Path ());
      goto ErrorExit;
    }

    ReturnCode = DumpRawString ("{ Create Index\n"
      "  Name=");
    if (ReturnCode != B_OK) goto ErrorExit;

    ReturnCode = DumpEscapedString (IndexName, 7 /* already output */);
    if (ReturnCode != B_OK) goto ErrorExit;

    ReturnCode = DumpRawString ("\n  PrimitiveType=");
    if (ReturnCode != B_OK) goto ErrorExit;

    m_Recursive.inputBuffer[0] = (char) (IndexInfo.type >> 24);
    m_Recursive.inputBuffer[1] = (char) (IndexInfo.type >> 16);
    m_Recursive.inputBuffer[2] = (char) (IndexInfo.type >> 8);
    m_Recursive.inputBuffer[3] = (char) (IndexInfo.type);
    ReturnCode = DumpEscapedData (m_Recursive.inputBuffer, 4,
      true /* Encode8Bit */, NULL /* SavedState */);
    if (ReturnCode != B_OK) goto ErrorExit;

    sprintf (m_Recursive.inputBuffer, "\n"
      "  MetaSize=%Lu\n"
      "  MetaOwnerID=%d\n"
      "  MetaGroupID=%d\n"
      "  MetaDateCreatedMicrosecondsSince1970=%Lu\n"
      "  MetaDateModifiedMicrosecondsSince1970=%Lu\n"
      "}\n\n",
      IndexInfo.size,
      IndexInfo.uid,
      IndexInfo.gid,
      IndexInfo.creation_time * 1000000ULL,
      IndexInfo.modification_time * 1000000ULL);
    ReturnCode = DumpRawString (m_Recursive.inputBuffer);
    if (ReturnCode != B_OK) goto ErrorExit;
  }

  ReturnCode = B_OK;

ErrorExit:
  if (IndexDirHandle != NULL)
    fs_close_index_dir (IndexDirHandle);
  return ReturnCode;
}


status_t AFSPApp::PackSymbolicLink (
  BSymLink &CurrentNode,
 BDirectory &ParentNode)
{
  status_t ErrorCode;
  char    *StringPntr;
  BString  TargetText;

  UpdateNowProcessingPath (m_Recursive.currentPathName.Path ());

  /* AGMS links and parent directory references show up as fake symbolic
  links.  You can tell them apart from the real symbolic links by the size
  of the link. */

  if (m_Recursive.fileStatus.st_size == 2)
    return B_OK; /* An automatically generated auxiliary parent directory. */

  if (m_Recursive.fileStatus.st_size != 0 &&
  m_Recursive.fileStatus.st_size != 1)
  {
    sprintf (m_Recursive.errorMessage, "PackSymbolicLink: "
      "Unknown kind of symbolic (or other kind of) link \"%s\", "
      "has a weird size of %d", m_Recursive.currentPathName.Path (),
      (int) m_Recursive.fileStatus.st_size);
    return B_MISMATCHED_VALUES;
  }

  m_NowProcessingCurrentTotal++; /* Links count as one byte in progress bar. */

  ErrorCode = DumpRawString ("{ Create Link\n  LinkType=");
  if (ErrorCode != B_OK) return ErrorCode;

  if (m_Recursive.fileStatus.st_size == 0)
    ErrorCode = DumpRawString ("Symbolic");
  else
    ErrorCode = DumpRawString ("AliasGeneratedMorphingSymbolic");
  if (ErrorCode != B_OK) return ErrorCode;

  ErrorCode = DumpRawString ("\n  Path=");
  if (ErrorCode != B_OK) return ErrorCode;

  ErrorCode = DumpEscapedString (m_Recursive.currentPathName.Path () +
    m_Recursive.discardedPathPrefixLength, 7 /* already output */);
  if (ErrorCode != B_OK) return ErrorCode;

  StringPntr = TargetText.LockBuffer (PATH_MAX * 2);
  ErrorCode = CurrentNode.ReadLink (StringPntr, PATH_MAX * 2);
  TargetText.UnlockBuffer (-1);
  if (ErrorCode < 0)
  {
    sprintf (m_Recursive.errorMessage, "PackSymbolicLink: "
      "Unable to get link's target string for link \"%s\"",
      m_Recursive.currentPathName.Path ());
    return ErrorCode;
  }

  /* Write out the target path, relative to the top level directory if it is an
  absolute link pointing to something inside the top level directory, or as-is
  if it points to something outside the root directory. */

  StringPntr = (char *) TargetText.String ();
  if (strncmp (StringPntr, m_TopLevelDirectoryName.Path (),
  m_Recursive.discardedPathPrefixLength) == 0)
  {
    ErrorCode = DumpRawString ("\n  RootRelativeTarget=");
    if (ErrorCode != B_OK) return ErrorCode;
    StringPntr += m_Recursive.discardedPathPrefixLength;
    ErrorCode = DumpEscapedString (StringPntr, 21 /* already output */);
    if (ErrorCode != B_OK) return ErrorCode;
  }
  else
  {
    ErrorCode = DumpRawString ("\n  ExplicitTarget=");
    if (ErrorCode != B_OK) return ErrorCode;
    ErrorCode = DumpEscapedString (StringPntr, 17 /* already output */);
    if (ErrorCode != B_OK) return ErrorCode;
  }

  ErrorCode = DumpRawString ("\n");
  if (ErrorCode != B_OK) return ErrorCode;

  ErrorCode = DumpUnixStandardMetadata (false /* DumpSize */);
  if (ErrorCode != B_OK) return ErrorCode;

  ErrorCode = DumpRawString ("}\n\n");
  if (ErrorCode != B_OK) return ErrorCode;

  ErrorCode = PackAttributes (CurrentNode);
  if (ErrorCode != B_OK) return ErrorCode;

  return B_OK;
}


/* Do what the scripting command says.  A reply message will be sent back with
several fields: "error" containing the numerical error code (0 for success),
"CommandText" with a text representation of the command, "result" with the
resulting data for a get or count command.  If it isn't understood, then rather
than a B_REPLY kind of message, it will be a B_MESSAGE_NOT_UNDERSTOOD message
with an "error" number and an "message" string with a description. */

void AFSPApp::ProcessScriptingMessage (
  BMessage *MessagePntr,
  struct property_info *PropInfoPntr)
{
  bool        ArgumentBool = false;
  bool        ArgumentGotBool = false;
  bool        ArgumentGotInt32 = false;
  bool        ArgumentGotString = false;
  int32       ArgumentInt32 = 0;
  const char *ArgumentString = NULL;
  BString     CommandText;
  status_t    ErrorCode;
  int         i;
  BMessage    ReplyMessage (B_MESSAGE_NOT_UNDERSTOOD);
  ssize_t     StringBufferSize;
  BMessage    TempBMessage;
  BPath       TempPath;
  char        TempString [ERROR_MESSAGE_SIZE];

  SetCursor (g_BusyCursor);

  ErrorCode = MessagePntr->FindData (g_DataName, B_STRING_TYPE,
    (const void **) &ArgumentString, &StringBufferSize);
  if (ErrorCode == B_OK)
  {
    if (strlen (ArgumentString) >= PATH_MAX)
    {
      sprintf (TempString, "\"data\" string of a scripting message is too "
        "long, for SET %s action", PropInfoPntr->name);
      ErrorCode = B_NAME_TOO_LONG;
      goto ErrorExit;
    }
    ArgumentGotString = true;
  }
  else if (MessagePntr->FindBool (g_DataName, &ArgumentBool) == B_OK)
    ArgumentGotBool = true;
  else if (MessagePntr->FindInt32 (g_DataName, &ArgumentInt32) == B_OK)
    ArgumentGotInt32 = true;

  /* Prepare a Human readable description of the scripting command. */

  switch (PropInfoPntr->commands[0])
  {
    case B_SET_PROPERTY:
      CommandText.SetTo ("Set ");
      break;

    case B_GET_PROPERTY:
      CommandText.SetTo ("Get ");
      break;

    case B_COUNT_PROPERTIES:
      CommandText.SetTo ("Count ");
      break;

    case B_CREATE_PROPERTY:
      CommandText.SetTo ("Create ");
      break;

    case B_DELETE_PROPERTY:
      CommandText.SetTo ("Delete ");
      break;

    case B_EXECUTE_PROPERTY:
      CommandText.SetTo ("Execute ");
      break;

    default:
      sprintf (TempString, "Bug: scripting command for \"%s\" has an unknown "
        "action code %d", PropInfoPntr->name,
        (int) PropInfoPntr->commands[0]);
      ErrorCode = -1;
      goto ErrorExit;
  }
  CommandText.Append (PropInfoPntr->name);

  /* Add on the argument value to our readable command, if there is one. */

  if (ArgumentGotString)
  {
    CommandText.Append (" \"");
    CommandText.Append (ArgumentString);
    CommandText.Append ("\"");
  }
  if (ArgumentGotBool)
    CommandText.Append (ArgumentBool ? " true" : " false");
  if (ArgumentGotInt32)
  {
    sprintf (TempString, " %ld", ArgumentInt32);
    CommandText.Append (TempString);
  }

  /* From now on the scripting command has been recognized and is in the
  correct format, so it always returns a B_REPLY message.  A readable version
  of the command is also added to make debugging easier. */

  ReplyMessage.what = B_REPLY;
  ReplyMessage.AddString ("CommandText", CommandText);

  /* Now actually do the command.  First prepare a default error message. */

  sprintf (TempString, "Operation code %d (get, set, count, etc) "
    "unsupported for property %s",
    (int) PropInfoPntr->commands[0], PropInfoPntr->name);
  ErrorCode = B_BAD_INDEX;

  switch (PropInfoPntr->extra_data)
  {
    case PN_PACKED_FILE:
      switch (PropInfoPntr->commands[0])
      {
        case B_GET_PROPERTY: /* Get the packed file name. */
          ReplyMessage.AddString (g_ResultName, m_PackedFileName.Path());
          break;

        case B_SET_PROPERTY: /* Set the packed file name to a new one. */
          if (!ArgumentGotString)
          {
            ErrorCode = B_BAD_TYPE;
            sprintf (TempString, "You need to specify a string for the "
              "SET %s command", PropInfoPntr->name);
            goto ErrorExit;
          }
          ErrorCode = TempPath.SetTo (ArgumentString, NULL /* leaf */,
            true /* normalize - verifies parent directories exist */);
          if (ErrorCode != B_OK)
          {
            sprintf (TempString, "New packed file name of \"%s\" is invalid "
              "(parent directories must exist)", ArgumentString);
            goto ErrorExit;
          }
          m_PackedFileName = TempPath;
          m_SettingsHaveChanged = true;
          break;

        default: /* Unknown operation code, error message already set. */
          goto ErrorExit;
      }
      break;

    case PN_TOP_LEVEL_DIRECTORY:
      switch (PropInfoPntr->commands[0])
      {
        case B_GET_PROPERTY: /* Get the load directory name. */
          ReplyMessage.AddString (g_ResultName, m_TopLevelDirectoryName.Path());
          break;

        case B_SET_PROPERTY: /* Set the load directory name to a new one. */
          if (!ArgumentGotString)
          {
            ErrorCode = B_BAD_TYPE;
            sprintf (TempString, "You need to specify a string for the "
              "SET %s command", PropInfoPntr->name);
            goto ErrorExit;
          }
          ErrorCode = TempPath.SetTo (ArgumentString, NULL /* leaf */,
            true /* normalize - verifies parent directories exist */);
          if (ErrorCode != B_OK)
          {
            sprintf (TempString, "New load directory name of \"%s\" is "
              "invalid (parent directories must exist)", ArgumentString);
            goto ErrorExit;
          }
          m_TopLevelDirectoryName = TempPath;
          m_SettingsHaveChanged = true;
          break;

        default: /* Unknown operation code, error message already set. */
          goto ErrorExit;
      }
      break;

    case PN_CREATE_FILE_SYSTEM:
      if (PropInfoPntr->commands[0] == B_EXECUTE_PROPERTY &&
      (ErrorCode = DoCreateFS (TempString)) == B_OK)
        break;
      goto ErrorExit;

    case PN_PACK:
      if (PropInfoPntr->commands[0] == B_EXECUTE_PROPERTY &&
      (ErrorCode = DoPacking (TempString)) == B_OK)
        break;
      goto ErrorExit;

    case PN_UNPACK:
      if (PropInfoPntr->commands[0] == B_EXECUTE_PROPERTY &&
      (ErrorCode = DoUnpacking (TempString)) == B_OK)
        break;
      goto ErrorExit;

    case PN_COMPRESSION_METHOD:
      switch (PropInfoPntr->commands[0])
      {
        case B_GET_PROPERTY: /* Get the compression method name. */
          ReplyMessage.AddString (g_ResultName,
            g_CompressionMethodNames [m_CompressionMethod]);
          break;

        case B_SET_PROPERTY: /* Set the compression method to a new one. */
          if (!ArgumentGotString)
          {
            ErrorCode = B_BAD_TYPE;
            sprintf (TempString, "You need to specify a string for the "
              "SET %s command", PropInfoPntr->name);
            goto ErrorExit;
          }
          for (i = 0; i < CM_MAX; i++)
          {
            if (strcasecmp (g_CompressionMethodNames[i], ArgumentString) == 0)
            {
              m_CompressionMethod = (CompressionMethod) i;
              break;
            }
          }
          if (i >= CM_MAX)
          {
            sprintf (TempString, "Compression method of \"%s\" isn't known, "
              "want one of: %s", ArgumentString, g_CompressionMethodNames[0]);
            for (i = 1; i < CM_MAX; i++)
              sprintf (TempString + strlen (TempString),
                ", %s", g_CompressionMethodNames[i]);
            ErrorCode = B_BAD_VALUE;
            goto ErrorExit;
          }
          m_SettingsHaveChanged = true;
          break;

        default: /* Unknown operation code, error message already set. */
          goto ErrorExit;
      }
      break;

    case PN_ZLIB_COMPRESSION_LEVEL:
      switch (PropInfoPntr->commands[0])
      {
        case B_GET_PROPERTY:
          ReplyMessage.AddInt32 (g_ResultName, m_zlibCompressionLevel);
          break;

        case B_SET_PROPERTY:
          if (!ArgumentGotInt32)
          {
            ErrorCode = B_BAD_TYPE;
            sprintf (TempString, "You need to specify a 32 bit integer "
              "for the SET %s command", PropInfoPntr->name);
            goto ErrorExit;
          }
          m_zlibCompressionLevel = ArgumentInt32;
          if (m_zlibCompressionLevel < 0)
            m_zlibCompressionLevel = 0;
          else if (m_zlibCompressionLevel > 9)
            m_zlibCompressionLevel = 9;
          m_SettingsHaveChanged = true;
          break;

        default: /* Unknown operation code, error message already set. */
          goto ErrorExit;
      }
      break;

    case PN_MAKE_RAM_DISK:
      switch (PropInfoPntr->commands[0])
      {
        case B_GET_PROPERTY:
          ReplyMessage.AddBool (g_ResultName, m_MakeRAMDisk);
          break;

        case B_SET_PROPERTY:
          if (!ArgumentGotBool)
          {
            ErrorCode = B_BAD_TYPE;
            sprintf (TempString, "You need to specify a boolean "
              "for the SET %s command", PropInfoPntr->name);
            goto ErrorExit;
          }
          m_MakeRAMDisk = ArgumentBool;
          m_SettingsHaveChanged = true;
          break;

        default: /* Unknown operation code, error message already set. */
          goto ErrorExit;
      }
      break;

    case PN_MAX_RAM_DISK_SIZE:
      switch (PropInfoPntr->commands[0])
      {
        case B_GET_PROPERTY:
          ReplyMessage.AddInt32 (g_ResultName, m_MaxRAMDiskSize);
          break;

        case B_SET_PROPERTY:
          if (!ArgumentGotInt32)
          {
            ErrorCode = B_BAD_TYPE;
            sprintf (TempString, "You need to specify a 32 bit integer "
              "for the SET %s command", PropInfoPntr->name);
            goto ErrorExit;
          }
          m_MaxRAMDiskSize = ArgumentInt32;
          if (m_MaxRAMDiskSize < 0)
            m_MaxRAMDiskSize = 0;
          m_SettingsHaveChanged = true;
          break;

        default: /* Unknown operation code, error message already set. */
          goto ErrorExit;
      }
      break;

    case PN_RAM_DISK_VIRTUAL_MEMORY:
      switch (PropInfoPntr->commands[0])
      {
        case B_GET_PROPERTY:
          ReplyMessage.AddBool (g_ResultName, m_RAMDiskVirtualMemory);
          break;

        case B_SET_PROPERTY:
          if (!ArgumentGotBool)
          {
            ErrorCode = B_BAD_TYPE;
            sprintf (TempString, "You need to specify a boolean "
              "for the SET %s command", PropInfoPntr->name);
            goto ErrorExit;
          }
          m_RAMDiskVirtualMemory = ArgumentBool;
          m_SettingsHaveChanged = true;
          break;

        default: /* Unknown operation code, error message already set. */
          goto ErrorExit;
      }
      break;

    case PN_RECREATE_INDICES:
      switch (PropInfoPntr->commands[0])
      {
        case B_GET_PROPERTY:
          ReplyMessage.AddBool (g_ResultName, m_RecreateIndices);
          break;

        case B_SET_PROPERTY:
          if (!ArgumentGotBool)
          {
            ErrorCode = B_BAD_TYPE;
            sprintf (TempString, "You need to specify a boolean "
              "for the SET %s command", PropInfoPntr->name);
            goto ErrorExit;
          }
          m_RecreateIndices = ArgumentBool;
          m_SettingsHaveChanged = true;
          break;

        default: /* Unknown operation code, error message already set. */
          goto ErrorExit;
      }
      break;

    case PN_NOTES:
      switch (PropInfoPntr->commands[0])
      {
        case B_GET_PROPERTY:
          ReplyMessage.AddString (g_ResultName, m_Notes.String ());
          break;

        case B_SET_PROPERTY:
          if (!ArgumentGotString)
          {
            ErrorCode = B_BAD_TYPE;
            sprintf (TempString, "You need to specify a string for the "
              "SET %s command", PropInfoPntr->name);
            goto ErrorExit;
          }
          m_Notes = ArgumentString;
          m_SettingsHaveChanged = true;
          break;

        default: /* Unknown operation code, error message already set. */
          goto ErrorExit;
      }
      break;

    case PN_RESET_TO_DEFAULTS:
      if (PropInfoPntr->commands[0] == B_EXECUTE_PROPERTY)
      {
        DefaultSettings ();
        break;
      }
      goto ErrorExit;

    case PN_INSTALL_THINGS:
      if (PropInfoPntr->commands[0] == B_EXECUTE_PROPERTY &&
      (ErrorCode = InstallThings (TempString)) == B_OK)
        break;
      goto ErrorExit;

    default:
      sprintf (TempString, "Bug!  Unrecognized property identification "
        "number %d (should be between 0 and %d).  Fix the entry in "
        "the g_ScriptingPropertyList array!",
        (int) PropInfoPntr->extra_data, PN_MAX - 1);
      goto ErrorExit;
  }

  /* Success. */

  ReplyMessage.AddInt32 ("error", B_OK);
  ErrorCode = MessagePntr->SendReply (&ReplyMessage,
    this /* Reply's reply handler */, 500000 /* send timeout */);
  if (ErrorCode != B_OK)
    cerr << "ProcessScriptingMessage failed to send a reply message, code " <<
    ErrorCode << " (" << strerror (ErrorCode) << ")" << " for " <<
    CommandText.String () << endl;
  SetCursor (B_CURSOR_SYSTEM_DEFAULT);
  return;

ErrorExit: /* Error message in TempString, return code in ErrorCode. */
  ReplyMessage.AddInt32 ("error", ErrorCode);
  ReplyMessage.AddString ("message", TempString);
  DisplayErrorMessage (TempString, ErrorCode);
  ErrorCode = MessagePntr->SendReply (&ReplyMessage,
    this /* Reply's reply handler */, 500000 /* send timeout */);
  if (ErrorCode != B_OK)
    cerr << "ProcessScriptingMessage failed to send an error message, code " <<
    ErrorCode << " (" << strerror (ErrorCode) << ")" << " for " <<
    CommandText.String () << endl;
  SetCursor (B_CURSOR_SYSTEM_DEFAULT);
}


/* Since quitting stops the program before the results of a script command
are received, we use a time delay to do the quit and make sure there are no
pending commands being processed by the auxiliary looper which is sending us
commands. */

void AFSPApp::Pulse ()
{
  if (g_QuitCountdown == 0)
  {
    if (g_CommanderLooperPntr == NULL ||
    !g_CommanderLooperPntr->IsBusy ())
      PostMessage (B_QUIT_REQUESTED);
  }
  else if (g_QuitCountdown > 0)
  {
    printf ("AGMSFileSystemPacker quitting in %d.\n", g_QuitCountdown);
    g_QuitCountdown--;
  }
}


/* A quit request message has come in.  If the quit countdown has reached zero,
allow the request, otherwise reject it (and start the countdown if it hasn't
been started). */

bool AFSPApp::QuitRequested ()
{
  BMessage  *QuitMessage;
  team_info  RemoteInfo;
  BMessenger RemoteMessenger;
  team_id    RemoteTeam;

  /* See if the quit is from the system shutdown command (which goes through
  the registrar server), if so, quit immediately. */

  QuitMessage = CurrentMessage ();
  if (QuitMessage != NULL && QuitMessage->IsSourceRemote ())
  {
    RemoteMessenger = QuitMessage->ReturnAddress ();
    RemoteTeam = RemoteMessenger.Team ();
    if (get_team_info (RemoteTeam, &RemoteInfo) == B_OK &&
    strstr (RemoteInfo.args, "registrar") != NULL)
      g_QuitCountdown = 0;
  }

  if (g_QuitCountdown == 0)
    return BApplication::QuitRequested ();

  if (g_QuitCountdown < 0)
    g_QuitCountdown = 3; /* Start the countdown. */

  return false;
}


/* Read all the data from the current position to the next real end of line
into a string, then de-escape the string.  Doesn't include the end of line in
the string, and leaves the read position just after the end of line LF.
Returns B_OK if successful, READERR_END_OF_FILE at end of file, or the usual
error codes.  ValueString will have a partial line of text when it returns with
an error. */

status_t AFSPApp::ReadAndUnescapeValue (BString &ValueString)
{
  int      DestinationLength;
  char    *EndPntr;
  bool     FoundEndOfLine = false;
  status_t ReturnCode = -1;
  char    *StringPntr;
  int      SourceLength;

  ValueString.Truncate (0);
  EndPntr = m_Recursive.inputBuffer + m_Recursive.inputLengthExported;
  StringPntr = m_Recursive.inputBuffer + m_Recursive.inputUnusedDataStart;
  while (!FoundEndOfLine)
  {
    /* Need enough text in the buffers for the end of a line.  Avoid checking
    too often since filling buffers is expensive. */

    if (EndPntr - StringPntr < 16)
    {
      /* Fill the buffer with more data.  inputUnusedDataStart is used too. */
      ReturnCode = ReadFillInputBuffer ();
      if (ReturnCode != B_OK)
        break;
      if (m_Recursive.inputUnusedDataStart >= m_Recursive.inputLengthExported)
      {
        sprintf (m_Recursive.errorMessage, "ReadAndUnescapeValue: "
          "Reached end of file \"%s\"", m_PackedFileName.Path ());
        ReturnCode = READERR_END_OF_FILE;
        break;
      }

      EndPntr = m_Recursive.inputBuffer + m_Recursive.inputLengthExported;
      StringPntr = m_Recursive.inputBuffer + m_Recursive.inputUnusedDataStart;
    }

    while (StringPntr < EndPntr)
    {
      /* Look for the end of a line.  It's a LF character that is not preceeded
      by a backslash.  Note that when the buffer is refilled, it keeps the last
      two read characters before the new data. */

      if (*StringPntr == '\n' &&
      (StringPntr <= m_Recursive.inputBuffer || StringPntr[-1] != '\\'))
      {
        StringPntr++; /* Skip over linefeed character.  It isn't stored. */
        FoundEndOfLine = true;
        ReturnCode = B_OK;
        break;
      }
      ValueString.Append (*StringPntr, 1);
      StringPntr++;
    }

    m_Recursive.inputUnusedDataStart = StringPntr - m_Recursive.inputBuffer;
  }

  /* Remove escape codes from the string.  Since this results in a shorter
  string, we can do it in place, no need to use a separate buffer. */

  SourceLength = ValueString.Length ();
  DestinationLength = SourceLength;
  StringPntr = ValueString.LockBuffer (SourceLength + 1 /* avoid zero bug */);
  if (StringPntr == NULL)
  {
    strcpy (m_Recursive.errorMessage, "ReadAndUnescapeValue: "
      "Unable to lock string buffer");
    return -1;
  }
  RemoveBackslashEscapeCodes (StringPntr, &SourceLength,
    StringPntr, &DestinationLength, NULL);
  ValueString.UnlockBuffer (DestinationLength);

  return ReturnCode;
}


/* Fill the input buffer as full as possible with data from the packed archive
file.  This includes parsing the compression blocks and switching compression
modes, and all the buffering complications that implies.  At end of file,
inputUnusedDataStart stays equal to inputLengthExported. */

status_t AFSPApp::ReadFillInputBuffer ()
{
  int      AmountRead;
  int      AmountToMove;
  int      AmountToRemove;
  int      AmountToRead;
  char    *EndPntr;
  status_t ErrorCode;
  bool     FoundCompressionBlockEnd;
  char    *HeaderStartPntr;
  BString  MethodValue;
  int      NextHeaderSearchStartIndex;
  int      ReadIterations;
  char    *StringPntr;

  for (ReadIterations = 0; ReadIterations < 4; ReadIterations++)
  {
    if (m_CancelRequestedByGUI)
    {
      sprintf (m_Recursive.errorMessage, "ReadFillInputBuffer: "
        "Cancelled by user while processing \"%s\"",
        m_Recursive.currentPathName.Path ());
      return B_INTERRUPTED;
    }

    /* Fill the raw data buffer with data from the archive file.  For string
    search convenience when looking for the next compression header, put a NUL
    byte at the end of the buffer but don't include it in the buffer length. */

    AmountToRead = INPUT_OUTPUT_BUFFER_SIZE - 1 - m_Recursive.rawDataLength;
    if (AmountToRead > 0 && !feof (m_Recursive.stream))
    {
      AmountRead = fread (
        m_Recursive.rawDataBuffer + m_Recursive.rawDataLength,
        1, AmountToRead, m_Recursive.stream);
      if (AmountRead <= 0)
      {
        if (AmountRead < 0 || ferror (m_Recursive.stream))
        {
          ErrorCode = errno;
          if (ErrorCode >= 0)
            ErrorCode = -1;
          sprintf (m_Recursive.errorMessage, "ReadFillInputBuffer: "
            "Tried to read %d bytes from \"%s\", but got %d",
            AmountToRead, m_PackedFileName.Path (), AmountRead);
          return ErrorCode;
        }
      }
      else /* Successfully read some data. */
      {
        m_Recursive.rawDataLength += AmountRead;
        m_Recursive.rawDataBuffer [m_Recursive.rawDataLength] = 0;
      }
    }

    /* Remove most of the used up data from the input buffer.  But keep the
    previously read last couple of bytes so that higher level code can easily
    check to see if the previous two bytes were the start of a new line (a
    non-backslash followed by a linefeed character). */

    AmountToRemove = m_Recursive.inputUnusedDataStart - 2;
    if (AmountToRemove > 0)
    {
      AmountToMove = m_Recursive.inputLength - AmountToRemove;
      if (AmountToMove > 0)
        memmove (m_Recursive.inputBuffer,
        m_Recursive.inputBuffer + AmountToRemove, AmountToMove);
      m_Recursive.inputLength -= AmountToRemove;
      m_Recursive.inputLengthExported -= AmountToRemove;
      m_Recursive.inputUnusedDataStart -= AmountToRemove;
      m_Recursive.inputBuffer[m_Recursive.inputLength] = 0;
    }

    /* Move data from the raw data buffer to the input buffer. */

    if (m_Recursive.currentCompressionMethod == CM_ZLIB)
    {
      /* Decompress data from the raw data buffer to the input buffer.  Stops
      decompressing after the end of the data stream is reached, so that the
      next compression method can start immediately after. */

      if (!m_Recursive.zlibInitialised)
      {
        strcpy (m_Recursive.errorMessage, "ReadFillInputBuffer: "
          "Bug!  ZLib wasn't initialised");
        return -1;
      }
      if (m_Recursive.rawDataLength > 0 ||
      INPUT_OUTPUT_BUFFER_SIZE - 1 - m_Recursive.inputLength > 0)
      {
        m_Recursive.zlibStream.next_in = (uint8 *) m_Recursive.rawDataBuffer;
        m_Recursive.zlibStream.avail_in = m_Recursive.rawDataLength;
        m_Recursive.zlibStream.next_out =
          (uint8 *) m_Recursive.inputBuffer + m_Recursive.inputLength;
        m_Recursive.zlibStream.avail_out = INPUT_OUTPUT_BUFFER_SIZE -
          1 /* for NUL at end of inputBuffer */ - m_Recursive.inputLength;
        ErrorCode = inflate (&m_Recursive.zlibStream, Z_NO_FLUSH);
        if (ErrorCode >= 0) /* Can be one of several success codes. */
        {
          /* Move unread raw data to the front of the buffer. */
          AmountRead =
            m_Recursive.rawDataLength - m_Recursive.zlibStream.avail_in;
          m_Recursive.rawDataLength -= AmountRead;
          if (m_Recursive.rawDataLength > 0)
            memmove (m_Recursive.rawDataBuffer,
            m_Recursive.rawDataBuffer + AmountRead,
            m_Recursive.rawDataLength);
          /* Update the inputBuffer size for the added decompressed data. */
          AmountRead =
            INPUT_OUTPUT_BUFFER_SIZE - 1 - m_Recursive.inputLength -
            m_Recursive.zlibStream.avail_out;
          m_Recursive.inputLength += AmountRead;
          m_Recursive.inputBuffer[m_Recursive.inputLength] = 0;
        }
        if (ErrorCode == Z_STREAM_END)
        {
          inflateEnd (&m_Recursive.zlibStream);
          m_Recursive.zlibInitialised = false;
          m_Recursive.currentCompressionMethod = CM_MAX; /* Now no method. */
        }
        else if (ErrorCode != Z_OK)
        {
          sprintf (m_Recursive.errorMessage, "ReadFillInputBuffer: "
            "ZLib decompression failed, ZLib error code %ld, message \"%s\", "
            "while reading file \"%s\"",
            ErrorCode, m_Recursive.zlibStream.msg, m_PackedFileName.Path ());
          return -1;
        }
      }
    }
    else if (m_Recursive.currentCompressionMethod == CM_NONE)
    {
      /* Copy data up to but not including the next compression header
      (presumed to be under 2K bytes long), until we have the whole compression
      header.  When the header end has been found then it gets copied too,
      stopping just after the last byte in the header.  If we can't find a
      compression header, then leave about 2K of data in the buffer for next
      time, in case the compression header is split between buffers.  Search
      for the "{ Compression" marker in the raw data by first searching for the
      start of the next real line (not just a continuation line) of text.  It's
      a linefeed with the previous character not being a backslash (the saving
      code escapes blackslashes as numbers rather than a double backslash to
      avoid problems here). */

      FoundCompressionBlockEnd = false;
      EndPntr = m_Recursive.rawDataBuffer + m_Recursive.rawDataLength;
      StringPntr = m_Recursive.rawDataBuffer + 1;
      while (StringPntr < EndPntr)
      {
        /* Look for end of line, not preceeded by a backslash. */
        if (StringPntr[0] != '\n' || StringPntr[-1] == '\\')
        {
          StringPntr++;
          continue;
        }
        StringPntr++;
        /* Skip leading spaces.  Can't use isspace() since it does LFs too. */
        while (*StringPntr == ' ' || *StringPntr == '\t')
          StringPntr++;
        if (*StringPntr != '{')
          continue;
        StringPntr++;
        while (*StringPntr == ' ' || *StringPntr == '\t')
          StringPntr++;
        /* Continue searching if it is some other kind of block start. */
        if (strncasecmp (StringPntr, "Compression", 11) != 0)
          continue;
        StringPntr += 11;

        /* Found the start of the compression block, look for the end (a line
        with a closing curly bracket), assume there are no nested blocks. */

        while (StringPntr < EndPntr)
        {
          /* Advance to the start of the next line. */
          while (*StringPntr != '\n' && StringPntr < EndPntr)
            StringPntr++;
          if (StringPntr < EndPntr)
            StringPntr++;
          /* Skip leading spaces. */
          while (*StringPntr == ' ' || *StringPntr == '\t')
            StringPntr++;
          /* Bracket marking the end of the block? */
          if (*StringPntr != '}')
            continue;
          while (*StringPntr != '\n' && StringPntr < EndPntr)
            StringPntr++;
          break;
        }
        if (StringPntr < EndPntr && *StringPntr == '\n')
        {
          /* Found the end of the compression header.  Can pass on the data
          up to the end of the compression header to the next higher
          buffering level.  But stop exactly at the end.  Also have to pass
          on the whole header, keep it in our buffer if there isn't space in
          the next higher buffer. */

          StringPntr++;
          FoundCompressionBlockEnd = true;
          break;
        }
      }
      if (FoundCompressionBlockEnd)
      {
        AmountToRead = StringPntr - m_Recursive.rawDataBuffer;
        if (AmountToRead >
        INPUT_OUTPUT_BUFFER_SIZE - 1 - m_Recursive.inputLength)
        {
          /* Oops, can't fit the whole compression header into the input
          buffer, so avoid copying any of the header for now.  Assume it is
          less than 2K long. */
          AmountToRead -= 2048;
          if (AmountToRead < 0)
            AmountToRead = 0;
        }
        else /* Finished this compression method's data. */
          m_Recursive.currentCompressionMethod = CM_MAX; /* Now no method. */
      }
      else
      {
        /* Didn't find the compression header.  Leave 2K of data in the buffer
        to avoid splitting the compression header between buffers, except when
        at end of file (no upcoming compression header exists). */
        if (feof (m_Recursive.stream))
          AmountToRead = m_Recursive.rawDataLength;
        else
          AmountToRead = m_Recursive.rawDataLength - 2048;
      }

      if (AmountToRead >
      INPUT_OUTPUT_BUFFER_SIZE - 1 -  m_Recursive.inputLength)
        AmountToRead = INPUT_OUTPUT_BUFFER_SIZE -
        1 /* save space for NUL at end */ - m_Recursive.inputLength;
      if (AmountToRead > 0)
      {
        memcpy (m_Recursive.inputBuffer + m_Recursive.inputLength,
          m_Recursive.rawDataBuffer, AmountToRead);
        m_Recursive.inputLength += AmountToRead;
        m_Recursive.inputBuffer[m_Recursive.inputLength] = 0;
        if (m_Recursive.rawDataLength - AmountToRead > 0)
          memmove (m_Recursive.rawDataBuffer,
          m_Recursive.rawDataBuffer + AmountToRead,
          m_Recursive.rawDataLength - AmountToRead);
        m_Recursive.rawDataLength -= AmountToRead;
        m_Recursive.rawDataBuffer[m_Recursive.rawDataLength] = 0;
      }
      if (m_Recursive.rawDataLength <= 0 && feof (m_Recursive.stream))
        m_Recursive.currentCompressionMethod = CM_MAX; /* Finished data. */
    }

    /* If the last compression method has finished, see if there is a
    compression header in the inputBuffer, which will tell us the next
    compression method to use.  The header also gets removed from the input
    buffer so that the higher levels just see pure decompressed data. */

    if (m_Recursive.currentCompressionMethod == CM_MAX)
    {
      /* We only have to start the search in the last 2K of the buffer since
      the decompression methods will have decompressed the header as the last
      thing they did - so it should be near the end of the buffer.  And if that
      includes a bit of previously read data, that doesn't matter since it
      won't have any compression headers in it. */

      NextHeaderSearchStartIndex = m_Recursive.inputLength - 2048;
      if (NextHeaderSearchStartIndex < 0)
        NextHeaderSearchStartIndex = 0;
      FoundCompressionBlockEnd = false;
      HeaderStartPntr = NULL;
      EndPntr = m_Recursive.inputBuffer + m_Recursive.inputLength;
      StringPntr = m_Recursive.inputBuffer + NextHeaderSearchStartIndex + 1;
      while (StringPntr < EndPntr)
      {
        /* Look for end of line, not preceeded by a backslash. */
        if (StringPntr[0] != '\n' || StringPntr[-1] == '\\')
        {
          StringPntr++;
          continue;
        }
        StringPntr++;
        HeaderStartPntr = StringPntr;
        /* Skip leading spaces.  Can't use isspace() since it does LFs too. */
        while (*StringPntr == ' ' || *StringPntr == '\t')
          StringPntr++;
        if (*StringPntr != '{')
          continue;
        StringPntr++;
        while (*StringPntr == ' ' || *StringPntr == '\t')
          StringPntr++;
        /* Continue searching if it is some other kind of block start. */
        if (strncasecmp (StringPntr, "Compression", 11) != 0)
          continue;
        StringPntr += 11;

        /* Found the start of the compression block, look for the compression
        method (save the value found there) and the end of the compression
        block.  The method is on a line that starts with "Method=". */

        MethodValue = "Unspecified";
        while (StringPntr < EndPntr)
        {
          /* Advance to the start of the next line. */
          while (*StringPntr != '\n' && StringPntr < EndPntr)
            StringPntr++;
          if (StringPntr < EndPntr)
            StringPntr++;
          /* Skip leading spaces. */
          while (*StringPntr == ' ' || *StringPntr == '\t')
            StringPntr++;
          if (*StringPntr == '}')
          {
            while (*StringPntr != '\n' && StringPntr < EndPntr)
              StringPntr++;
            if (StringPntr < EndPntr)
              StringPntr++;
            FoundCompressionBlockEnd = true;
            break; /* Got to the end of the block. */
          }
          if (strncasecmp (StringPntr, "Method=", 7) != 0)
            continue; /* Wrong line, some other keyword we ignore. */
          StringPntr += 7;
          MethodValue.Truncate (0);
          while (*StringPntr != '\n' && StringPntr < EndPntr)
            MethodValue.Append (*StringPntr++, 1);
        }
      }

      if (FoundCompressionBlockEnd)
      {
        /* Remove the compression block from the buffer, so that the higher
        levels don't see it at all. */

        AmountToMove =
          m_Recursive.inputLength - (StringPntr - m_Recursive.inputBuffer);
        if (AmountToMove < 0)
        {
          strcpy (m_Recursive.errorMessage, "ReadFillInputBuffer: "
            "Bug!  Negative AmountToMove when removing the "
            "compression block");
          return -1;
        }
        if (AmountToMove > 0)
          memmove (HeaderStartPntr, StringPntr, AmountToMove);
        m_Recursive.inputLength = AmountToMove +
          (HeaderStartPntr - m_Recursive.inputBuffer);
        /* Exported length etc. updated later. */

        /* Switch to the new compression method. */

        m_Recursive.currentCompressionMethod = (CompressionMethod) 0;
        while (m_Recursive.currentCompressionMethod < CM_MAX)
        {
          if (0 == strcasecmp (g_CompressionMethodNames
          [m_Recursive.currentCompressionMethod], MethodValue.String ()))
            break;
          m_Recursive.currentCompressionMethod =
            (CompressionMethod) (m_Recursive.currentCompressionMethod + 1);
        }
        if (m_Recursive.currentCompressionMethod >= CM_MAX)
        {
          MethodValue.Truncate (100);
          sprintf (m_Recursive.errorMessage, "ReadFillInputBuffer: "
            "Didn't recognize compression method \"%s\" in a "
            "compression block, while reading \"%s\"",
            MethodValue.String (), m_PackedFileName.Path ());
          return -1;
        }
        if (m_Recursive.currentCompressionMethod == CM_ZLIB)
        {
          memset (&m_Recursive.zlibStream, 0, sizeof (m_Recursive.zlibStream));
          m_Recursive.zlibStream.zalloc = MyAllocForZLib;
          m_Recursive.zlibStream.zfree = MyFreeForZLib;
          m_Recursive.zlibStream.opaque = this;

          ErrorCode = inflateInit (&m_Recursive.zlibStream);
          if (ErrorCode != Z_OK)
          {
            sprintf (m_Recursive.errorMessage, "ReadFillInputBuffer: "
              "ZLib initialisation failed, ZLib error code %ld, "
              "message \"%s\"", ErrorCode, m_Recursive.zlibStream.msg);
            return -1;
          }
          m_Recursive.zlibInitialised = true;
        }
      }
    }

    if (m_Recursive.currentCompressionMethod == CM_MAX)
    {
      m_Recursive.inputLengthExported = m_Recursive.inputLength;
      break; /* No more reading possible.  Done. */
    }
    else
    {
      /* Haven't reached the next compression header, but may have part of one,
      so keep aside the last 2K of data until the current compression method
      has reached the end of its stream. */

      m_Recursive.inputLengthExported = m_Recursive.inputLength - 2048;
      if (m_Recursive.inputLengthExported < 0)
        m_Recursive.inputLengthExported = 0;
      if (m_Recursive.inputLength >= INPUT_OUTPUT_BUFFER_SIZE / 2)
        break; /* Input buffer is reasonably full.  Done. */
    }
  }

  return B_OK;
}


status_t AFSPApp::ReadIndexCommand ()
{
  UpdateNowProcessingPath ("Read index command...");
// bleeble;
  return B_OK;
}


status_t AFSPApp::ReadLinkCommand (bool CreateCommand)
{
  UpdateNowProcessingPath ("Read link command...");
// bleeble;
  return B_OK;
}


/* Scans forward in the input until it finds the start of the next line,
returns the command words it finds there (skips leading spaces, includes
everything up to and including the equals sign, or up to but not including the
end of line, whichever comes first).  Also updates the commandNestingLevel,
incrementing it whenever a block is entered (a line starting with "{") and
decrementing it when a block finishes (a line starting with "}".  Returns the
usual error codes for errors, READERR_END_OF_FILE if it can't find the next
line (end of file reached), B_OK if successful. */

status_t AFSPApp::ReadNextLine (BString &InitialText)
{
  char    *EndPntr;
  status_t ErrorCode;
  char    *StringPntr;

  InitialText.Truncate (0);
  EndPntr = m_Recursive.inputBuffer + m_Recursive.inputLengthExported;
  StringPntr = m_Recursive.inputBuffer + m_Recursive.inputUnusedDataStart;
  while (true)
  {
    /* Need enough text in the buffers for the start of a line, which is either
    a command or a parameter name or a comment line.  None are very long. */

    if (EndPntr - StringPntr < 1024)
    {
      /* Fill the buffer with more data.  inputUnusedDataStart is used too. */
      ErrorCode = ReadFillInputBuffer ();
      if (ErrorCode != B_OK)
        return ErrorCode;
      if (m_Recursive.inputUnusedDataStart >= m_Recursive.inputLengthExported)
      {
        sprintf (m_Recursive.errorMessage, "ReadNextLine: "
          "Reached end of file \"%s\", so can't find the next line",
          m_PackedFileName.Path ());
        return READERR_END_OF_FILE;
      }

      EndPntr = m_Recursive.inputBuffer + m_Recursive.inputLengthExported;
      StringPntr = m_Recursive.inputBuffer + m_Recursive.inputUnusedDataStart;
    }

    while (StringPntr < EndPntr)
    {
      /* Look for the start of a line.  It's either at the start of the buffer
      (no previous lines read yet) or the previous two characters are an end of
      line, not preceeded by a backslash.  Note that when the buffer is
      refilled, it keeps the last two read characters before the new data. */
      if (StringPntr - m_Recursive.inputBuffer >= 2 &&
      (StringPntr[-1] != '\n' || StringPntr[-2] == '\\'))
      {
        StringPntr++;
        continue;
      }
      /* Skip leading spaces.  Can't use isspace() since it does LFs too. */
      while (StringPntr < EndPntr &&
      (*StringPntr == ' ' || *StringPntr == '\t'))
        StringPntr++;
      /* Update the block nesting level - lines starting with "{" increment it
      and starting with "}" decrement it.  Presumably all command lines get
      read through this function. */
      if (StringPntr < EndPntr)
      {
        if (*StringPntr == '{')
          m_Recursive.commandNestingLevel++;
        else if (*StringPntr == '}')
        {
          m_Recursive.commandNestingLevel--;
          if (m_Recursive.commandNestingLevel < 0)
            m_Recursive.commandNestingLevel = 0;
        }
      }
      /* Save the rest of the line as the returned text. */
      while (StringPntr < EndPntr && *StringPntr != '\n' && *StringPntr != '=')
        InitialText.Append (*StringPntr++, 1);
      if (StringPntr < EndPntr)
      {
        if (*StringPntr == '=')
          InitialText.Append (*StringPntr++, 1);
        else /* Just skip the line feed, don't store it. */
          StringPntr++;
      }
      m_Recursive.inputUnusedDataStart = StringPntr - m_Recursive.inputBuffer;
      return B_OK;
    }
    m_Recursive.inputUnusedDataStart = StringPntr - m_Recursive.inputBuffer;
  }
  return -1; /* Should never get here. */
}


/* Read the next parameter name.  Returns the name (with leading spaces and
trailing equals sign removed) if successful, leaving the stream positioned just
after the equals sign.  Returns the usual IO errors or READERR_END_OF_FILE if
it hit end of file, or READERR_END_OF_NESTING if it read the end of a nested
command block (a line with a "}").  Will skip nested command blocks if
requested, otherwise it enters them and returns the names in them. */

status_t AFSPApp::ReadNextParameterName (
  BString &ParameterName,
  bool SkipNestedBlocks)
{
  status_t ErrorCode;
  int      StartingNestingLevel;

  StartingNestingLevel = m_Recursive.commandNestingLevel;
  while (true)
  {
    ErrorCode = ReadNextLine (ParameterName);
    if (ErrorCode != B_OK)
      return ErrorCode;

    if (m_Recursive.commandNestingLevel < StartingNestingLevel)
      return READERR_END_OF_NESTING;

    if (SkipNestedBlocks &&
    m_Recursive.commandNestingLevel > StartingNestingLevel)
      continue; /* Ignore lines from inside a nested block. */

    if (ParameterName.ByteAt (0) == '#')
      continue; /* Ignore comment lines. */

    if (ParameterName.ByteAt (ParameterName.Length () - 1
    /* negative indices return a NUL byte */) != '=')
      continue; /* Not a parameter name, doesn't end with equals sign. */

    /* Remove the trailing equals sign for easier comparisons. */
    ParameterName.Truncate (ParameterName.Length () - 1);
    break;
  }

  return B_OK;
}


/* Process one of the standard metadata (date stamps, security permissions,
etc) named values and apply it to the given BStatable file system object.
Returns TRUE if it actually tried to use the named value (was given a
recognized name in NamePntr and StatablePntr was non-NULL).  FALSE if it didn't
try to do anything (NULL StatablePntr or name is unknown).  If it returned
TRUE, then it will also update ErrorCode with the usual error codes. */

bool AFSPApp::ReadMetaData (
  const char *NamePntr,
  const char *ValuePntr,
  BEntry *StatablePntr,
  bool IsAttribute,
  off_t ActualDataSize,
  const char *UserPathName,
  status_t &ErrorCode)
{
  char *EndPntr;
  int64 IntegerValue;

  if (strncasecmp (NamePntr, "Meta", 4) != 0)
    return false;

  if (StatablePntr == NULL || StatablePntr->InitCheck () != B_OK)
    return false;

  IntegerValue = strtoll (ValuePntr, &EndPntr, 0 /* base */);
  if (EndPntr == ValuePntr)
  {
    sprintf (m_Recursive.errorMessage, "ReadMetaData: "
      "Unable to convert %s value \"", NamePntr);
    EndPntr = m_Recursive.errorMessage + strlen (m_Recursive.errorMessage);
    strncpy (EndPntr, ValuePntr, NAME_MAX);
    EndPntr [NAME_MAX] = 0;
    sprintf (m_Recursive.errorMessage + strlen (m_Recursive.errorMessage),
      "\" into a number, while processing \"%s\"", UserPathName);
    ErrorCode = B_BAD_VALUE;
    return true;
  }

  /* Attributes have a size, as do files and directories (zero). */

  if (strcasecmp (NamePntr, "MetaSize") == 0)
  {
    ErrorCode = B_OK;
    if (IntegerValue != ActualDataSize)
    {
      sprintf (m_Recursive.errorMessage, "ReadMetaData: "
        "Actual data size of %Ld doesn't match recorded size of %Ld for "
        "Thing \"%s\"", ActualDataSize, IntegerValue, UserPathName);
      ErrorCode = B_MISMATCHED_VALUES;
    }
    return true;
  }

  if (IsAttribute)
    return false; /* Attributes don't have any other kinds of metadata. */

  if (strcasecmp (NamePntr, "MetaOwnerID") == 0)
    ErrorCode = StatablePntr->SetOwner (IntegerValue);
  else if (strcasecmp (NamePntr, "MetaGroupID") == 0)
    ErrorCode = StatablePntr->SetGroup (IntegerValue);
  else if (strcasecmp (NamePntr, "MetaModeBits") == 0)
    ErrorCode = StatablePntr->SetPermissions (IntegerValue & S_IUMSK);
  else if (strcasecmp (NamePntr,
  "MetaDateCreatedMicrosecondsSince1970") == 0)
    ErrorCode = StatablePntr->SetCreationTime (IntegerValue / 1000000);
  else if (strcasecmp (NamePntr,
  "MetaDateAccessedMicrosecondsSince1970") == 0)
    ErrorCode = StatablePntr->SetAccessTime (IntegerValue / 1000000);
  else if (strcasecmp (NamePntr,
  "MetaDateModifiedMicrosecondsSince1970") == 0)
    ErrorCode = StatablePntr->SetModificationTime (IntegerValue / 1000000);
  else if (strcasecmp (NamePntr,
  "MetaDateMetaChangedMicrosecondsSince1970") == 0)
    ErrorCode = B_OK; /* For some reason, security?, you can't set it. */
  else
    return false; /* Unrecognized parameter name. */

  if (ErrorCode != B_OK)
    sprintf (m_Recursive.errorMessage, "ReadMetaData: "
      "Unable to set %s to %Ld for Thing \"%s\"",
      NamePntr, IntegerValue, UserPathName);

  return true;
}


/* Read a Thing from the archive and create it (if CreateCommand is TRUE) or
modify it (CreateCommand FALSE).  The Thing can be a directory, file or
attribute.  Returns the usual errors if it fails. */

status_t AFSPApp::ReadThingCommand (bool CreateCommand)
{
  enum ThingTypesEnum {
    THING_ATTRIBUTE,
    THING_DIRECTORY,
    THING_FILE,
    THING_MAX
  }           ThingType;

  off_t       ActualDataSize = -12345;
  type_code   AttributeTypeCode = B_STRING_TYPE;
  int         DestinationLength;
  BDirectory  Directory;
  bool        DoubleEscapedPathLeaf = false;
  char       *EndPntr;
  status_t    ErrorCode;
  int32       EscapeState;
  BFile       File;
  BString     LeafName;
  char       *LeafPntr;
  BNode       Node;
  BString     ParameterName;
  BString     ParameterValue;
  BPath       ParentPath;
  BString     PathNameForUser;
  int         SourceLength;
  BEntry      StatableEntry;
  char       *StringPntr;
  BString     TempBString;
  const char *TopLevelPathString;
  int         TopLevelPathStringLength;

  PathNameForUser = "No path known yet...";
  m_Recursive.currentPathName.Unset ();
  ThingType = THING_MAX;

  while (true)
  {
    /* Read the parameter name. */

    ErrorCode = ReadNextParameterName (ParameterName, true /* skip nested */);
    if (ErrorCode == READERR_END_OF_NESTING)
      break;
    if (ErrorCode != B_OK)
    {
      sprintf (m_Recursive.errorMessage, "ReadThingCommand: "
        "Didn't get to end of the Thing information block for \"%s\"",
        PathNameForUser.String ());
      return ErrorCode;
    }
    ParameterName.Truncate (100); /* So later sprintf doesn't overflow. */

    /* Read the parameter value, but only for small parameters, since BStrings
    can't hold multi-gigabyte sized data items. */

    if (ParameterName.ICompare ("Data") == 0)
      ParameterValue.Truncate (0); /* Don't read Data= since it may be huge. */
    else
    {
      ErrorCode = ReadAndUnescapeValue (ParameterValue);
      if (ErrorCode != B_OK)
      {
        sprintf (m_Recursive.errorMessage, "ReadThingCommand: "
          "Didn't get the value for parameter \"%s\" in Thing \"%s\"",
          ParameterName.String(), PathNameForUser.String ());
        return ErrorCode;
      }
    }

    /* Handle each kind of parameter... */

    if (ThingType == THING_MAX &&
    ParameterName.ICompare ("EquivalentOldThing") != 0)
    {
      sprintf (m_Recursive.errorMessage, "ReadThingCommand: "
        "Since this isn't a true Fildirute file system, first parameter "
        "in a Thing must be \"EquivalentOldThing\", not \"%s\"",
        ParameterName.String ());
      return B_BAD_VALUE;
    }

    if (ParameterName.ICompare ("EquivalentOldThing") == 0)
    {
      if (ParameterValue.ICompare ("Attribute") == 0)
        ThingType = THING_ATTRIBUTE;
      else if (ParameterValue.ICompare ("Directory") == 0)
        ThingType = THING_DIRECTORY;
      else if (ParameterValue.ICompare ("File") == 0)
        ThingType = THING_FILE;
      else
      {
        ParameterValue.Truncate (PATH_MAX); /* Avoid buffer overflow. */
        sprintf (m_Recursive.errorMessage, "ReadThingCommand: "
          "Don't recognize type \"%s\" for the %s setting "
          "(want Attribute, Directory or File), ignoring the rest "
          "of the Thing", ParameterValue.String (), ParameterName.String());
        return READERR_UNKNOWN_FUTURE_THING;
      }
    }
    else if (ParameterName.ICompare ("Path") == 0)
    {
      /* Remove the redundant old volume name from the front of the path. */

      if (m_Recursive.discardedPathPrefixLength > 0)
      {
        /* Does the path start with the volume name?  Invalid if not. */

        if (ParameterValue.Compare (m_Recursive.oldVolumePrefix,
        m_Recursive.discardedPathPrefixLength) != 0)
        {
          ParameterValue.Truncate (PATH_MAX); /* Avoid buffer overflow. */
          sprintf (m_Recursive.errorMessage, "ReadThingCommand: "
            "Path in packed archive is \"%s\", which doesn't "
            "start with the old volume name/directory of \"%s\"",
            ParameterValue.String (), m_Recursive.oldVolumePrefix);
          return B_BAD_VALUE;
        }

        /* Remove the volume name and slash after it, if the slash exists.  Can
        get an empty string for the top level directory itself. */

        ParameterValue.Remove (0 /* start */,
          m_Recursive.discardedPathPrefixLength /* amount */);
        while (ParameterValue.ByteAt (0) == '/')
          ParameterValue.Remove (0 /* start */, 1 /* amount */);
      }

      /* Move the leaf name from the path and into LeafName. */

      StringPntr = (char *) ParameterValue.String ();
      LeafPntr = strrchr (StringPntr, '/');
      if (LeafPntr != NULL)
      {
        LeafName = LeafPntr + 1;
        ParameterValue.Truncate (LeafPntr - StringPntr);
      }
      else /* The whole path is the leaf name. */
      {
        LeafName = ParameterValue;
        ParameterValue.Truncate (0);
      }

      if (DoubleEscapedPathLeaf)
      {
        /* De-escape the leaf name at the end of the path.  It may have odd
        characters, like slashes, in it. */

        SourceLength = LeafName.Length ();
        DestinationLength = SourceLength;
        StringPntr = LeafName.LockBuffer (SourceLength + 1 /* avoid zero */);
        RemoveBackslashEscapeCodes (StringPntr, &SourceLength,
          StringPntr, &DestinationLength, NULL);
        LeafName.UnlockBuffer (DestinationLength);
      }

      if (LeafName.Length () >= NAME_MAX)
      {
        LeafName.Truncate (PATH_MAX);
        sprintf (m_Recursive.errorMessage, "ReadThingCommand: "
          "Leaf name of \"%s\" is too long", LeafName.String ());
        return B_NAME_TOO_LONG;
      }

      /* Put the leaf back in the relative path, if this is a file or a
      directory.  Attributes in BeOS R5 have a separate name space so the leaf
      stays separate for them. */

      if (ThingType != THING_ATTRIBUTE)
      {
        if (ParameterValue.Length () > 0 &&
        ParameterValue.ByteAt (ParameterValue.Length () - 1) != '/')
          ParameterValue.Append ('/', 1);
        ParameterValue.Append (LeafName);
        LeafName.Truncate (0);
      }

      /* Combine the top level directory and the thing's relative path to get
      the absolute path to the BeOS entity we will be fiddling with (file or
      directory itself, or the thing (file, dir, link) that the attribute will
      be added to). */

      TopLevelPathString = (char *) m_TopLevelDirectoryName.Path ();
      TopLevelPathStringLength = strlen (TopLevelPathString);

      ErrorCode = m_Recursive.currentPathName.SetTo (TopLevelPathString,
        ParameterValue.String (), true /* normalize */);
      if (ErrorCode != B_OK)
      {
        ParameterValue.Truncate (PATH_MAX); /* Avoid buffer overflow. */
        sprintf (m_Recursive.errorMessage, "ReadThingCommand: "
          "Combining the top level directory with path of \"%s\" "
          "from the archive didn't work (maybe a subdirectory doesn't exist?)",
          ParameterValue.String ());
        return ErrorCode;
      }

      /* Verify that symbolic links didn't take the BeOS entity (a directory
      or file) out of the top level directory - avoids cracking tricks. */

      StringPntr = (char *) m_Recursive.currentPathName.Path ();
      if (strncmp (TopLevelPathString, StringPntr,
      TopLevelPathStringLength) != 0 ||
      ((int) strlen (StringPntr) > TopLevelPathStringLength &&
      StringPntr[TopLevelPathStringLength] != '/'))
      {
        ParameterValue.Truncate (PATH_MAX); /* Avoid buffer overflow. */
        sprintf (m_Recursive.errorMessage, "ReadThingCommand: "
          "Security cracking attempt detected: path for \"%s\" "
          "uses symbolic links or other tricks to get out of the top level "
          "directory", ParameterValue.String ());
        return B_BAD_VALUE;
      }

      ErrorCode = StatableEntry.SetTo (StringPntr, false /* traverse */);
      if (ErrorCode != B_OK)
      {
        sprintf (m_Recursive.errorMessage, "ReadThingCommand: "
          "Unable to set a BEntry to \"%s\"", StringPntr);
        return ErrorCode;
      }

      PathNameForUser = StringPntr;
      if (LeafName.Length () > 0)
      {
        PathNameForUser.Append ("/");
        PathNameForUser.Append (LeafName);
      }
      if (PathNameForUser.Length () > PATH_MAX + NAME_MAX)
      {
        PathNameForUser.Truncate (
          PATH_MAX + NAME_MAX - strlen (B_UTF8_ELLIPSIS));
        PathNameForUser.Append (B_UTF8_ELLIPSIS);
      }
      UpdateNowProcessingPath (PathNameForUser.String ());
    }
    else if (ParameterName.ICompare ("Data") == 0)
    {
      if (!CreateCommand)
      {
        sprintf (m_Recursive.errorMessage, "ReadThingCommand: "
          "Got an unallowed Data= line in a Modify Thing command, "
          "for thing \"%s\"", PathNameForUser.String ());
        return B_BAD_VALUE;
      }

      /* Data is coming, this is our cue to actually create the Thing.  We need
      to know what variety of Thing it is, and the path name (actually BEntry
      StatableEntry to cut down on path traversal costs). */

      if (ThingType == THING_MAX || StatableEntry.InitCheck () != B_OK)
      {
        sprintf (m_Recursive.errorMessage, "ReadThingCommand: "
          "Haven't got the path or Thing type yet, while processing \"%s\".  "
          "They need to be known before the Data= line",
          PathNameForUser.String ());
        return B_BAD_VALUE;
      }

      switch (ThingType)
      {
        case THING_ATTRIBUTE:
          ErrorCode = Node.SetTo (&StatableEntry);
          if (ErrorCode != B_OK)
          {
            sprintf (m_Recursive.errorMessage, "ReadThingCommand: "
              "Can't open \"%s\" for setting attributes",
              m_Recursive.currentPathName.Path ());
            return ErrorCode;
          }

          /* Assume attribute data is small enough to fit in a BString. */

          ErrorCode = ReadAndUnescapeValue (ParameterValue);
          if (ErrorCode != B_OK)
          {
            sprintf (m_Recursive.errorMessage, "ReadThingCommand: "
              "Problems reading attribute data for \"%s\"",
              PathNameForUser.String ());
            return ErrorCode;
          }

          ActualDataSize = ParameterValue.Length ();
          ErrorCode = Node.WriteAttr (LeafName.String (), AttributeTypeCode,
            0, ParameterValue.String (), ActualDataSize);
          if (ErrorCode != ActualDataSize)
          {
            sprintf (m_Recursive.errorMessage, "ReadThingCommand: "
              "When writing attribute \"%s\", only wrote %ld bytes instead "
              "of %Ld", PathNameForUser.String (),  ErrorCode, ActualDataSize);
            if (ErrorCode >= 0)
              ErrorCode = -1;
            return ErrorCode;
          }
          Node.Unset ();
          break;

        case THING_DIRECTORY:
          ErrorCode = Directory.SetTo (&StatableEntry);
          if (ErrorCode == B_ENTRY_NOT_FOUND)
          {
            if (mkdir (m_Recursive.currentPathName.Path (), 0700))
              ErrorCode = errno;
            else
              ErrorCode = Directory.SetTo (&StatableEntry);
          }
          if (ErrorCode != B_OK)
          {
            sprintf (m_Recursive.errorMessage, "ReadThingCommand: "
              "Can't open or create directory \"%s\"",
              m_Recursive.currentPathName.Path ());
            return ErrorCode;
          }
          /* Don't actually save the Data= stuff for directories, since this
          isn't a fildirute file system.  Don't even need to read it,
          ReadNextLine will skip over the unused data. */
          ActualDataSize = 0;
          Directory.Unset ();
          break;

        case THING_FILE:
          ErrorCode = File.SetTo (&StatableEntry,
            B_WRITE_ONLY | B_CREATE_FILE | B_FAIL_IF_EXISTS);
          if (ErrorCode != B_OK)
          {
            sprintf (m_Recursive.errorMessage, "ReadThingCommand: "
              "Can't create file \"%s\"", m_Recursive.currentPathName.Path ());
            return ErrorCode;
          }

          /* Read some data, remove escape codes, write it to the file.  Repeat
          until the end of the Data= line is encountered. */

          EscapeState = 0;
          ActualDataSize = 0;
          while (true)
          {
            /* Need enough text in the buffers for the end of a line marker. */

            if (m_Recursive.inputLengthExported -
            m_Recursive.inputUnusedDataStart < 10)
            {
              ErrorCode = ReadFillInputBuffer ();
              if (ErrorCode != B_OK)
                return ErrorCode;
              if (m_Recursive.inputUnusedDataStart >=
              m_Recursive.inputLengthExported)
              {
                sprintf (m_Recursive.errorMessage, "ReadThingCommand: "
                  "Reached end of file, so can't find the end of the Data= "
                  "line for \"%s\"", PathNameForUser.String ());
                return READERR_END_OF_FILE;
              }
            }
            EndPntr =
              m_Recursive.inputBuffer + m_Recursive.inputLengthExported;
            StringPntr =
              m_Recursive.inputBuffer + m_Recursive.inputUnusedDataStart;

            /* Advance StringPntr to the end of the line (or end of the buffer
            if the line is longer than the buffer).  End of line is a LF
            character, not preceeded by a backslash.  Note that when the buffer
            is refilled, it keeps the last two read characters before the new
            data. */

            while (StringPntr < EndPntr)
            {
              if (StringPntr[0] == '\n' && StringPntr[-1] != '\\')
                break;
              StringPntr++;
            }

            /* Remove the escape codes and put the result in outputBuffer. */

            SourceLength = (StringPntr - m_Recursive.inputBuffer) -
              m_Recursive.inputUnusedDataStart;
            DestinationLength = INPUT_OUTPUT_BUFFER_SIZE;
            RemoveBackslashEscapeCodes (
              m_Recursive.inputBuffer + m_Recursive.inputUnusedDataStart,
              &SourceLength, m_Recursive.outputBuffer, &DestinationLength,
              &EscapeState);
            if (DestinationLength > 0)
            {
              ErrorCode =
                File.Write (m_Recursive.outputBuffer, DestinationLength);
              if (ErrorCode != DestinationLength)
              {
                sprintf (m_Recursive.errorMessage, "ReadThingCommand: "
                  "Error while writing file contents \"%s\".  Only wrote "
                  "%ld of %d bytes", PathNameForUser.String (),
                  ErrorCode, DestinationLength);
                return ErrorCode;
              }
              ActualDataSize += DestinationLength;
            }

            m_Recursive.inputUnusedDataStart =
              StringPntr - m_Recursive.inputBuffer;

            if (StringPntr < EndPntr)
            {
              m_Recursive.inputUnusedDataStart++; /* Skip the final LF. */
              break; /* Finished reading data. */
            }
          }
          File.Unset ();
          break;

        default:
          strcpy (m_Recursive.errorMessage, "ReadThingCommand: "
            "Internal bug #1");
          return -1;
      }
    }
    else if (ParameterName.ICompare ("PathDoubleEscapedLeaf") == 0)
    {
      if (ParameterValue.ICompare ("Yes") == 0 ||
      ParameterValue.ICompare ("True") == 0 ||
      ParameterValue.ICompare ("1") == 0)
        DoubleEscapedPathLeaf = true;
      else
        DoubleEscapedPathLeaf = false;
    }
    else if (ParameterName.ICompare ("PrimitiveType") == 0)
    {
      AttributeTypeCode = 0;
      while (ParameterValue.Length () > 0)
      {
        AttributeTypeCode <<= 8;
        AttributeTypeCode |= ParameterValue.ByteAt (0);
        ParameterValue.Remove (0 /* start */, 1 /* count */);
      }
    }
    else if (ReadMetaData (ParameterName.String (), ParameterValue.String (),
    &StatableEntry, (ThingType == THING_ATTRIBUTE), ActualDataSize,
    PathNameForUser.String (), ErrorCode))
    {
      if (ErrorCode != B_OK)
        return ErrorCode;
    }
    else /* Unknown parameter, ignore it. */
    {
      /* Print a warning to stdout, not a full fledged error. */
      ParameterValue.Truncate (256);
      printf ("ReadThingCommand: "
        "Ignoring unknown parameter \"%s\" with value \"%s\" for \"%s\".\n",
        ParameterName.String (), ParameterValue.String (),
        PathNameForUser.String ());
    }
  }

  return B_OK;
}


/* Have already read the "{ Create Volume" line.  Read the information about
the volume and save it in the global variables (volume name, size, etc).  The
data can then be just displayed to the user (so they know what's in a file) or
actual reading of the volume can follow.  If OverrideUserSettings is TRUE then
settings specified by the user will be replaced by ones from the volume header,
this includes m_TopLevelDirectoryName, m_MaxRAMDiskSize, m_Notes. */

status_t AFSPApp::ReadVolumeCommand (bool OverrideUserSettings)
{
  status_t ErrorCode;
  BString  ParameterName;
  BString  ParameterValue;
  char    *StringPntr;
  BString  TempBString;

  UpdateNowProcessingPath ("Processing Create Volume command...");

  while (true)
  {
    ErrorCode = ReadNextParameterName (ParameterName, true /* skip nested */);
    if (ErrorCode == READERR_END_OF_NESTING)
      break;
    if (ErrorCode != B_OK)
    {
      sprintf (m_Recursive.errorMessage, "ReadVolumeCommand: "
        "Didn't get to end of the volume information block in file \"%s\"",
        m_PackedFileName.Path ());
      return ErrorCode;
    }

    ErrorCode = ReadAndUnescapeValue (ParameterValue);
    if (ErrorCode != B_OK)
    {
      ParameterName.Truncate (PATH_MAX); /* Avoid buffer overflow. */
      sprintf (m_Recursive.errorMessage, "ReadVolumeCommand: "
        "Didn't get the value for parameter \"%s\" in file \"%s\"",
        ParameterName.String(), m_PackedFileName.Path ());
      return ErrorCode;
    }

    if (ParameterName.ICompare ("Name") == 0)
    {
      if (OverrideUserSettings)
      {
        /* Make the volume name into a nice safe root directory name. */

        TempBString = ParameterValue;
        TempBString.ReplaceAll ('/', '_');
        TempBString.Prepend ("/");
        m_TopLevelDirectoryName.SetTo (TempBString.String ());
      }

      /* Save the volume name so that it can be removed from path names. */

      free (m_Recursive.oldVolumePrefix);
      m_Recursive.oldVolumePrefix =
        (char *) malloc (ParameterValue.Length () + 1);
      if (m_Recursive.oldVolumePrefix == NULL)
      {
        strcpy (m_Recursive.errorMessage, "ReadVolumeCommand: "
          "Unable to allocate volume name string");
        return ENOMEM;
      }
      strcpy (m_Recursive.oldVolumePrefix, ParameterValue.String ());
      StringPntr = strchr (m_Recursive.oldVolumePrefix, '/');
      if (StringPntr != NULL)
        *StringPntr = 0;
      m_Recursive.discardedPathPrefixLength =
        strlen (m_Recursive.oldVolumePrefix);
    }
    else if (ParameterName.ICompare ("MaxVolumeSize") == 0)
    {
      if (OverrideUserSettings)
        m_MaxRAMDiskSize = atol (ParameterValue.String ());
    }
    else if (ParameterName.ICompare ("Notes") == 0)
    {
      if (OverrideUserSettings)
        m_Notes = ParameterValue;
    }
    else if (ParameterName.ICompare ("RoughSizeEstimateNoAttributes") == 0)
    {
      m_NowProcessingEstimatedFinalTotal =
        strtoull (ParameterValue.String (), &StringPntr, 0);
    }
  }

  return B_OK;
}


void AFSPApp::ReadyToRun ()
{
  AFSPSettingsWindow *SettingsWindowPntr;
  float                JunkFloat;
  BButton             *TempButtonPntr;
  BCheckBox           *TempCheckBoxPntr;
  font_height          TempFontHeight;
  BRadioButton        *TempRadioButtonPntr;
  BRect                TempRect;
  const char          *TempString = "Testing My Things";
  BStringView         *TempStringViewPntr;
  BTextControl        *TempTextPntr;
  BWindow             *TempWindowPntr;

  /* This batch of code gets some measurements which will be used for laying
  out controls and other GUI elements.  Set the spacing between buttons and
  other controls to the width of the letter "M" in the user's desired font. */

 g_MarginBetweenControls = (int) be_plain_font->StringWidth ("M");

  /* Also find out how much space a line of text uses. */

  be_plain_font->GetHeight (&TempFontHeight);
  g_LineOfTextHeight = ceilf (
    TempFontHeight.ascent + TempFontHeight.descent + TempFontHeight.leading);

  /* Start finding out the height of various user interface gadgets, which can
  vary based on the current font size.  Make a temporary gadget, which is
  attached to our window, then resize it to its prefered size so that it
  accomodates the font size and other frills it needs. */

  TempWindowPntr = new BWindow (BRect (10, 20, 200, 200), "Temporary Window",
    B_DOCUMENT_WINDOW, B_NO_WORKSPACE_ACTIVATION | B_ASYNCHRONOUS_CONTROLS);
  if (TempWindowPntr == NULL)
  {
    DisplayErrorMessage ("Unable to create temporary window for finding "
      "sizes of controls.");
    g_QuitCountdown = 0;
    return;
  }

  TempRect = TempWindowPntr->Bounds ();

  /* Find the height of a single line of text in a BStringView. */

  TempStringViewPntr = new BStringView (TempRect, TempString, TempString);
  if (TempStringViewPntr != NULL)
  {
    TempWindowPntr->Lock ();
    TempWindowPntr->AddChild (TempStringViewPntr);
    TempStringViewPntr->GetPreferredSize (&JunkFloat, &g_StringViewHeight);
    TempWindowPntr->RemoveChild (TempStringViewPntr);
    TempWindowPntr->Unlock ();
    delete TempStringViewPntr;
  }

  /* Find the height of a button, which seems to be larger than a text
  control and can make life difficult.  Make a temporary button, which
  is attached to our window so that it resizes to accomodate the font size. */

  TempButtonPntr = new BButton (TempRect, TempString, TempString, NULL);
  if (TempButtonPntr != NULL)
  {
    TempWindowPntr->Lock ();
    TempWindowPntr->AddChild (TempButtonPntr);
    TempButtonPntr->GetPreferredSize (&JunkFloat, &g_ButtonHeight);
    TempWindowPntr->RemoveChild (TempButtonPntr);
    TempWindowPntr->Unlock ();
    delete TempButtonPntr;
  }

  /* Find the height of a text box. */

  TempTextPntr = new BTextControl (TempRect, TempString, NULL /* label */,
    TempString, NULL);
  if (TempTextPntr != NULL)
  {
    TempWindowPntr->Lock ();
    TempWindowPntr->AddChild (TempTextPntr);
    TempTextPntr->GetPreferredSize (&JunkFloat, &g_TextBoxHeight);
    TempWindowPntr->RemoveChild (TempTextPntr);
    TempWindowPntr->Unlock ();
    delete TempTextPntr;
  }

  /* Find the height of a checkbox control. */

  TempCheckBoxPntr = new BCheckBox (TempRect, TempString, TempString, NULL);
  if (TempCheckBoxPntr != NULL)
  {
    TempWindowPntr->Lock ();
    TempWindowPntr->AddChild (TempCheckBoxPntr);
    TempCheckBoxPntr->GetPreferredSize (&JunkFloat, &g_CheckBoxHeight);
    TempWindowPntr->RemoveChild (TempCheckBoxPntr);
    TempWindowPntr->Unlock ();
    delete TempCheckBoxPntr;
  }

  /* Find the height of a radio button control. */

  TempRadioButtonPntr =
    new BRadioButton (TempRect, TempString, TempString, NULL);
  if (TempRadioButtonPntr != NULL)
  {
    TempWindowPntr->Lock ();
    TempWindowPntr->AddChild (TempRadioButtonPntr);
    TempRadioButtonPntr->GetPreferredSize (&JunkFloat, &g_RadioButtonHeight);
    TempWindowPntr->RemoveChild (TempRadioButtonPntr);
    TempWindowPntr->Unlock ();
    delete TempRadioButtonPntr;
  }

  /* The height of a row of controls, tall enough to fit in any of the controls
  we use, plus a bit of margin. */

  g_RowHeight = g_ButtonHeight;
  if (g_TextBoxHeight > g_RowHeight)
    g_RowHeight = g_TextBoxHeight;
  g_RowHeight = ceilf (g_RowHeight * 1.1);

  TempWindowPntr->Lock ();
  TempWindowPntr->Quit ();

  if (g_CommandLineMode)
    g_QuitCountdown = 3; /* Quit as soon as queued up commands done. */
  else /* GUI mode, make a window. */
  {
    SettingsWindowPntr = new AFSPSettingsWindow ();
    if (SettingsWindowPntr == NULL)
    {
      DisplayErrorMessage ("Unable to create window.");
      g_QuitCountdown = 0;
    }
    else
      SettingsWindowPntr->Show (); /* Starts the window's message loop. */
  }

  SetPulseRate (500000);
}


status_t AFSPApp::RecursiveTotalSizeEstimate (BDirectory &CurrentNode)
{
  BEntry    ChildEntry;
  status_t  ErrorCode;

  m_NowProcessingEstimatedFinalTotal++; /* Add one for a directory. */

  CurrentNode.Rewind ();
  while ((ErrorCode = CurrentNode.GetNextEntry (&ChildEntry,
  false /* traverse */)) == B_OK)
  {
    if (m_CancelRequestedByGUI)
    {
      sprintf (m_Recursive.errorMessage,
        "Initial file scan cancelled by user, was listing \"%s\" contents",
        m_Recursive.currentPathName.Path ());
      return B_INTERRUPTED;
    }

    ChildEntry.GetName (m_Recursive.errorMessage);
    m_Recursive.currentPathName.Append (m_Recursive.errorMessage);

    ErrorCode = ChildEntry.GetStat (&m_Recursive.fileStatus);
    if (ErrorCode != B_OK)
    {
      sprintf (m_Recursive.errorMessage, "RecursiveTotalSizeEstimate: "
        "Unable to get status of entry \"%s\"",
        m_Recursive.currentPathName.Path ());
      return ErrorCode;
    }

    if (S_ISDIR (m_Recursive.fileStatus.st_mode))
    {
      BDirectory ChildDir;
      ErrorCode = ChildDir.SetTo (&ChildEntry);
      if (ErrorCode != B_OK)
      {
        sprintf (m_Recursive.errorMessage, "RecursiveTotalSizeEstimate: "
          "Unable to open directory \"%s\"",
          m_Recursive.currentPathName.Path ());
        return ErrorCode;
      }
      ErrorCode = RecursiveTotalSizeEstimate (ChildDir);
      if (ErrorCode != B_OK) return ErrorCode;
    }
    else if (S_ISREG (m_Recursive.fileStatus.st_mode))
      m_NowProcessingEstimatedFinalTotal +=
      m_Recursive.fileStatus.st_size; /* File size is used for files. */
    else if (S_ISLNK (m_Recursive.fileStatus.st_mode))
    {
      if (m_Recursive.fileStatus.st_size == 0 ||
      m_Recursive.fileStatus.st_size == 1)
        /* Links count as one.  However we don't count automatically generated
        parent directory links (they have size 2). */
        m_NowProcessingEstimatedFinalTotal++;
    }
    /* else Mysterious other things, like block devices, don't count. */

    m_Recursive.currentPathName.GetParent (&m_Recursive.currentPathName);
  }

  if (ErrorCode != B_ENTRY_NOT_FOUND)
  {
    sprintf (m_Recursive.errorMessage, "RecursiveTotalSizeEstimate: "
      "Had problems reading directory \"%s\" contents",
      m_Recursive.currentPathName.Path ());
    return ErrorCode;
  }

  return B_OK;
}


/* The user has tried to open a file or several files with this application,
via Tracker's open-with menu item.  If it is a pack type file, then change the
pack file name to it and restore it.  If it is a directory, make it the top
level directory.  The command is sent to an intermediary thread where it is
asynchronously converted into a scripting message(s) that are sent back to this
BApplication.  The intermediary is needed since we can't recursively execute
scripting messages while processing a message (this RefsReceived one). */

void AFSPApp::RefsReceived (BMessage *MessagePntr)
{
  if (g_CommanderLooperPntr != NULL)
    g_CommanderLooperPntr->CommandReferences (MessagePntr);
}


/* A scripting command is looking for something to execute it.  See if it is
targetted at ourself. */

BHandler * AFSPApp::ResolveSpecifier (
  BMessage *MessagePntr,
  int32 Index,
  BMessage *SpecifierMsgPntr,
  int32 SpecificationKind,
  const char *PropertyPntr)
{
  int i;

  /* See if it is one of our commands. */

  if (SpecificationKind == B_DIRECT_SPECIFIER)
  {
    for (i = PN_MAX - 1; i >= 0; i--)
    {
      if (strcasecmp (PropertyPntr, g_PropertyNames [i]) == 0)
        return this; /* Found it!  Return the Handler (which is us). */
    }
  }

  /* Handle an unrecognized scripting command, let the parent figure it out. */

  return BApplication::ResolveSpecifier (
    MessagePntr, Index, SpecifierMsgPntr, SpecificationKind, PropertyPntr);
}


/* Copy as much of the given string that fits into the buffer which gets
displayed by the progress window as the path that's being worked on.  Not that
it has to be a path. */

void AFSPApp::UpdateNowProcessingPath (const char *StringPntr)
{
  char *EndPntr;
  char *WritePntr;

  WritePntr = m_NowProcessingPath;
  EndPntr = m_NowProcessingPath + (sizeof (m_NowProcessingPath) - 1);

  while (*StringPntr != 0 && WritePntr != EndPntr)
    *WritePntr++ = *StringPntr++;
  *WritePntr = 0;
}



/******************************************************************************
 * Implementation of the CommanderLooper class, constructor, destructor and the
 * rest of the member functions in mostly alphabetical order.
 */

CommanderLooper::CommanderLooper ()
: BLooper ("CommanderLooper", B_NORMAL_PRIORITY),
  m_IsBusy (false)
{
}


CommanderLooper::~CommanderLooper ()
{
  g_CommanderLooperPntr = NULL;
  delete g_CommanderMessenger;
  g_CommanderMessenger = NULL;
}


/* Process some command line arguments.  Basically just send a message to this
looper itself to do the work later.  That way the caller can continue doing
whatever they're doing, particularly if it's the BApplication. */

void CommanderLooper::CommandArguments (int argc, char **argv)
{
  int      i;
  BMessage InternalMessage;

  InternalMessage.what = MSG_COMMAND_ARGUMENTS;
  for (i = 0; i < argc; i++)
    InternalMessage.AddString ("arg", argv[i]);

  PostMessage (&InternalMessage);
}


/* Copy the refs out of the given message and stuff them into an internal
message to ourself (so that the original message can be returned to the caller,
and if it is Tracker, it can close the file handles it has open). */

void CommanderLooper::CommandReferences (
  BMessage *MessagePntr)
{
  entry_ref EntryRef;
  int       i;
  BMessage  InternalMessage;

  InternalMessage.what = MSG_COMMAND_FILE_REFS;
  for (i = 0; MessagePntr->FindRef ("refs", i, &EntryRef) == B_OK; i++)
    InternalMessage.AddRef ("refs", &EntryRef);

  PostMessage (&InternalMessage);
}


/* This function is called by other threads to see if the CommanderLooper is
busy working on something. */

bool CommanderLooper::IsBusy ()
{
  if (m_IsBusy)
    return true;

  if (IsLocked () || !MessageQueue()->IsEmpty ())
    return true;

  return false;
}


void CommanderLooper::MessageReceived (BMessage *MessagePntr)
{
  m_IsBusy = true;

  if (MessagePntr->what == MSG_COMMAND_ARGUMENTS)
    ProcessArgs (MessagePntr);
  else if (MessagePntr->what == MSG_COMMAND_FILE_REFS)
    ProcessRefs (MessagePntr);
  else
    BLooper::MessageReceived (MessagePntr);

  m_IsBusy = false;
}


/* Process the command line by converting it into a series of scripting
messages (possibly thousands) and sent them to the BApplication synchronously
(so we can print the result). */

void CommanderLooper::ProcessArgs (BMessage *MessagePntr)
{
  int32                 argc = 0;
  const char          **argv = NULL;
  int                   ArgumentIndex;
  uint32                CommandCode;
  const char           *CommandWord;
  status_t              ErrorCode;
  const char           *ErrorTitle = "ProcessArgs";
  char                 *EndPntr;
  int32                 i;
  BMessage              ReplyMessage;
  BMessage              ScriptMessage;
  struct property_info *PropInfoPntr;
  const char           *PropertyName;
  bool                  TempBool;
  float                 TempFloat;
  int32                 TempInt32;
  const char           *TempStringPntr;
  type_code             TypeCode;
  const char           *ValuePntr;

  /* Get the argument count and pointers to arguments out of the message and
  into our argc and argv. */

  ErrorCode = MessagePntr->GetInfo ("arg", &TypeCode, &argc);
  if (ErrorCode != B_OK || TypeCode != B_STRING_TYPE)
  {
    DisplayErrorMessage ("Unable to find argument strings in message",
      ErrorCode, ErrorTitle);
    goto ErrorExit;
  }

  if (argc < 2)
  {
    cerr << PrintUsage;
    DisplayErrorMessage ("You need to specify a command word, like GET, SET "
      "and so on followed by a property, like PackedFile, and maybe "
      "followed by a value of some sort", -1, ErrorTitle);
    goto ErrorExit;
  }

  argv = (const char **) malloc (sizeof (char *) * argc);
  if (argv == NULL)
  {
    DisplayErrorMessage ("Out of memory when allocating argv array",
      ENOMEM, ErrorTitle);
    goto ErrorExit;
  }

  for (i = 0; i < argc; i++)
  {
    if ((ErrorCode = MessagePntr->FindString ("arg", i, &argv[i])) != B_OK)
    {
      DisplayErrorMessage ("Unable to find argument in the BMessage",
        ErrorCode, ErrorTitle);
      goto ErrorExit;
    }
  }

  CommandWord = argv[1];

  /* Special case for the Quit command since it isn't a scripting command. */

  if (strcasecmp (CommandWord, "quit") == 0)
  {
    g_QuitCountdown = 3;
    goto ErrorExit;
  }

  /* Find the corresponding scripting command. */

  if (strcasecmp (CommandWord, "set") == 0)
    CommandCode = B_SET_PROPERTY;
  else if (strcasecmp (CommandWord, "get") == 0)
    CommandCode = B_GET_PROPERTY;
  else if (strcasecmp (CommandWord, "count") == 0)
    CommandCode = B_COUNT_PROPERTIES;
  else if (strcasecmp (CommandWord, "create") == 0)
    CommandCode = B_CREATE_PROPERTY;
  else if (strcasecmp (CommandWord, "delete") == 0)
    CommandCode = B_DELETE_PROPERTY;
  else
    CommandCode = B_EXECUTE_PROPERTY;

  if (CommandCode == B_EXECUTE_PROPERTY)
  {
    PropertyName = CommandWord;
    ArgumentIndex = 2; /* Arguments to the command start at this index. */
  }
  else
  {
    if (CommandCode == B_SET_PROPERTY)
    {
      /* SET commands require at least one argument value. */
      if (argc < 4)
      {
        cerr << PrintUsage;
        DisplayErrorMessage ("SET commands require at least one "
          "argument value after the property name", -1, ErrorTitle);
        goto ErrorExit;
      }
    }
    else
      if (argc < 3)
      {
        cerr << PrintUsage;
        DisplayErrorMessage ("You need to specify a property to act on",
          -1, ErrorTitle);
        goto ErrorExit;
      }
    PropertyName = argv[2];
    ArgumentIndex = 3;
  }

  /* See if it is one of our commands. */

  for (PropInfoPntr = g_ScriptingPropertyList + 0; true; PropInfoPntr++)
  {
    if (PropInfoPntr->name == 0)
    {
      cerr << PrintUsage;
      DisplayErrorMessage ("The property specified isn't known or "
        "doesn't support the requested action (usually means it is an "
        "unknown command)", -1, ErrorTitle);
      goto ErrorExit; /* Unrecognized command. */
    }

    if (PropInfoPntr->commands[0] == CommandCode &&
    strcasecmp (PropertyName, PropInfoPntr->name) == 0)
      break;
  }

  /* Make the equivalent command message.  For commands with multiple
  arguments, repeat the message for each single argument and just change the
  data portion for each extra argument.  Send the command and wait for a reply,
  which we'll print out. */

  ScriptMessage.MakeEmpty ();
  ScriptMessage.what = CommandCode;
  ScriptMessage.AddSpecifier (PropertyName);
  while (true)
  {
    if (ArgumentIndex < argc) /* If there are arguments to be added. */
    {
      ValuePntr = argv[ArgumentIndex];

      /* Convert the value into the likely kind of data. */

      if (strcasecmp (ValuePntr, "yes") == 0 ||
      strcasecmp (ValuePntr, "true") == 0)
        ScriptMessage.AddBool (g_DataName, true);
      else if (strcasecmp (ValuePntr, "no") == 0 ||
      strcasecmp (ValuePntr, "false") == 0)
        ScriptMessage.AddBool (g_DataName, false);
      else
      {
        /* See if it is a number. */
        i = strtol (ValuePntr, &EndPntr, 0);
        if (*EndPntr == 0)
          ScriptMessage.AddInt32 (g_DataName, i);
        else /* Nope, it's just a string. */
          ScriptMessage.AddString (g_DataName, ValuePntr);
      }
    }

    ErrorCode = be_app_messenger.SendMessage (&ScriptMessage, &ReplyMessage);
    if (ErrorCode != B_OK)
    {
      DisplayErrorMessage ("Unable to send scripting command",
        ErrorCode, ErrorTitle);
      goto ErrorExit;
    }

    /* Print the reply to the scripting command.  Even in server mode.  To
    standard output. */

    if (ReplyMessage.FindString ("CommandText", &TempStringPntr) == B_OK)
    {
      TempInt32 = -1;
      if (ReplyMessage.FindInt32 ("error", &TempInt32) == B_OK &&
      TempInt32 == B_OK)
      {
        /* It's a successful reply to one of our scripting messages.  Print out
        the returned values code for command line users to see. */

        cout << "Result of command to " << TempStringPntr << " is:\t";
        if (ReplyMessage.FindString (g_ResultName, &TempStringPntr) == B_OK)
          cout << "\"" << TempStringPntr << "\"";
        else if (ReplyMessage.FindInt32 (g_ResultName, &TempInt32) == B_OK)
          cout << TempInt32;
        else if (ReplyMessage.FindFloat (g_ResultName, &TempFloat) == B_OK)
          cout << TempFloat;
        else if (ReplyMessage.FindBool (g_ResultName, &TempBool) == B_OK)
          cout << (TempBool ? "true" : "false");
        else
          cout << "just plain success";
        cout << endl;
      }
      else /* An error reply, print out the error, even in server mode. */
      {
        cout << "Failure of command " << TempStringPntr << ", error ";
        cout << TempInt32 << " (" << strerror (TempInt32) << ")";
        if (ReplyMessage.FindString ("message", &TempStringPntr) == B_OK)
          cout << ", message: " << TempStringPntr;
        cout << "." << endl;
      }
    }

    /* Advance to the next argument and its scripting message. */

    ScriptMessage.RemoveName (g_DataName);
    if (++ArgumentIndex >= argc)
      break;
  }

ErrorExit:
  free (argv);
}


/* Given a bunch of references to files or directories, open them by sending
the appropriate scripting messages to the application. */

void CommanderLooper::ProcessRefs (BMessage *MessagePntr)
{
  BEntry      Entry;
  entry_ref   EntryRef;
  status_t    ErrorCode;
  const char *ErrorTitle = "CommanderLooper::ProcessRefs";
  int         i;
  int32       NumberOfRefs = 0;
  BPath       Path;
  BMessage    ReplyMessage;
  BMessage    ScriptingMessage;
  BFile       TempFile;
  int32       TempInt32;
  char        TempString [PATH_MAX + 1024];
  type_code   TypeCode;

  ErrorCode = MessagePntr->GetInfo ("refs", &TypeCode, &NumberOfRefs);
  if (ErrorCode != B_OK || TypeCode != B_REF_TYPE || NumberOfRefs <= 0)
  {
    DisplayErrorMessage ("Unable to get refs from the message",
      ErrorCode, ErrorTitle);
    return;
  }

  for (i = 0; MessagePntr->FindRef ("refs", i, &EntryRef) == B_OK; i++)
  {
    ScriptingMessage.MakeEmpty ();
    ScriptingMessage.what = 0; /* Haven't figured out what to do yet. */

    /* See if the entry is a valid file or directory or other thing. */

    ErrorCode = Entry.SetTo (&EntryRef, true /* traverse symbolic links */);
    if (ErrorCode != B_OK ||
    ((ErrorCode = /* assignment */ B_ENTRY_NOT_FOUND) != 0 /* this pacifies
    mwcc -nwhitehorn */ && !Entry.Exists ()) ||
    ((ErrorCode = Entry.GetPath (&Path)) != B_OK))
    {
      DisplayErrorMessage ("Bad entry reference encountered, will skip it",
        ErrorCode, ErrorTitle);
      continue; /* Bad file reference, try the next one. */
    }

    /* If it's a file, open it. */

    if (Entry.IsFile ())
    {
      ScriptingMessage.what = B_SET_PROPERTY;
      ScriptingMessage.AddSpecifier (g_PropertyNames[PN_PACKED_FILE]);
      ScriptingMessage.AddString (g_DataName, Path.Path ());
      TempFile.Unset ();
    }
    else if (Entry.IsDirectory ())
    {
      ScriptingMessage.what = B_SET_PROPERTY;
      ScriptingMessage.AddSpecifier (g_PropertyNames[PN_TOP_LEVEL_DIRECTORY]);
      ScriptingMessage.AddString (g_DataName, Path.Path ());
      TempFile.Unset ();
    }

    if (ScriptingMessage.what == 0)
    {
      sprintf (TempString, "\"%s\" is not a file or directory, "
        "can't do anything with it", Path.Path ());
      DisplayErrorMessage (TempString, -1, ErrorTitle);
      continue;
    }

    /* Tell the BApplication to do the work, and wait for it to finish.  The
    BApplication will display any error messages for us. */

    ErrorCode =
      be_app_messenger.SendMessage (&ScriptingMessage, &ReplyMessage);
    if (ErrorCode != B_OK)
    {
      DisplayErrorMessage ("Unable to send scripting command",
        ErrorCode, ErrorTitle);
      return;
    }

    /* If there was an error, stop. */

    if (ReplyMessage.FindInt32 ("error", &TempInt32) != B_OK ||
    TempInt32 != B_OK)
      return;
  }
}



/******************************************************************************
 * Implementation of the ProgressArrow class, constructor, destructor and the
 * rest of the member functions in mostly alphabetical order.
 */

ProgressArrow::ProgressArrow (BRect NewBounds, bool PointingUp)
: BView (NewBounds, "ProgressArrow", B_FOLLOW_TOP | B_FOLLOW_LEFT,
    B_WILL_DRAW | B_PULSE_NEEDED),
  m_BackingBitmapPntr (NULL),
  m_BackingView (BRect (0, 0, 100, 100), "BackingView",
    B_FOLLOW_ALL_SIDES, B_FULL_UPDATE_ON_RESIZE),
  m_Red (255),
  m_Green (255),
  m_Blue (255),
  m_DRed (2),
  m_DGreen (3),
  m_DBlue (5),
  m_PercentageTextFont (be_plain_font),
  m_PercentageTextHeightOffset (0.0F),
  m_PointingUp (PointingUp),
  m_Proportion (-1.0F)
{
}


ProgressArrow::~ProgressArrow ()
{
  if (m_BackingBitmapPntr != NULL)
  {
    m_BackingBitmapPntr->RemoveChild (&m_BackingView);
    delete m_BackingBitmapPntr;
    m_BackingBitmapPntr = NULL;
  }
}


void ProgressArrow::AttachedToWindow ()
{
  font_height FontHeightStuff;
  BRect       TempRect;

  if (m_BackingBitmapPntr != NULL)
    return; // Initialisation already done.

  SetViewColor (B_TRANSPARENT_COLOR);

  m_BackingBitmapPntr =
    new BBitmap (Bounds(), B_RGBA32, true /* Accepts subviews */);
  if (!m_BackingBitmapPntr->IsValid ())
  {
    delete m_BackingBitmapPntr;
    m_BackingBitmapPntr = NULL;
  }
  else
  {
    m_BackingBitmapPntr->Lock ();
    TempRect = m_BackingBitmapPntr->Bounds ();
    m_BackingView.MoveTo (TempRect.left, TempRect.top);
    m_BackingView.ResizeTo (TempRect.Width(), TempRect.Height());
    m_BackingBitmapPntr->AddChild (&m_BackingView);
    m_PercentageTextFont.SetSize (m_PercentageTextFont.Size () * 1.5);
    m_PercentageTextFont.GetHeight (&FontHeightStuff);
    m_PercentageTextHeightOffset = FontHeightStuff.ascent -
      (FontHeightStuff.ascent + FontHeightStuff.descent) / 2;
    m_BackingView.SetFont (&m_PercentageTextFont);
    m_BackingView.SetViewColor (B_TRANSPARENT_COLOR);
    m_BackingBitmapPntr->Unlock ();
  }
}


void ProgressArrow::Draw (BRect UpdateRect)
{
  typedef enum ArrowPartEnum {
    AP_BLANK_AT_TOP,
    AP_HEAD_DIAGONALS,
    AP_HEAD_TO_BODY_JUNCTION,
    AP_BODY_VERTICALS,
    AP_BODY_BOTTOM,
    AP_BLANK_AT_BOTTOM,
    AP_MAX
  } ArrowParts;

  float      ActualY;
  ArrowParts ArrowPart;
  float      BodyWidthHalf;
  BRect      BoundsRect (Bounds());
  float      CenterX;
  int        CurBlue, CurGreen, CurRed;
  int        CurDBlue, CurDGreen, CurDRed;
  int        IndexY;
  int        IndexYEnd;
  BPoint     LineEnd;
  BPoint     LineStart;
  float      MaxArrowWidthHalf;
  float      NextArrowPartY;
  float      OneSixtyFourth;
  int        OneSixtyFourthInt;
  BPoint     PercentageTextPoint;
  char       PercentageString [10];
  int        ProgressIndexYColourEnd;
  float      ScanFarX;
  float      ScanNearX;
  BRect      TempRect;

  if (m_BackingBitmapPntr == NULL)
    return; /* Something isn't initialised. */
  m_BackingBitmapPntr->Lock ();

  /* Erase the bitmap to the background colour. */

  m_BackingView.SetDrawingMode (B_OP_COPY);
  m_BackingView.SetHighColor (ui_color (B_PANEL_BACKGROUND_COLOR));
  m_BackingView.FillRect (BoundsRect);

  /* Draw the big arrow.  4/64 of total height is blank at top, 18/64 is used
  for head, 1/64 for the junction, 36/64 for the body, 1/64 for the bottom
  edge, and 4/64 blank at the bottom. */

  CurRed = m_Red;
  CurGreen = m_Green;
  CurBlue = m_Blue;
  CurDRed = 2;
  CurDGreen = 3;
  CurDBlue = 1;

  MaxArrowWidthHalf = ceilf (BoundsRect.Width() / 2 * 0.9);
  BodyWidthHalf = ceilf (BoundsRect.Width() / 2 * 0.5);
  CenterX = BoundsRect.left + BoundsRect.Width() / 2;

  IndexYEnd = (int) BoundsRect.Height();
  OneSixtyFourth = IndexYEnd / 64.0F;
  OneSixtyFourthInt = (int) ceilf (OneSixtyFourth);
  ProgressIndexYColourEnd = (int) (OneSixtyFourth * 4 +
    OneSixtyFourth * (18 + 1 + 36) * (1.0 - m_Proportion));
  ArrowPart = AP_BLANK_AT_TOP;
  NextArrowPartY = 4 * OneSixtyFourth;
  for (IndexY = 0; IndexY <= IndexYEnd; IndexY++)
  {
    /* If we reached the end of the drawing area for one part, as we move down
    the screen, switch to the next part. */

    if (IndexY >= NextArrowPartY)
    {
      ArrowPart = (ArrowParts) ((int) ArrowPart + 1);
      switch (ArrowPart)
      {
        case AP_HEAD_DIAGONALS:
          NextArrowPartY = 22 * OneSixtyFourth; break;
        case AP_HEAD_TO_BODY_JUNCTION:
          NextArrowPartY = IndexY + OneSixtyFourthInt; break;
        case AP_BODY_VERTICALS:
          NextArrowPartY = 59 * OneSixtyFourth; break;
        case AP_BODY_BOTTOM:
          NextArrowPartY = IndexY + OneSixtyFourthInt; break;
        case AP_BLANK_AT_BOTTOM:
          NextArrowPartY = 64 * OneSixtyFourth; break;
        default:
          ArrowPart = AP_BLANK_AT_BOTTOM; break;
      }
    }

    /* Draw the current scan line of the arrow.  First find out the start and
    end of the line.  The far part is the outer part, the near part is the part
    closer to the centre.  A solid black line is drawn from far to near, and
    colour cycling stuff is drawn from near to the center.  The mirror image
    lines are also drawn. */

    switch (ArrowPart)
    {
      case AP_HEAD_DIAGONALS:
        ScanFarX = ceilf ((IndexY - 4 * OneSixtyFourth) /
          (18 * OneSixtyFourth) * MaxArrowWidthHalf);
        ScanNearX = ScanFarX - OneSixtyFourthInt;
        break;

      case AP_HEAD_TO_BODY_JUNCTION:
        ScanFarX = MaxArrowWidthHalf;
        ScanNearX = BodyWidthHalf - OneSixtyFourthInt;
        break;

      case AP_BODY_VERTICALS:
        ScanFarX = BodyWidthHalf;
        ScanNearX = ScanFarX - OneSixtyFourthInt;
        break;

      case AP_BODY_BOTTOM:
        ScanFarX = BodyWidthHalf;
        ScanNearX = 0;
        break;

      default:
        ScanFarX = -1;
        ScanNearX = -1;
        break;
    }

    /* Finally draw the actual scan line. */

    if (m_PointingUp)
      ActualY = IndexY + BoundsRect.top;
    else
      ActualY = IndexYEnd - IndexY + BoundsRect.top;

    if (ScanNearX >= 0)
    {
      /* Draw the outline. */

      m_BackingView.SetHighColor (0, 0, 0);
      LineStart.y = LineEnd.y = ActualY;
      LineStart.x = CenterX - ScanFarX;
      LineEnd.x = CenterX - ScanNearX;
      m_BackingView.StrokeLine (LineStart, LineEnd);

      /* Mirror image of the outline. */

      LineStart.x = CenterX + ScanNearX;
      LineEnd.x = CenterX + ScanFarX;
      m_BackingView.StrokeLine (LineStart, LineEnd);

      /* Filler inside the outline.  Starts animating at the boundary
      corresponding to the percentage complete. */

      if (ScanNearX > 0)
      {
        if (IndexY > ProgressIndexYColourEnd)
        {
          CurRed += CurDRed;
          CurGreen += CurDGreen;
          CurBlue += CurDBlue;

          // Note range 159-255 has 97 possible values, a prime number, so it
          // won't repeat the colours for a long time.  Also, 159-255 is bright
          // enough to keep black text readable.

          if (CurRed < 159 || CurRed > 255)
          {
            CurDRed = -CurDRed;
            CurRed += CurDRed;
          }
          if (CurGreen < 159 || CurGreen > 255)
          {
            CurDGreen = -CurDGreen;
            CurGreen += CurDGreen;
          }
          if (CurBlue < 159 || CurBlue > 255)
          {
            CurDBlue = -CurDBlue;
            CurBlue += CurDBlue;
          }
        }

        m_BackingView.SetHighColor (CurRed, CurGreen, CurBlue);
        LineStart.x = CenterX - ScanNearX;
        LineEnd.x = CenterX + ScanNearX;
        m_BackingView.StrokeLine (LineStart, LineEnd);
      }
    }
  }

  /* Draw the percentage text. */

  if (m_Proportion >= 0)
  {
    sprintf (PercentageString, "%d%%", (int) ceilf (m_Proportion * 100));
    TempRect = Bounds ();
    PercentageTextPoint.x = (TempRect.left + TempRect.right) / 2;
    PercentageTextPoint.y = (TempRect.top + TempRect.bottom) / 2;
    PercentageTextPoint.y = ceilf (
      PercentageTextPoint.y + m_PercentageTextHeightOffset);
    PercentageTextPoint.x = ceilf (PercentageTextPoint.x -
      m_PercentageTextFont.StringWidth (PercentageString) / 2);
    m_BackingView.MovePenTo (PercentageTextPoint);
    m_BackingView.SetDrawingMode (B_OP_OVER);
    m_BackingView.DrawString (PercentageString);
  }

  /* Copy the backing bitmap to the screen. */

  m_BackingView.Sync ();
  m_BackingBitmapPntr->Unlock ();
  MovePenTo (0, 0);
  DrawBitmap (m_BackingBitmapPntr);
}


void ProgressArrow::Pulse ()
{
  m_Red += m_DRed;
  if (m_Red < 159 || m_Red > 255)
  {
    m_DRed = -m_DRed;
    m_Red += m_DRed;
  }

  m_Green += m_DGreen;
  if (m_Green < 159 || m_Green > 255)
  {
    m_DGreen = -m_DGreen;
    m_Green += m_DGreen;
  }

  m_Blue += m_DBlue;
  if (m_Blue < 159 || m_Blue > 255)
  {
    m_DBlue = -m_DBlue;
    m_Blue += m_DBlue;
  }

  Invalidate ();
}


void ProgressArrow::SetProportion (float Proportion)
{
  m_Proportion = Proportion;
  Invalidate ();
}



/******************************************************************************
 * Implementation of the ControlsView class, constructor, destructor and the
 * rest of the member functions in mostly alphabetical order.
 */

ControlsView::ControlsView (BRect NewBounds)
: BView (NewBounds, "ControlsView", B_FOLLOW_TOP_BOTTOM | B_FOLLOW_LEFT_RIGHT,
    B_WILL_DRAW | B_PULSE_NEEDED | B_NAVIGABLE_JUMP),
  m_NowShowing (DM_UNPACKING),
  m_PackFileBrowsePanelPntr (NULL),
  m_TextControlsContainInvalidData (true),
  m_TimeOfLastPoll (0),
  m_TopLevelDirectoryBrowsePanelPntr (NULL)
{
  m_AppPntr = dynamic_cast<AFSPApp *> (be_app);
  if (m_AppPntr == NULL)
    DisplayErrorMessage ("Unable to convert be_app to a class pointer.");
  NullifyChildViewPointers ();
}


ControlsView::~ControlsView ()
{
  /* Delete things which don't automatically get deleted by the destruction of
  this BView. */

  if (m_PackFileBrowsePanelPntr != NULL)
  {
    delete m_PackFileBrowsePanelPntr;
    m_PackFileBrowsePanelPntr = NULL;
  }

  if (m_TopLevelDirectoryBrowsePanelPntr != NULL)
  {
    delete m_TopLevelDirectoryBrowsePanelPntr;
    m_TopLevelDirectoryBrowsePanelPntr = NULL;
  }
}


void ControlsView::AttachedToWindow ()
{
  SetViewColor (ui_color (B_PANEL_BACKGROUND_COLOR));
  SwitchToDisplayMode (m_NowShowing);
}


void ControlsView::BrowseForPackFile ()
{
  if (m_PackFileBrowsePanelPntr == NULL)
  {
    BEntry      DirectoryEntry;
    entry_ref   DirectoryEntryRef;
    BMessage    GetPackFileNameCommand;
    BMessage    GetPackFileNameResult;
    const char *StringPntr = NULL;

    /* Create a new file panel.  First set up the entry ref stuff so that the
    file panel can open to show the initial directory (the current pack file's
    directory).  Note that we have to create it after the window and view are
    up and running, otherwise the BMessenger won't point to a valid
    looper/handler. */

    GetPackFileNameCommand.what = B_GET_PROPERTY;
    GetPackFileNameCommand.AddSpecifier (
      g_PropertyNames[PN_PACKED_FILE]);
    be_app_messenger.SendMessage (&GetPackFileNameCommand,
      &GetPackFileNameResult, 5000000 /* delivery timeout */,
      5000000 /* reply timeout */);
    if (GetPackFileNameResult.FindString (g_ResultName, &StringPntr) != B_OK ||
    DirectoryEntry.SetTo (StringPntr) != B_OK ||
    DirectoryEntry.GetParent (&DirectoryEntry) != B_OK)
      DirectoryEntry.SetTo ("."); /* Default directory if we can't find it. */
    if (DirectoryEntry.GetRef (&DirectoryEntryRef) != B_OK)
    {
      DisplayErrorMessage ("BrowseForPackFile: "
        "Unable to set up the file requestor starting directory.  Sorry.");
      return;
    }

    m_PackFileBrowsePanelPntr = new BFilePanel (
      B_OPEN_PANEL /* mode */,
      &be_app_messenger /* target for event messages */,
      &DirectoryEntryRef /* starting directory */,
      B_FILE_NODE,
      false /* true for multiple selections */,
      NULL /* canned message, default is B_REFS_RECEIVED */,
      NULL /* ref filter */,
      false /* true for modal */,
      true /* true to hide when done */);
  }

  if (m_PackFileBrowsePanelPntr != NULL)
    m_PackFileBrowsePanelPntr->Show ();
}


void ControlsView::BrowseForTopLevelDirectory ()
{
  if (m_TopLevelDirectoryBrowsePanelPntr == NULL)
  {
    BEntry      DirectoryEntry;
    entry_ref   DirectoryEntryRef;
    BMessage    GetTopLevelDirectoryCommand;
    BMessage    GetTopLevelDirectoryResult;
    const char *StringPntr = NULL;

    /* Create a new file panel.  First set up the entry ref stuff so that the
    file panel can open to show the initial directory (the current top level
    directory).  Note that we have to create it after the window and view are
    up and running, otherwise the BMessenger won't point to a valid
    looper/handler. */

    GetTopLevelDirectoryCommand.what = B_GET_PROPERTY;
    GetTopLevelDirectoryCommand.AddSpecifier (
      g_PropertyNames[PN_TOP_LEVEL_DIRECTORY]);
    be_app_messenger.SendMessage (&GetTopLevelDirectoryCommand,
      &GetTopLevelDirectoryResult, 5000000 /* delivery timeout */,
      5000000 /* reply timeout */);
    if (GetTopLevelDirectoryResult.FindString (g_ResultName, &StringPntr)
    != B_OK || DirectoryEntry.SetTo (StringPntr) != B_OK)
      DirectoryEntry.SetTo ("."); /* Default directory if we can't find it. */
    if (DirectoryEntry.GetRef (&DirectoryEntryRef) != B_OK)
    {
      DisplayErrorMessage ("BrowseForTopLevelDirectory: "
        "Unable to set up the file requestor starting directory.  Sorry.");
      return;
    }

    m_TopLevelDirectoryBrowsePanelPntr = new BFilePanel (
      B_OPEN_PANEL /* mode */,
      &be_app_messenger /* target for event messages */,
      &DirectoryEntryRef /* starting directory */,
      B_DIRECTORY_NODE,
      false /* true for multiple selections */,
      NULL /* canned message, default is B_REFS_RECEIVED */,
      NULL /* ref filter */,
      false /* true for modal */,
      true /* true to hide when done */);
  }

  if (m_TopLevelDirectoryBrowsePanelPntr != NULL)
    m_TopLevelDirectoryBrowsePanelPntr->Show ();
}


void ControlsView::MessageReceived (BMessage *MessagePntr)
{
  BMessage CommandMessage;
  bool     TempBool;

  if (MessagePntr->what >= MSG_COMPRESSION_METHOD &&
  MessagePntr->what < MSG_COMPRESSION_METHOD + CM_MAX)
  {
    SubmitCommandString (PN_COMPRESSION_METHOD, B_SET_PROPERTY,
      g_CompressionMethodNames [MessagePntr->what - MSG_COMPRESSION_METHOD]);
    return;
  }

  switch (MessagePntr->what)
  {
    case MSG_MAX_RAM_DISK_SIZE_TEXT:
    case MSG_NOTES_TEXT:
    case MSG_PACK_FILE_NAME_TEXT:
    case MSG_TOP_LEVEL_PATH_TEXT:
    case MSG_ZLIB_LEVEL_TEXT:
      ReadAllTextControls ();
      break;

    case MSG_DISPLAY_MODE_CREATE_FS:
      if (m_NowShowing != DM_CREATE_FILE_SYSTEM)
      {
        ReadAllTextControls ();
        SwitchToDisplayMode (DM_CREATE_FILE_SYSTEM);
      }
      break;

    case MSG_DISPLAY_MODE_PACK:
      if (m_NowShowing != DM_PACKING)
      {
        ReadAllTextControls ();
        SwitchToDisplayMode (DM_PACKING);
      }
      break;

    case MSG_DISPLAY_MODE_UNPACK:
      if (m_NowShowing != DM_UNPACKING)
      {
        ReadAllTextControls ();
        SwitchToDisplayMode (DM_UNPACKING);
      }
      break;

    case MSG_MAKE_RAM_DISK_CHECK:
      TempBool = (m_MakeRAMDiskCheckboxPntr->Value() == B_CONTROL_ON);
      if (m_MakeRAMDiskCachedValue != TempBool)
        SubmitCommandBool (PN_MAKE_RAM_DISK, B_SET_PROPERTY, TempBool);
      break;

    case MSG_PACK_FILE_NAME_BROWSE:
      BrowseForPackFile ();
      break;

    case MSG_RAM_VIRTUAL_MEMORY_CHECK:
      TempBool = (m_RAMDiskVirtualMemoryCheckboxPntr->Value() == B_CONTROL_ON);
      if (m_RAMDiskVirtualMemoryCachedValue != TempBool)
        SubmitCommandBool (PN_RAM_DISK_VIRTUAL_MEMORY,
          B_SET_PROPERTY, TempBool);
      break;

    case MSG_RECREATE_INDICES_CHECK:
      TempBool = (m_RecreateIndicesCheckboxPntr->Value() == B_CONTROL_ON);
      if (m_RecreateIndicesCachedValue != TempBool)
        SubmitCommandBool (PN_RECREATE_INDICES, B_SET_PROPERTY, TempBool);
      break;

    case MSG_TOP_LEVEL_PATH_BROWSE:
      BrowseForTopLevelDirectory ();
      break;

    case MSG_DO_OPERATION_OR_CANCEL:
      if (m_AppPntr->m_NowProcessingCurrentTotal < 0)
      {
        /* Currently the BApplication isn't busy packing or unpacking, so
        request that it start the operation implied by the current display
        mode. */

        BMessage CommandMessage (B_EXECUTE_PROPERTY);

        ReadAllTextControls ();
        CommandMessage.AddSpecifier (g_PropertyNames[
          (m_NowShowing == DM_PACKING) ? PN_PACK :
          ((m_NowShowing == DM_UNPACKING) ? PN_UNPACK :
          PN_CREATE_FILE_SYSTEM)]);
        SubmitCommand (CommandMessage);
      }
      else /* BApplication is currenty working, so this becomes a cancel. */
      {
        m_AppPntr->m_CancelRequestedByGUI = true;
      }
      break;

    default:
      BView::MessageReceived (MessagePntr);
  }
}


void ControlsView::NullifyChildViewPointers ()
{
  int i;

  m_AboutButtonPntr = NULL;
  m_DoOperationButtonPntr = NULL;
  m_InstallThingsButtonPntr = NULL;
  m_MakeRAMDiskCheckboxPntr = NULL;
  m_MaxRAMDiskSizeTextboxPntr = NULL;
  m_ModeCreateFSRadioButtonPntr = NULL;
  m_ModePackRadioButtonPntr = NULL;
  m_ModeUnpackRadioButtonPntr = NULL;
  m_NotesTextboxPntr = NULL;
  m_PackFileBrowseButtonPntr = NULL;
  m_PackFileNameTextboxPntr = NULL;
  m_ProgressArrowPntr = NULL;
  m_ProgressPathTextboxPntr = NULL;
  m_RAMDiskVirtualMemoryCheckboxPntr = NULL;
  m_RecreateIndicesCheckboxPntr = NULL;
  m_ResetToDefaultsButtonPntr = NULL;
  m_TopLevelDirectoryBrowseButtonPntr = NULL;
  m_TopLevelDirectoryNameTextboxPntr = NULL;
  m_ZLibCompressionLevelTextboxPntr = NULL;

  for (i = 0; i < CM_MAX; i++)
    m_CompressionMethodRadioButtons [i] = NULL;
}


/* Check the server for changes in the global state, and if there are any
changes, update the displayed values.  Since this is a read only examination of
the server, we go directly to the application rather than sending it messages.
Also, when sending messages, we can't find out what it is doing while it is
busy.  Instead, we lock the BApplication (so it isn't changing things while
we're looking) and retrieve our values. */

void ControlsView::PollServerForChanges ()
{
  float            CurrentProportion;
  OperButtonStates DesiredOperationButtonState;
  float            FinalTotalCopy;
  char             TempString [PATH_MAX];
  BWindow         *WindowPntr;

  /* We need a pointer to our window, for changing the title etc. */

  WindowPntr = Window ();
  if (WindowPntr == NULL)
    return; /* No window, no point in updating the display! */

  /* Check various read only values, which can be read from the BApplication
  without having to lock it (so you can show what it is doing in the main
  thread while it is busy doing it).  First, see what the operation button name
  should be displaying. */

  if (m_AppPntr->m_NowProcessingCurrentTotal < 0)
    /* BApplication isn't busy. */
    DesiredOperationButtonState = OBS_DO_OPERATION;
  else if (m_AppPntr->m_CancelRequestedByGUI)
    /* Cancel has already been requested. */
    DesiredOperationButtonState = OBS_NOW_CANCELLING;
  else
    /* BApplication is busy doing something, not cancelled yet. */
    DesiredOperationButtonState = OBS_CANCEL;

  if (m_DoOperationButtonPntr != NULL &&
  m_DoOperationButtonCachedState != DesiredOperationButtonState)
  {
    if (DesiredOperationButtonState == OBS_DO_OPERATION)
    {
      m_DoOperationButtonPntr->SetLabel (
        (m_NowShowing == DM_UNPACKING) ? "Do Unpacking" :
        ((m_NowShowing == DM_PACKING) ? "Do Packing" : "Create FS"));
      system_beep (g_BeepPackDone);
      WindowPntr->PostMessage (AFSPSettingsWindow::MSG_SLOWER_PULSE);
    }
    else if (DesiredOperationButtonState == OBS_CANCEL)
    {
      m_DoOperationButtonPntr->SetLabel ("Cancel");
      WindowPntr->PostMessage (AFSPSettingsWindow::MSG_FASTER_PULSE);
    }
    else
      m_DoOperationButtonPntr->SetLabel ("Cancelling...");
    m_DoOperationButtonCachedState = DesiredOperationButtonState;
  }

  /* Update the text box showing the file currently being worked on. */

  if (m_ProgressPathTextboxPntr != NULL && 0 !=
  strcmp (m_ProgressPathTextboxPntr->Text(), m_AppPntr->m_NowProcessingPath))
    m_ProgressPathTextboxPntr->SetText (m_AppPntr->m_NowProcessingPath);

  /* Update the progress graphical indicator.  Make a copy of volatile
  variables that might change and cause trouble with the change. */

  FinalTotalCopy = m_AppPntr->m_NowProcessingEstimatedFinalTotal;
  if (m_AppPntr->m_NowProcessingCurrentTotal <= 0)
    CurrentProportion = -1; /* Currently not processing. */
  else if (FinalTotalCopy <= 0)
    CurrentProportion = 0.5; /* Processing but no size estimate available. */
  else
    CurrentProportion =
    m_AppPntr->m_NowProcessingCurrentTotal / FinalTotalCopy;

  if (m_ProgressArrowPntr != NULL &&
  m_ProgressProportionCachedValue != CurrentProportion)
  {
    m_ProgressArrowPntr->SetProportion (CurrentProportion);
    m_ProgressProportionCachedValue = CurrentProportion;
  }

  /* Lock the application.  This will stop it from processing any further
  messages until we are done.  Or if it is busy, the lock will fail. */

  if (m_AppPntr->LockWithTimeout (100000) != B_OK)
    return; /* It's probably busy doing something. */

  /* See if the compression method has changed. */

  if (m_CompressionMethodRadioButtons [0] != NULL &&
  m_AppPntr->m_CompressionMethod != m_CompressionMethodCachedValue)
  {
    m_CompressionMethodCachedValue = m_AppPntr->m_CompressionMethod;
    m_CompressionMethodRadioButtons [m_CompressionMethodCachedValue]->
      SetValue (B_CONTROL_ON);
    if (m_ZLibCompressionLevelTextboxPntr != NULL)
      m_ZLibCompressionLevelTextboxPntr->SetEnabled (
        m_CompressionMethodCachedValue == CM_ZLIB);
  }

  /* See if the make RAM disk setting has changed. */

  if (m_MakeRAMDiskCheckboxPntr != NULL &&
  m_AppPntr->m_MakeRAMDisk != m_MakeRAMDiskCachedValue)
  {
    m_MakeRAMDiskCachedValue = m_AppPntr->m_MakeRAMDisk;
    m_MakeRAMDiskCheckboxPntr->SetValue (
      m_MakeRAMDiskCachedValue ? B_CONTROL_ON : B_CONTROL_OFF);

    /* Also enable or disable the RAM disk options if they are being displayed.
    Only relevant for unpacking mode, in Create File System mode the
    m_MakeRAMDiskCheckboxPntr is NULL so this code doesn't get executed and the
    options stay enabled all the time. */

    if (m_MaxRAMDiskSizeTextboxPntr != NULL)
      m_MaxRAMDiskSizeTextboxPntr->SetEnabled (m_MakeRAMDiskCachedValue);
    if (m_RAMDiskVirtualMemoryCheckboxPntr != NULL)
      m_RAMDiskVirtualMemoryCheckboxPntr->SetEnabled (
        m_MakeRAMDiskCachedValue);
  }

  /* See if the maximum RAM disk size has changed. */

  if (m_MaxRAMDiskSizeTextboxPntr != NULL &&
  m_AppPntr->m_MaxRAMDiskSize != m_MaxRAMDiskSizeCachedValue)
  {
    m_MaxRAMDiskSizeCachedValue = m_AppPntr->m_MaxRAMDiskSize;
    sprintf (TempString, "%ld", m_MaxRAMDiskSizeCachedValue);
    m_MaxRAMDiskSizeTextboxPntr->SetText (TempString);
  }

  /* See if the notes text has changed. */

  if (m_NotesTextboxPntr != NULL &&
  strcmp (m_AppPntr->m_Notes.String (), m_NotesCachedValue.String ()) != 0)
  {
    m_NotesCachedValue = m_AppPntr->m_Notes.String ();
    m_NotesTextboxPntr->SetText (m_NotesCachedValue.String ());
  }

  /* See if the pack file name has changed. */

  if (m_PackFileNameTextboxPntr != NULL &&
  strcmp (m_AppPntr->m_PackedFileName.Path (),
  m_PackFileNameCachedValue.String ()) != 0)
  {
    m_PackFileNameCachedValue =
      m_AppPntr->m_PackedFileName.Path ();
    m_PackFileNameTextboxPntr->SetText (
      m_PackFileNameCachedValue.String ());
    WindowPntr->SetTitle (m_PackFileNameCachedValue.String ());
  }


  /* See if the RAM disk virtual memory setting has changed. */

  if (m_RAMDiskVirtualMemoryCheckboxPntr != NULL &&
  m_AppPntr->m_RAMDiskVirtualMemory != m_RAMDiskVirtualMemoryCachedValue)
  {
    m_RAMDiskVirtualMemoryCachedValue = m_AppPntr->m_RAMDiskVirtualMemory;
    m_RAMDiskVirtualMemoryCheckboxPntr->SetValue (
      m_RAMDiskVirtualMemoryCachedValue ? B_CONTROL_ON : B_CONTROL_OFF);
  }

  /* See if the recreate indices checkbox has changed. */

  if (m_RecreateIndicesCheckboxPntr != NULL &&
  m_AppPntr->m_RecreateIndices != m_RecreateIndicesCachedValue)
  {
    m_RecreateIndicesCachedValue = m_AppPntr->m_RecreateIndices;
    m_RecreateIndicesCheckboxPntr->SetValue (
      m_RecreateIndicesCachedValue ? B_CONTROL_ON : B_CONTROL_OFF);
  }

  /* See if the top level directory name has changed. */

  if (m_TopLevelDirectoryNameTextboxPntr != NULL &&
  strcmp (m_AppPntr->m_TopLevelDirectoryName.Path (),
  m_TopLevelDirectoryNameCachedValue.String ()) != 0)
  {
    m_TopLevelDirectoryNameCachedValue =
      m_AppPntr->m_TopLevelDirectoryName.Path ();
    m_TopLevelDirectoryNameTextboxPntr->SetText (
      m_TopLevelDirectoryNameCachedValue.String ());
  }

  /* See if the ZLib compression level has changed. */

  if (m_ZLibCompressionLevelTextboxPntr != NULL &&
  m_AppPntr->m_zlibCompressionLevel != m_ZLibCompressionLevelCachedValue)
  {
    m_ZLibCompressionLevelCachedValue = m_AppPntr->m_zlibCompressionLevel;
    sprintf (TempString, "%lu", m_ZLibCompressionLevelCachedValue);
    m_ZLibCompressionLevelTextboxPntr->SetText (TempString);
  }

  /* Done.  Let the BApplication continue processing messages. */

  m_TextControlsContainInvalidData = false;
  m_AppPntr->Unlock ();
}


void ControlsView::Pulse ()
{
  if (system_time () > m_TimeOfLastPoll + 100000)
  {
    PollServerForChanges ();
    m_TimeOfLastPoll = system_time ();
  }
}


void ControlsView::ReadAllTextControls ()
{
  /* Since text controls don't send a notification until the user hits enter,
  sometimes the typed text would be lost unless we explicitly read it, which
  you can do by calling this function at the right moment. */

  if (m_TextControlsContainInvalidData)
    return;

  if (m_MaxRAMDiskSizeTextboxPntr != NULL &&
  m_MaxRAMDiskSizeCachedValue != atol (m_MaxRAMDiskSizeTextboxPntr->Text ()))
  {
    SubmitCommandInt32 (PN_MAX_RAM_DISK_SIZE, B_SET_PROPERTY,
      atol (m_MaxRAMDiskSizeTextboxPntr->Text ()));
    m_MaxRAMDiskSizeCachedValue = -19275261;
  }

  if (m_NotesTextboxPntr != NULL &&
  strcmp (m_NotesCachedValue.String (), m_NotesTextboxPntr->Text ()) != 0)
  {
    SubmitCommandString (PN_NOTES, B_SET_PROPERTY,
      m_NotesTextboxPntr->Text ());
    m_NotesCachedValue = g_UnknownText;
  }

  if (m_PackFileNameTextboxPntr != NULL &&
  strcmp (m_PackFileNameCachedValue.String (),
  m_PackFileNameTextboxPntr->Text ()) != 0)
  {
    SubmitCommandString (PN_PACKED_FILE, B_SET_PROPERTY,
      m_PackFileNameTextboxPntr->Text ());
    m_PackFileNameCachedValue = g_UnknownText;
  }

  if (m_TopLevelDirectoryNameTextboxPntr != NULL &&
  strcmp (m_TopLevelDirectoryNameCachedValue.String (),
  m_TopLevelDirectoryNameTextboxPntr->Text ()) != 0)
  {
    SubmitCommandString (PN_TOP_LEVEL_DIRECTORY, B_SET_PROPERTY,
      m_TopLevelDirectoryNameTextboxPntr->Text ());
    m_TopLevelDirectoryNameCachedValue = g_UnknownText;
  }

  if (m_ZLibCompressionLevelTextboxPntr != NULL &&
  m_ZLibCompressionLevelCachedValue !=
  (uint32) atol (m_ZLibCompressionLevelTextboxPntr->Text ()))
  {
    SubmitCommandInt32 (PN_ZLIB_COMPRESSION_LEVEL, B_SET_PROPERTY,
      atol (m_ZLibCompressionLevelTextboxPntr->Text ()));
    m_ZLibCompressionLevelCachedValue = 12834156;
  }
}


void ControlsView::SwitchToDisplayMode (DisplayModes NewMode)
{
  BBox         *BBoxPntr;
  BMessage      CommandMessage;
  int           CompressionMethodIndex;
  float         GapBetweenRadioButtons;
  float         Margin;
  BWindow      *WindowPntr;
  char         *RAMDiskMaxSizeTitle = "RAM disk maximum size (bytes):";
  float         RowTop;
  char         *StringPntr;
  BBox         *SubBBoxPntr;
  float         SubRowTop;
  float         SubSubRowTop;
  BView        *TempBViewPntr;
  BRadioButton *TempRadioButtonPntr;
  BRect         TempRect;
  float         X;
  char         *ZLibLevelTitle =
                  "ZLib Compression Level (0 for none, to 9 for maximum):";

  /* Remove all the current child views. */

  while ((TempBViewPntr = ChildAt (0)) != NULL)
  {
    RemoveChild (TempBViewPntr);
    delete TempBViewPntr;
  }
  NullifyChildViewPointers ();
  m_TextControlsContainInvalidData = true;
  m_NowShowing = NewMode;

  /* Set up the basic row size, and current row creation Y value. */

  TempRect = Bounds ();
  RowTop = TempRect.top;

  /* Make the Top Level Directory cluster.  It contains the browse button and a
  text box for showing or manually entering the top level directory path. */

  TempRect = Bounds ();
  TempRect.top = RowTop;
  TempRect.InsetBy (2, 2);
  TempRect.bottom = g_RowHeight + TempRect.top + g_StringViewHeight + 4;
  BBoxPntr = new BBox (TempRect, "TopLevelGroup",
    B_FOLLOW_LEFT_RIGHT | B_FOLLOW_TOP);
  AddChild (BBoxPntr);
  BBoxPntr->SetLabel ("Top Level Directory");
  RowTop = BBoxPntr->Frame().bottom + 1;
  SubRowTop = BBoxPntr->Bounds().top + g_StringViewHeight + 2;

  /* Top level directory browse button. */

  Margin = ceilf ((g_RowHeight - g_ButtonHeight) / 2);
  TempRect = BBoxPntr->Bounds ();
  TempRect.top = SubRowTop;
  TempRect.bottom = TempRect.top + g_ButtonHeight;
  TempRect.InsetBy (g_MarginBetweenControls, 0);
  m_TopLevelDirectoryBrowseButtonPntr = new BButton (TempRect,
    "TopLevelBrowse", "Browse", new BMessage (MSG_TOP_LEVEL_PATH_BROWSE));
  BBoxPntr->AddChild (m_TopLevelDirectoryBrowseButtonPntr);
  m_TopLevelDirectoryBrowseButtonPntr->SetTarget (this);
  m_TopLevelDirectoryBrowseButtonPntr->ResizeToPreferred ();
  X = m_TopLevelDirectoryBrowseButtonPntr->Frame().right +
    g_MarginBetweenControls;

  /* Top level directory text box. */

  Margin = ceilf ((g_RowHeight - g_TextBoxHeight) / 2);
  TempRect = BBoxPntr->Bounds ();
  TempRect.top = SubRowTop;
  TempRect.bottom = TempRect.top + g_TextBoxHeight;
  TempRect.InsetBy (g_MarginBetweenControls, 0);
  TempRect.left = X;

  StringPntr = "Top Level Directory:";
  m_TopLevelDirectoryNameCachedValue = g_UnknownText;
  m_TopLevelDirectoryNameTextboxPntr = new BTextControl (TempRect,
    "TopLevelPath",
    StringPntr /* label */,
    m_TopLevelDirectoryNameCachedValue.String () /* text */,
    new BMessage (MSG_TOP_LEVEL_PATH_TEXT),
    B_FOLLOW_LEFT_RIGHT | B_FOLLOW_TOP,
    B_WILL_DRAW | B_NAVIGABLE | B_NAVIGABLE_JUMP);
  BBoxPntr->AddChild (m_TopLevelDirectoryNameTextboxPntr);
  m_TopLevelDirectoryNameTextboxPntr->SetTarget (this);
  m_TopLevelDirectoryNameTextboxPntr->SetDivider (
    be_plain_font->StringWidth (StringPntr) + g_MarginBetweenControls);

  /* Make the operation cluster.  It has the mode buttons for Pack/Unpack, a
  progress indicator, a button for doing the operation, and various extra
  controls for options that depend on the particular operation. */

  TempRect = Bounds ();
  TempRect.top = RowTop;
  TempRect.InsetBy (2, 2);
  if (m_NowShowing == DM_UNPACKING)
    TempRect.bottom = TempRect.top + g_RowHeight * 6 +
      2 * /* boxes have a title and borders */ (g_StringViewHeight + 4);
  else if (m_NowShowing == DM_PACKING)
    TempRect.bottom = TempRect.top + g_RowHeight * 5 + g_StringViewHeight + 4;
  else /* DM_CREATE_FILE_SYSTEM */
    TempRect.bottom = TempRect.top + g_RowHeight * 4 + g_StringViewHeight + 4;
  BBoxPntr = new BBox (TempRect, "OperationGroup",
    B_FOLLOW_LEFT_RIGHT | B_FOLLOW_TOP);
  AddChild (BBoxPntr);
  BBoxPntr->SetLabel ((BString ("Operation - ") <<
    ((m_NowShowing == DM_UNPACKING) ? "Unpack an Archive File into a Directory" :
    ((m_NowShowing == DM_PACKING) ? "Pack Contents of a Directory into an Archive File" :
    "Create an Empty RAM File System Mounted at a Directory"))).String ()
  );
  RowTop = BBoxPntr->Frame().bottom + 1;
  SubRowTop = BBoxPntr->Bounds().top + g_StringViewHeight + 2;

  /* The Pack, Unpack and Create radio buttons.  Centered in the middle of the
  group's visible area (1/4 of the empty space above, below and between the
  buttons). */

  TempRect = BBoxPntr->Bounds ();
  TempRect.top = SubRowTop;
  TempRect.bottom -= g_RowHeight /* current path box below radio area */ +
    4 /* don't include group box border */;
  GapBetweenRadioButtons =
    ceilf ((TempRect.Height() - 3 * g_RadioButtonHeight) / 4);
  TempRect.top += GapBetweenRadioButtons;
  TempRect.bottom = TempRect.top + g_RadioButtonHeight;
  TempRect.InsetBy (g_MarginBetweenControls, 0);
  m_ModePackRadioButtonPntr = new BRadioButton (TempRect,
    "ModePack", "Pack", new BMessage (MSG_DISPLAY_MODE_PACK));
  BBoxPntr->AddChild (m_ModePackRadioButtonPntr);
  m_ModePackRadioButtonPntr->SetTarget (this);
  m_ModePackRadioButtonPntr->ResizeToPreferred ();
  m_ModePackRadioButtonPntr->SetValue (m_NowShowing == DM_PACKING);

  TempRect.top = TempRect.bottom + GapBetweenRadioButtons;
  TempRect.bottom = TempRect.top + g_RadioButtonHeight;
  m_ModeUnpackRadioButtonPntr = new BRadioButton (TempRect,
    "ModeUnpack", "Unpack", new BMessage (MSG_DISPLAY_MODE_UNPACK));
  BBoxPntr->AddChild (m_ModeUnpackRadioButtonPntr);
  m_ModeUnpackRadioButtonPntr->SetTarget (this);
  m_ModeUnpackRadioButtonPntr->ResizeToPreferred ();
  m_ModeUnpackRadioButtonPntr->SetValue (m_NowShowing == DM_UNPACKING);

  TempRect.top = TempRect.bottom + GapBetweenRadioButtons;
  TempRect.bottom = TempRect.top + g_RadioButtonHeight;
  m_ModeCreateFSRadioButtonPntr = new BRadioButton (TempRect,
    "ModeCreateFS", "Create", new BMessage (MSG_DISPLAY_MODE_CREATE_FS));
  BBoxPntr->AddChild (m_ModeCreateFSRadioButtonPntr);
  m_ModeCreateFSRadioButtonPntr->SetTarget (this);
  m_ModeCreateFSRadioButtonPntr->ResizeToPreferred ();
  m_ModeCreateFSRadioButtonPntr->SetValue (
    m_NowShowing == DM_CREATE_FILE_SYSTEM);

  X = m_ModeUnpackRadioButtonPntr->Frame().right +
    g_MarginBetweenControls;

  /* The progress arrow. */

  TempRect = BBoxPntr->Bounds ();
  TempRect.bottom -= 4 /* Don't overwrite the group box border */ +
    g_RowHeight;
  TempRect.left = X;
  TempRect.right = TempRect.left + g_MarginBetweenControls * 10;
  if (m_NowShowing == DM_PACKING)
    TempRect.top = SubRowTop + g_RowHeight;
  else /* Arrow is pointing up, Unpacking or Create button is below arrow. */
  {
    TempRect.top = SubRowTop;
    TempRect.bottom -= g_RowHeight;
  }
  m_ProgressProportionCachedValue = -12345;
  m_ProgressArrowPntr = new ProgressArrow (TempRect,
    m_NowShowing != DM_PACKING /* TRUE for Pointing Up */);
  BBoxPntr->AddChild (m_ProgressArrowPntr);

  /* The Do Operation / Cancel button. */

  Margin = ceilf ((g_RowHeight - g_ButtonHeight) / 2);
  TempRect = BBoxPntr->Bounds ();
  TempRect.bottom -= 4 /* Don't overwrite the group box border */ +
    g_RowHeight;
  TempRect.left = X;
  TempRect.right = TempRect.left + g_MarginBetweenControls * 10;
  if (m_NowShowing != DM_PACKING)
    TempRect.top = TempRect.bottom - g_RowHeight + Margin;
  else /* Packing, arrow is pointing down. */
    TempRect.top = SubRowTop + Margin;
  TempRect.bottom = TempRect.top + g_ButtonHeight;
  m_DoOperationButtonCachedState = OBS_DO_OPERATION;
  m_DoOperationButtonPntr = new BButton (TempRect,
    "DoOp", (m_NowShowing == DM_UNPACKING) ? "Do Unpacking" :
    ((m_NowShowing == DM_PACKING) ? "Do Packing" : "Create FS"),
    new BMessage (MSG_DO_OPERATION_OR_CANCEL));
  BBoxPntr->AddChild (m_DoOperationButtonPntr);
  m_DoOperationButtonPntr->SetTarget (this);

  /* The text line displaying the currently processed file name, covers most of
  the bottom row of the cluster. */

  Margin = ceilf ((g_RowHeight - g_StringViewHeight) / 2);
  TempRect = BBoxPntr->Bounds ();
  TempRect.InsetBy (g_MarginBetweenControls, 0);
  TempRect.bottom -= 4; /* Don't overwrite the group box border. */
  TempRect.top = TempRect.bottom - g_RowHeight + Margin;
  m_ProgressPathTextboxPntr = new BTextControl (TempRect,
    "ProgressPath", "Now processing:", g_UnknownText, NULL,
    B_FOLLOW_LEFT_RIGHT | B_FOLLOW_TOP);
  BBoxPntr->AddChild (m_ProgressPathTextboxPntr);
  m_ProgressPathTextboxPntr->SetTarget (this);
  m_ProgressPathTextboxPntr->SetDivider (be_plain_font->StringWidth (
    m_ProgressPathTextboxPntr->Label ()) + g_MarginBetweenControls);
  m_ProgressPathTextboxPntr->SetEnabled (false);

  X = m_ProgressArrowPntr->Frame().right + g_MarginBetweenControls;

  /* The Compression Method subgroup in the operation group, which exists when
  packing.  Has radio buttons for the methods, and underneath is the ZLib
  compression level. */

  if (m_NowShowing == DM_PACKING)
  {
    StringPntr = "Compression Method";
    TempRect = BBoxPntr->Bounds ();
    TempRect.left = X + 2;
    TempRect.right = TempRect.left +
      be_plain_font->StringWidth (ZLibLevelTitle) +
      g_MarginBetweenControls * 8;
    TempRect.top = SubRowTop + 2;
    TempRect.bottom = TempRect.top + g_RowHeight * CM_MAX +
      g_RowHeight + g_StringViewHeight + 4;
    SubBBoxPntr = new BBox (TempRect, "MethodGroup");
    BBoxPntr->AddChild (SubBBoxPntr);
    SubBBoxPntr->SetLabel (StringPntr);
    SubRowTop = SubBBoxPntr->Frame().bottom + 1;
    SubSubRowTop = SubBBoxPntr->Bounds().top + g_StringViewHeight + 2;

    m_CompressionMethodCachedValue = CM_MAX;
    for (CompressionMethodIndex = 0; CompressionMethodIndex < CM_MAX;
    CompressionMethodIndex++)
    {
      Margin = ceilf ((g_RowHeight - g_RadioButtonHeight) / 2);
      TempRect = SubBBoxPntr->Bounds ();
      TempRect.top = SubSubRowTop + Margin;
      TempRect.bottom = TempRect.top + g_RadioButtonHeight;
      TempRect.InsetBy (g_MarginBetweenControls, 0);
      TempRadioButtonPntr = new BRadioButton (TempRect,
        g_CompressionMethodNames [CompressionMethodIndex],
        g_CompressionMethodNames [CompressionMethodIndex],
        new BMessage (MSG_COMPRESSION_METHOD + CompressionMethodIndex));
      m_CompressionMethodRadioButtons [CompressionMethodIndex] =
        TempRadioButtonPntr;
      SubBBoxPntr->AddChild (TempRadioButtonPntr);
      TempRadioButtonPntr->SetTarget (this);
      SubSubRowTop += g_RowHeight;
    }

    /* ZLib compression level. */

    Margin = ceilf ((g_RowHeight - g_TextBoxHeight) / 2);
    TempRect = SubBBoxPntr->Bounds ();
    TempRect.InsetBy (g_MarginBetweenControls, 0);
    TempRect.top = SubSubRowTop + Margin;
    TempRect.bottom = TempRect.top + g_TextBoxHeight;
    StringPntr = ZLibLevelTitle;
    m_ZLibCompressionLevelCachedValue = 123456;
    m_ZLibCompressionLevelTextboxPntr = new BTextControl (TempRect,
      "ZLibLevel",
      StringPntr /* label */,
      g_UnknownText /* text */,
      new BMessage (MSG_ZLIB_LEVEL_TEXT));
    SubBBoxPntr->AddChild (m_ZLibCompressionLevelTextboxPntr);
    m_ZLibCompressionLevelTextboxPntr->SetTarget (this);
    m_ZLibCompressionLevelTextboxPntr->SetDivider (
      be_plain_font->StringWidth (StringPntr) + g_MarginBetweenControls);
  }

  /* The Unpacking Options subgroup in the operation group.  Contains a couple
  of checkboxes. */

  if (m_NowShowing == DM_UNPACKING)
  {
    StringPntr = "Unpacking Options";
    TempRect.left = X + 2;
    TempRect.right = TempRect.left +
      be_plain_font->StringWidth (RAMDiskMaxSizeTitle) +
      g_MarginBetweenControls * 12; /* Matches width of RAM FS options. */
    TempRect.top = SubRowTop + 2;
    TempRect.bottom = TempRect.top + g_RowHeight * 2 + g_StringViewHeight + 4;
    SubBBoxPntr = new BBox (TempRect, "UnpackingGroup");
    BBoxPntr->AddChild (SubBBoxPntr);
    SubBBoxPntr->SetLabel (StringPntr);
    SubRowTop = SubBBoxPntr->Frame().bottom + 1;
    SubSubRowTop = SubBBoxPntr->Bounds().top + g_StringViewHeight + 2;

    /* The Recreate Indices checkbox. */

    Margin = ceilf ((g_RowHeight - g_CheckBoxHeight) / 2);
    TempRect = SubBBoxPntr->Bounds ();
    TempRect.top = SubSubRowTop + Margin;
    TempRect.InsetBy (g_MarginBetweenControls, 0);
    m_RecreateIndicesCachedValue = false;
    m_RecreateIndicesCheckboxPntr = new BCheckBox (TempRect,
      "RecreateIndices", "Recreate indices",
      new BMessage (MSG_RECREATE_INDICES_CHECK));
    SubBBoxPntr->AddChild (m_RecreateIndicesCheckboxPntr);
    m_RecreateIndicesCheckboxPntr->SetTarget (this);
    m_RecreateIndicesCheckboxPntr->ResizeToPreferred ();
    SubSubRowTop += g_RowHeight;

    /* The Make RAM Disk checkbox. */

    Margin = ceilf ((g_RowHeight - g_CheckBoxHeight) / 2);
    TempRect = SubBBoxPntr->Bounds ();
    TempRect.top = SubSubRowTop + Margin;
    TempRect.InsetBy (g_MarginBetweenControls, 0);
    m_MakeRAMDiskCachedValue = true; /* Reflect RAM options initial enable. */
    m_MakeRAMDiskCheckboxPntr = new BCheckBox (TempRect,
      "MakeRAMDisk", "Create AGMSRAMFileSystem RAM disk",
      new BMessage (MSG_MAKE_RAM_DISK_CHECK));
    SubBBoxPntr->AddChild (m_MakeRAMDiskCheckboxPntr);
    m_MakeRAMDiskCheckboxPntr->SetTarget (this);
    m_MakeRAMDiskCheckboxPntr->ResizeToPreferred ();
    m_MakeRAMDiskCheckboxPntr->SetValue (B_CONTROL_ON);
    SubSubRowTop += g_RowHeight;
  }

  /* The RAM File System Options subgroup.  Used when unpacking and when
  creating an empty file system. */

  if (m_NowShowing == DM_UNPACKING || m_NowShowing == DM_CREATE_FILE_SYSTEM)
  {
    StringPntr = "RAM File System Options";
    TempRect.left = X + 2;
    TempRect.right = TempRect.left +
      be_plain_font->StringWidth (RAMDiskMaxSizeTitle) +
      g_MarginBetweenControls * 12;
    TempRect.top = SubRowTop + 2;
    TempRect.bottom = TempRect.top + g_RowHeight * 2 + g_StringViewHeight + 4;
    SubBBoxPntr = new BBox (TempRect, "RAMOptionsGroup");
    BBoxPntr->AddChild (SubBBoxPntr);
    SubBBoxPntr->SetLabel (StringPntr);
    SubRowTop = SubBBoxPntr->Frame().bottom + 1;
    SubSubRowTop = SubBBoxPntr->Bounds().top + g_StringViewHeight + 2;

    /* RAM disk maximum size. */

    Margin = ceilf ((g_RowHeight - g_TextBoxHeight) / 2);
    TempRect = SubBBoxPntr->Bounds ();
    TempRect.InsetBy (g_MarginBetweenControls, 0);
    TempRect.top = SubSubRowTop + Margin;
    TempRect.bottom = TempRect.top + g_TextBoxHeight;
    StringPntr = RAMDiskMaxSizeTitle;
    m_MaxRAMDiskSizeCachedValue = -12345;
    m_MaxRAMDiskSizeTextboxPntr = new BTextControl (TempRect,
      "MaxRAMDisk",
      StringPntr /* label */,
      g_UnknownText /* text */,
      new BMessage (MSG_MAX_RAM_DISK_SIZE_TEXT));
    SubBBoxPntr->AddChild (m_MaxRAMDiskSizeTextboxPntr);
    m_MaxRAMDiskSizeTextboxPntr->SetTarget (this);
    m_MaxRAMDiskSizeTextboxPntr->SetDivider (
      be_plain_font->StringWidth (StringPntr) + g_MarginBetweenControls);
    SubSubRowTop += g_RowHeight;

    /* The RAM disk virtual memory checkbox. */

    Margin = ceilf ((g_RowHeight - g_CheckBoxHeight) / 2);
    TempRect = SubBBoxPntr->Bounds ();
    TempRect.top = SubSubRowTop + Margin;
    TempRect.InsetBy (g_MarginBetweenControls, 0);
    m_RAMDiskVirtualMemoryCachedValue = false;
    m_RAMDiskVirtualMemoryCheckboxPntr = new BCheckBox (TempRect,
      "RAMVM", "Use virtual memory for RAM disk",
      new BMessage (MSG_RAM_VIRTUAL_MEMORY_CHECK));
    SubBBoxPntr->AddChild (m_RAMDiskVirtualMemoryCheckboxPntr);
    m_RAMDiskVirtualMemoryCheckboxPntr->SetTarget (this);
    m_RAMDiskVirtualMemoryCheckboxPntr->ResizeToPreferred ();
    SubSubRowTop += g_RowHeight;
  }

  /* Make the Packed File System File cluster.  It contains the browse button
  and a text box for showing or manually entering the packed file path. */

  if (m_NowShowing != DM_CREATE_FILE_SYSTEM)
  {
    TempRect = Bounds ();
    TempRect.top = RowTop;
    TempRect.InsetBy (2, 2);
    TempRect.bottom = 2 * g_RowHeight + TempRect.top + g_StringViewHeight + 4;
    BBoxPntr = new BBox (TempRect, "PackedFileGroup",
      B_FOLLOW_LEFT_RIGHT | B_FOLLOW_TOP);
    AddChild (BBoxPntr);
    BBoxPntr->SetLabel ("Packed File System Archive File");
    RowTop = BBoxPntr->Frame().bottom + 1;
    SubRowTop = BBoxPntr->Bounds().top + g_StringViewHeight + 2;

    /* Packed file browse button. */

    Margin = ceilf ((g_RowHeight - g_ButtonHeight) / 2);
    TempRect = BBoxPntr->Bounds ();
    TempRect.top = SubRowTop;
    TempRect.bottom = TempRect.top + g_ButtonHeight;
    TempRect.InsetBy (g_MarginBetweenControls, 0);
    m_PackFileBrowseButtonPntr = new BButton (TempRect,
      "PackedFileBrowse", "Browse", new BMessage (MSG_PACK_FILE_NAME_BROWSE));
    BBoxPntr->AddChild (m_PackFileBrowseButtonPntr);
    m_PackFileBrowseButtonPntr->SetTarget (this);
    m_PackFileBrowseButtonPntr->ResizeToPreferred ();
    X = m_PackFileBrowseButtonPntr->Frame().right +
      g_MarginBetweenControls;

    /* Packed file path text box. */

    Margin = ceilf ((g_RowHeight - g_TextBoxHeight) / 2);
    TempRect = BBoxPntr->Bounds ();
    TempRect.top = SubRowTop;
    TempRect.bottom = TempRect.top + g_TextBoxHeight;
    TempRect.InsetBy (g_MarginBetweenControls, 0);
    TempRect.left = X;

    StringPntr = "Packed file:";
    m_PackFileNameCachedValue = g_UnknownText;
    m_PackFileNameTextboxPntr = new BTextControl (TempRect,
      "PackedFilePath",
      StringPntr /* label */,
      m_PackFileNameCachedValue.String () /* text */,
      new BMessage (MSG_PACK_FILE_NAME_TEXT),
      B_FOLLOW_LEFT_RIGHT | B_FOLLOW_TOP,
      B_WILL_DRAW | B_NAVIGABLE | B_NAVIGABLE_JUMP);
    BBoxPntr->AddChild (m_PackFileNameTextboxPntr);
    m_PackFileNameTextboxPntr->SetTarget (this);
    m_PackFileNameTextboxPntr->SetDivider (
      be_plain_font->StringWidth (StringPntr) + g_MarginBetweenControls);

    SubRowTop += g_RowHeight;

    /* Notes text box. */

    Margin = ceilf ((g_RowHeight - g_TextBoxHeight) / 2);
    TempRect = BBoxPntr->Bounds ();
    TempRect.top = SubRowTop;
    TempRect.bottom = TempRect.top + g_TextBoxHeight;
    TempRect.InsetBy (g_MarginBetweenControls, 0);

    StringPntr = "Notes:";
    m_NotesCachedValue = g_UnknownText;
    m_NotesTextboxPntr = new BTextControl (TempRect,
      "Notes",
      StringPntr /* label */,
      m_NotesCachedValue.String () /* text */,
      new BMessage (MSG_NOTES_TEXT),
      B_FOLLOW_LEFT_RIGHT | B_FOLLOW_TOP,
      B_WILL_DRAW | B_NAVIGABLE | B_NAVIGABLE_JUMP);
    BBoxPntr->AddChild (m_NotesTextboxPntr);
    m_NotesTextboxPntr->SetTarget (this);
    m_NotesTextboxPntr->SetDivider (
      be_plain_font->StringWidth (StringPntr) + g_MarginBetweenControls);
  }

  /* Make the Miscellaneous controls cluster.  It contains the rarely used
  buttons, separated from the other ones to avoid confusing the user. */

  TempRect = Bounds ();
  TempRect.top = RowTop;
  TempRect.InsetBy (2, 2);
  TempRect.bottom = g_RowHeight + TempRect.top + g_StringViewHeight + 4;
  BBoxPntr = new BBox (TempRect, "MiscGroup",
    B_FOLLOW_LEFT_RIGHT | B_FOLLOW_TOP);
  AddChild (BBoxPntr);
  BBoxPntr->SetLabel ("Miscellaneous");
  RowTop = BBoxPntr->Frame().bottom + 1;
  SubRowTop = BBoxPntr->Bounds().top + g_StringViewHeight + 2;

  /* The About button. */

  Margin = ceilf ((g_RowHeight - g_ButtonHeight) / 2);
  TempRect = BBoxPntr->Bounds ();
  TempRect.top = SubRowTop;
  TempRect.bottom = TempRect.top + g_ButtonHeight;
  TempRect.InsetBy (g_MarginBetweenControls, 0);
  m_AboutButtonPntr = new BButton (TempRect, "AboutButton",
    "About this fine program...",  new BMessage (B_ABOUT_REQUESTED));
  BBoxPntr->AddChild (m_AboutButtonPntr);
  m_AboutButtonPntr->SetTarget (be_app);
  m_AboutButtonPntr->ResizeToPreferred ();
  X = m_AboutButtonPntr->Frame().right + g_MarginBetweenControls;

  /* Install MIME Types button. */

  Margin = ceilf ((g_RowHeight - g_ButtonHeight) / 2);
  TempRect = BBoxPntr->Bounds ();
  TempRect.top = SubRowTop;
  TempRect.bottom = TempRect.top + g_ButtonHeight;
  TempRect.left = X;
  CommandMessage.MakeEmpty ();
  CommandMessage.what = B_EXECUTE_PROPERTY;
  CommandMessage.AddSpecifier (g_PropertyNames[PN_INSTALL_THINGS]);
  m_InstallThingsButtonPntr = new BButton (TempRect, "Install Button",
    "Install MIME Types", new BMessage (CommandMessage));
  BBoxPntr->AddChild (m_InstallThingsButtonPntr);
  m_InstallThingsButtonPntr->SetTarget (be_app);
  m_InstallThingsButtonPntr->ResizeToPreferred ();
  X = m_InstallThingsButtonPntr->Frame().right + g_MarginBetweenControls;

  /* The Reset to Defaults button. */

  Margin = ceilf ((g_RowHeight - g_ButtonHeight) / 2);
  TempRect = BBoxPntr->Bounds ();
  TempRect.top = SubRowTop;
  TempRect.bottom = TempRect.top + g_ButtonHeight;
  TempRect.left = X;
  CommandMessage.MakeEmpty ();
  CommandMessage.what = B_EXECUTE_PROPERTY;
  CommandMessage.AddSpecifier (g_PropertyNames[PN_RESET_TO_DEFAULTS]);
  m_ResetToDefaultsButtonPntr = new BButton (TempRect, "Reset Button",
    "Default Settings", new BMessage (CommandMessage));
  BBoxPntr->AddChild (m_ResetToDefaultsButtonPntr);
  m_ResetToDefaultsButtonPntr->SetTarget (be_app);
  m_ResetToDefaultsButtonPntr->ResizeToPreferred ();
  X = m_ResetToDefaultsButtonPntr->Frame().right + g_MarginBetweenControls;


  WindowPntr = Window ();
  if (WindowPntr != NULL)
  {
    /* Change the window height to include the space used by the controls. */

    WindowPntr->ResizeTo (WindowPntr->Frame().Width(), RowTop + 4);

    /* For some reason the window forgets the pulse rate after adding controls
    the second time (probably the text controls set it to their favorite rate).
    Reset it. */

    WindowPntr->PostMessage (AFSPSettingsWindow::MSG_SLOWER_PULSE);
  }
}



/******************************************************************************
 * Implementation of the AFSPSettingsWindow class, constructor, destructor and the
 * rest of the member functions in mostly alphabetical order.
 */

AFSPSettingsWindow::AFSPSettingsWindow ()
: BWindow (BRect (30, g_RowHeight,
    30 + g_MarginBetweenControls * 63, g_RowHeight * 15),
    "Alexander G. M. Smith's File System Packer",
    B_DOCUMENT_WINDOW,
    B_ASYNCHRONOUS_CONTROLS | B_NOT_V_RESIZABLE,
    B_CURRENT_WORKSPACE)
{
  m_ControlsViewPntr = new ControlsView (Bounds ());
  AddChild (m_ControlsViewPntr);
}


void AFSPSettingsWindow::MessageReceived (BMessage *MessagePntr)
{
  if (MessagePntr->what == MSG_FASTER_PULSE)
  {
    SetPulseRate (100000);
    return;
  }

  if (MessagePntr->what == MSG_SLOWER_PULSE)
  {
    SetPulseRate (500000);
    return;
  }

  return BWindow::MessageReceived (MessagePntr);
}


bool AFSPSettingsWindow::QuitRequested ()
{
  be_app->PostMessage (B_QUIT_REQUESTED);
  return true;
}



/******************************************************************************
 * Finally, the main program which drives it all.
 */

int main (int argc, char**)
{
  g_CommandLineMode = (argc > 1);
  if (!g_CommandLineMode)
    cout << PrintUsage; /* In case no arguments specified. */

  g_CommanderLooperPntr = new CommanderLooper;
  if (g_CommanderLooperPntr != NULL)
  {
    g_CommanderMessenger = new BMessenger (NULL, g_CommanderLooperPntr);
    g_CommanderLooperPntr->Run ();
  }

  AFSPApp MyApp;

  if (MyApp.InitCheck () == 0)
  {
    MyApp.LoadSaveSettings (true /* DoLoad */);
    MyApp.Run ();
  }

  if (g_CommanderLooperPntr != NULL)
  {
    g_CommanderLooperPntr->PostMessage (B_QUIT_REQUESTED);
    snooze (100000); /* Let the CommanderLooper thread run so it quits. */
  }

  printf ("AGMSFileSystemPacker shutting down...\n");
  return 0;
}
